@@ -33,10 +33,10 @@ extern "C" {
 #include <math.h>
 
 #include <alsa/asoundlib.h>
+#include <alsa/use-case.h>
 
 #include <spa/support/plugin.h>
 #include <spa/support/loop.h>
-#include <spa/support/log.h>
 #include <spa/utils/list.h>
 #include <spa/utils/json.h>
 
@@ -45,13 +45,16 @@ extern "C" {
 #include <spa/node/io.h>
 #include <spa/debug/types.h>
 #include <spa/param/param.h>
+#include <spa/param/latency-utils.h>
 #include <spa/param/audio/format-utils.h>
 
+#include "alsa.h"
+
 #include "dll.h"
 
-#define MIN_LATENCY	16
-#define MAX_LATENCY	8192
+#define MAX_RATES	16
 
+#define DEFAULT_PERIOD		1024u
 #define DEFAULT_RATE		48000u
 #define DEFAULT_CHANNELS	2u
 #define DEFAULT_USE_CHMAP	false
@@ -60,8 +63,6 @@ struct props {
 	char device[64];
 	char device_name[128];
 	char card_name[128];
-	uint32_t min_latency;
-	uint32_t max_latency;
 	bool use_chmap;
 };
 
@@ -85,6 +86,18 @@ struct channel_map {
 	uint32_t channels;
 	uint32_t pos[SPA_AUDIO_MAX_CHANNELS];
 };
+
+
+struct card {
+	struct spa_list link;
+	int ref;
+	uint32_t index;
+	snd_use_case_mgr_t *ucm;
+	char *ucm_prefix;
+	int format_ref;
+	uint32_t rate;
+};
+
 struct state {
 	struct spa_handle handle;
 	struct spa_node node;
@@ -93,6 +106,8 @@ struct state {
 	struct spa_system *data_system;
 	struct spa_loop *data_loop;
 
+	uint32_t card_index;
+	struct card *card;
 	snd_pcm_stream_t stream;
 	snd_output_t *output;
 
@@ -101,25 +116,34 @@ struct state {
 
 	uint64_t info_all;
 	struct spa_node_info info;
-	struct spa_param_info params[8];
+#define NODE_PropInfo		0
+#define NODE_Props		1
+#define NODE_IO			2
+#define NODE_ProcessLatency	3
+#define N_NODE_PARAMS		4
+	struct spa_param_info params[N_NODE_PARAMS];
 	struct props props;
 
 	bool opened;
 	snd_pcm_t *hndl;
-	int card;
 
 	bool have_format;
 	struct spa_audio_info current_format;
 
 	uint32_t default_period_size;
+	uint32_t default_period_num;
 	uint32_t default_headroom;
 	uint32_t default_start_delay;
 	uint32_t default_format;
 	unsigned int default_channels;
 	unsigned int default_rate;
+	uint32_t allowed_rates[MAX_RATES];
+	uint32_t n_allowed_rates;
 	struct channel_map default_pos;
 	unsigned int disable_mmap;
 	unsigned int disable_batch;
+	char clock_name[64];
+	uint32_t quantum_limit;
 
 	snd_pcm_uframes_t buffer_frames;
 	snd_pcm_uframes_t period_frames;
@@ -128,13 +152,21 @@ struct state {
 	int channels;
 	size_t frame_size;
 	int blocks;
-	int rate_denom;
+	uint32_t rate_denom;
 	uint32_t delay;
 	uint32_t read_size;
 
 	uint64_t port_info_all;
 	struct spa_port_info port_info;
-	struct spa_param_info port_params[8];
+#define PORT_EnumFormat		0
+#define PORT_Meta		1
+#define PORT_IO			2
+#define PORT_Format		3
+#define PORT_Buffers		4
+#define PORT_Latency		5
+#define N_PORT_PARAMS		6
+	struct spa_param_info port_params[N_PORT_PARAMS];
+	enum spa_direction port_direction;
 	struct spa_io_buffers *io;
 	struct spa_io_clock *clock;
 	struct spa_io_position *position;
@@ -157,8 +189,6 @@ struct state {
 	uint32_t start_delay;
 
 	uint32_t duration;
-	uint32_t last_duration;
-	uint64_t last_position;
 	unsigned int alsa_started:1;
 	unsigned int alsa_sync:1;
 	unsigned int alsa_recovering:1;
@@ -167,11 +197,17 @@ struct state {
 	unsigned int resample:1;
 	unsigned int use_mmap:1;
 	unsigned int planar:1;
+	unsigned int freewheel:1;
+	unsigned int open_ucm:1;
+	unsigned int is_iec958:1;
+	unsigned int is_hdmi:1;
+	unsigned int multi_rate:1;
+
+	uint64_t iec958_codecs;
 
 	int64_t sample_count;
 
 	int64_t sample_time;
-	uint64_t current_time;
 	uint64_t next_time;
 	uint64_t base_time;
 
@@ -179,23 +215,34 @@ struct state {
 
 	struct spa_dll dll;
 	double max_error;
+
+	struct spa_latency_info latency[2];
+	struct spa_process_latency_info process_latency;
 };
 
-int
-spa_alsa_enum_format(struct state *state, int seq,
+struct spa_pod *spa_alsa_enum_propinfo(struct state *state,
+		uint32_t idx, struct spa_pod_builder *b);
+int spa_alsa_add_prop_params(struct state *state, struct spa_pod_builder *b);
+int spa_alsa_parse_prop_params(struct state *state, struct spa_pod *params);
+
+int spa_alsa_enum_format(struct state *state, int seq,
 		     uint32_t start, uint32_t num,
 		     const struct spa_pod *filter);
 
 int spa_alsa_set_format(struct state *state, struct spa_audio_info *info, uint32_t flags);
 
-int spa_alsa_open(struct state *state);
+int spa_alsa_init(struct state *state, const struct spa_dict *info);
+int spa_alsa_clear(struct state *state);
+
+int spa_alsa_open(struct state *state, const char *params);
 int spa_alsa_start(struct state *state);
 int spa_alsa_reassign_follower(struct state *state);
 int spa_alsa_pause(struct state *state);
 int spa_alsa_close(struct state *state);
 
 int spa_alsa_write(struct state *state);
-int spa_alsa_read(struct state *state, snd_pcm_uframes_t silence);
+int spa_alsa_read(struct state *state);
+int spa_alsa_skip(struct state *state);
 
 void spa_alsa_recycle_buffer(struct state *state, uint32_t buffer_id);
 
@@ -235,6 +282,62 @@ static inline void spa_alsa_parse_position(struct channel_map *map, const char *
 	}
 }
 
+static inline uint32_t spa_alsa_parse_rates(uint32_t *rates, uint32_t max, const char *val, size_t len)
+{
+	struct spa_json it[2];
+	char v[256];
+	uint32_t count;
+
+	spa_json_init(&it[0], val, len);
+        if (spa_json_enter_array(&it[0], &it[1]) <= 0)
+                spa_json_init(&it[1], val, len);
+
+	count = 0;
+	while (spa_json_get_string(&it[1], v, sizeof(v)) > 0 && count < max)
+		rates[count++] = atoi(v);
+	return count;
+}
+
+static inline uint32_t spa_alsa_iec958_codec_from_name(const char *name)
+{
+	int i;
+	for (i = 0; spa_type_audio_iec958_codec[i].name; i++) {
+		if (strcmp(name, spa_debug_type_short_name(spa_type_audio_iec958_codec[i].name)) == 0)
+			return spa_type_audio_iec958_codec[i].type;
+	}
+	return SPA_AUDIO_IEC958_CODEC_UNKNOWN;
+}
+
+static inline void spa_alsa_parse_iec958_codecs(uint64_t *codecs, const char *val, size_t len)
+{
+	struct spa_json it[2];
+	char v[256];
+
+	spa_json_init(&it[0], val, len);
+        if (spa_json_enter_array(&it[0], &it[1]) <= 0)
+                spa_json_init(&it[1], val, len);
+
+	*codecs = 0;
+	while (spa_json_get_string(&it[1], v, sizeof(v)) > 0)
+		*codecs |= 1ULL << spa_alsa_iec958_codec_from_name(v);
+}
+
+static inline uint32_t spa_alsa_get_iec958_codecs(struct state *state, uint32_t *codecs,
+		uint32_t max_codecs)
+{
+	uint64_t mask = state->iec958_codecs;
+	uint32_t i = 0, j = 0;
+	if (!(state->is_iec958 || state->is_hdmi))
+		return 0;
+	while (mask && i < max_codecs) {
+		if (mask & 1)
+			codecs[i++] = j;
+		mask >>= 1;
+		j++;
+	}
+	return i;
+}
+
 #ifdef __cplusplus
 } /* extern "C" */
 #endif
