@@ -32,15 +32,22 @@
 #include <spa/node/keys.h>
 #include <spa/utils/names.h>
 #include <spa/utils/result.h>
+#include <spa/utils/string.h>
 #include <spa/buffer/alloc.h>
 #include <spa/pod/parser.h>
 #include <spa/pod/filter.h>
+#include <spa/pod/dynamic.h>
 #include <spa/param/param.h>
 #include <spa/param/audio/format-utils.h>
+#include <spa/param/latency-utils.h>
 #include <spa/debug/format.h>
 #include <spa/debug/pod.h>
 
-#define NAME "audioadapter"
+#undef SPA_LOG_TOPIC_DEFAULT
+#define SPA_LOG_TOPIC_DEFAULT log_topic
+static struct spa_log_topic *log_topic = &SPA_LOG_TOPIC(0, "spa.audioadapter");
+
+#define DEFAULT_ALIGN	16
 
 #define MAX_PORTS	SPA_AUDIO_MAX_CHANNELS
 
@@ -62,6 +69,7 @@ struct impl {
 	struct spa_hook follower_listener;
 	uint32_t follower_flags;
 	struct spa_audio_info follower_current_format;
+	struct spa_audio_info default_format;
 
 	struct spa_handle *hnd_convert;
 	struct spa_node *convert;
@@ -73,6 +81,7 @@ struct impl {
 
 	struct spa_io_buffers io_buffers;
 	struct spa_io_rate_match io_rate_match;
+	struct spa_io_position *io_position;
 
 	uint64_t info_all;
 	struct spa_node_info info;
@@ -82,9 +91,12 @@ struct impl {
 #define IDX_Format		3
 #define IDX_EnumPortConfig	4
 #define IDX_PortConfig		5
-	struct spa_param_info params[6];
-	uint32_t convert_params_flags[6];
-	uint32_t follower_params_flags[6];
+#define IDX_Latency		6
+#define IDX_ProcessLatency	7
+#define N_NODE_PARAMS		8
+	struct spa_param_info params[N_NODE_PARAMS];
+	uint32_t convert_params_flags[N_NODE_PARAMS];
+	uint32_t follower_params_flags[N_NODE_PARAMS];
 
 	struct spa_hook_list hooks;
 	struct spa_callbacks callbacks;
@@ -94,17 +106,45 @@ struct impl {
 	unsigned int started:1;
 	unsigned int driver:1;
 	unsigned int async:1;
+	unsigned int passthrough:1;
+	unsigned int follower_removing:1;
 };
 
 /** \endcond */
 
+static int follower_enum_params(struct impl *this,
+				 uint32_t id,
+				 uint32_t idx,
+				 struct spa_result_node_params *result,
+				 const struct spa_pod *filter,
+				 struct spa_pod_builder *builder)
+{
+	int res;
+	if (result->next < 0x100000) {
+		if ((res = spa_node_enum_params_sync(this->convert,
+				id, &result->next, filter, &result->param, builder)) == 1)
+			return res;
+		result->next = 0x100000;
+	}
+	if (result->next < 0x200000 && this->follower_params_flags[idx] & SPA_PARAM_INFO_READ) {
+		result->next &= 0xfffff;
+		if ((res = spa_node_enum_params_sync(this->follower,
+				id, &result->next, filter, &result->param, builder)) == 1) {
+			result->next |= 0x100000;
+			return res;
+		}
+		result->next = 0x200000;
+	}
+	return 0;
+}
+
 static int impl_node_enum_params(void *object, int seq,
 				 uint32_t id, uint32_t start, uint32_t num,
 				 const struct spa_pod *filter)
 {
 	struct impl *this = object;
-	struct spa_pod_builder b = { 0 };
-	uint8_t buffer[1024];
+	uint8_t buffer[4096];
+	struct spa_pod_dynamic_builder b;
 	struct spa_result_node_params result;
 	uint32_t count = 0;
 	int res;
@@ -117,9 +157,9 @@ static int impl_node_enum_params(void *object, int seq,
 next:
 	result.index = result.next;
 
-	spa_log_debug(this->log, NAME" %p: %d id:%u", this, seq, id);
+	spa_log_debug(this->log, "%p: %d id:%u", this, seq, id);
 
-	spa_pod_builder_init(&b, buffer, sizeof(buffer));
+	spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
 
 	switch (id) {
 	case SPA_PARAM_EnumPortConfig:
@@ -127,38 +167,38 @@ next:
 		res = spa_node_enum_params(this->convert, seq, id, start, num, filter);
 		return res;
 	case SPA_PARAM_PropInfo:
+		res = follower_enum_params(this,
+				id, IDX_PropInfo, &result, filter, &b.b);
+		break;
 	case SPA_PARAM_Props:
-	{
-		if (result.next < 0x10000) {
-			if ((res = spa_node_enum_params_sync(this->convert,
-					id, &result.next, filter, &result.param, &b)) == 1)
-				break;
-			result.next = 0x10000;
-		}
-		if (result.next >= 0x10000) {
-			result.next &= 0xffff;
-			if ((res = spa_node_enum_params_sync(this->follower,
-					id, &result.next, filter, &result.param, &b)) == 1) {
-				result.next |= 0x10000;
-				break;
-			}
-		}
-		return res;
-	}
+		res = follower_enum_params(this,
+				id, IDX_Props, &result, filter, &b.b);
+		break;
+	case SPA_PARAM_ProcessLatency:
+		res = follower_enum_params(this,
+				id, IDX_ProcessLatency, &result, filter, &b.b);
+		break;
 	case SPA_PARAM_EnumFormat:
 	case SPA_PARAM_Format:
-		if ((res = spa_node_port_enum_params_sync(this->follower,
+	case SPA_PARAM_Latency:
+		res = spa_node_port_enum_params_sync(this->follower,
 				this->direction, 0,
-				id, &result.next, filter, &result.param, &b)) == 1)
-			break;
-		return res;
+				id, &result.next, filter, &result.param, &b.b);
+		break;
 	default:
 		return -ENOENT;
 	}
 
-	spa_node_emit_result(&this->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
+	if (res == 1) {
+		spa_node_emit_result(&this->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
+		count++;
+	}
+	spa_pod_dynamic_builder_clean(&b);
 
-	if (++count != num)
+	if (res != 1)
+		return res;
+
+	if (count != num)
 		goto next;
 
 	return 0;
@@ -171,7 +211,7 @@ static int link_io(struct impl *this)
 	if (this->convert == NULL)
 		return 0;
 
-	spa_log_debug(this->log, NAME " %p: controls", this);
+	spa_log_debug(this->log, "%p: controls", this);
 
 	spa_zero(this->io_rate_match);
 	this->io_rate_match.rate = 1.0;
@@ -180,14 +220,14 @@ static int link_io(struct impl *this)
 			this->direction, 0,
 			SPA_IO_RateMatch,
 			&this->io_rate_match, sizeof(this->io_rate_match))) < 0) {
-		spa_log_debug(this->log, NAME " %p: set RateMatch on follower disabled %d %s", this,
+		spa_log_debug(this->log, "%p: set RateMatch on follower disabled %d %s", this,
 			res, spa_strerror(res));
 	}
 	else if ((res = spa_node_port_set_io(this->convert,
 			SPA_DIRECTION_REVERSE(this->direction), 0,
 			SPA_IO_RateMatch,
 			&this->io_rate_match, sizeof(this->io_rate_match))) < 0) {
-		spa_log_warn(this->log, NAME " %p: set RateMatch on convert failed %d %s", this,
+		spa_log_warn(this->log, "%p: set RateMatch on convert failed %d %s", this,
 			res, spa_strerror(res));
 	}
 
@@ -197,7 +237,7 @@ static int link_io(struct impl *this)
 			this->direction, 0,
 			SPA_IO_Buffers,
 			&this->io_buffers, sizeof(this->io_buffers))) < 0) {
-		spa_log_warn(this->log, NAME " %p: set Buffers on follower failed %d %s", this,
+		spa_log_warn(this->log, "%p: set Buffers on follower failed %d %s", this,
 			res, spa_strerror(res));
 		return res;
 	}
@@ -205,7 +245,7 @@ static int link_io(struct impl *this)
 			SPA_DIRECTION_REVERSE(this->direction), 0,
 			SPA_IO_Buffers,
 			&this->io_buffers, sizeof(this->io_buffers))) < 0) {
-		spa_log_warn(this->log, NAME " %p: set Buffers on convert failed %d %s", this,
+		spa_log_warn(this->log, "%p: set Buffers on convert failed %d %s", this,
 			res, spa_strerror(res));
 		return res;
 	}
@@ -215,12 +255,13 @@ static int link_io(struct impl *this)
 static void emit_node_info(struct impl *this, bool full)
 {
 	uint32_t i;
+	uint64_t old = full ? this->info.change_mask : 0;
 
 	if (full)
 		this->info.change_mask = this->info_all;
 	if (this->info.change_mask) {
 		if (this->info.change_mask & SPA_NODE_CHANGE_MASK_PARAMS) {
-			for (i = 0; i < SPA_N_ELEMENTS(this->params); i++) {
+			for (i = 0; i < this->info.n_params; i++) {
 				if (this->params[i].user > 0) {
 					this->params[i].flags ^= SPA_PARAM_INFO_SERIAL;
 					this->params[i].user = 0;
@@ -228,7 +269,7 @@ static void emit_node_info(struct impl *this, bool full)
 			}
 		}
 		spa_node_emit_info(&this->hooks, &this->info);
-		this->info.change_mask = 0;
+		this->info.change_mask = old;
 	}
 }
 
@@ -283,7 +324,10 @@ static int negotiate_buffers(struct impl *this)
 	struct spa_data *datas;
 	uint32_t follower_flags, conv_flags;
 
-	spa_log_debug(this->log, NAME" %p: %d", this, this->n_buffers);
+	spa_log_debug(this->log, "%p: %d", this, this->n_buffers);
+
+	if (this->target == this->follower)
+		return 0;
 
 	if (this->n_buffers > 0)
 		return 0;
@@ -331,13 +375,15 @@ static int negotiate_buffers(struct impl *this)
 			follower_alloc = false;
 	}
 
+	align = DEFAULT_ALIGN;
+
 	if ((res = spa_pod_parse_object(param,
 			SPA_TYPE_OBJECT_ParamBuffers, NULL,
 			SPA_PARAM_BUFFERS_buffers, SPA_POD_Int(&buffers),
 			SPA_PARAM_BUFFERS_blocks,  SPA_POD_Int(&blocks),
 			SPA_PARAM_BUFFERS_size,    SPA_POD_Int(&size),
 			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(&stride),
-			SPA_PARAM_BUFFERS_align,   SPA_POD_Int(&align))) < 0)
+			SPA_PARAM_BUFFERS_align,   SPA_POD_OPT_Int(&align))) < 0)
 		return res;
 
 	spa_log_debug(this->log, "%p: buffers:%d, blocks:%d, size:%d, stride:%d align:%d %d:%d",
@@ -380,11 +426,11 @@ static int configure_format(struct impl *this, uint32_t flags, const struct spa_
 {
 	int res;
 
-	spa_log_debug(this->log, NAME "%p: configure format:", this);
+	spa_log_debug(this->log, "%p: configure format:", this);
 	if (format && spa_log_level_enabled(this->log, SPA_LOG_LEVEL_DEBUG))
 		spa_debug_format(0, NULL, format);
 
-	if (this->convert) {
+	if (this->target != this->follower && this->convert) {
 		if ((res = spa_node_port_set_param(this->convert,
 					   SPA_DIRECTION_REVERSE(this->direction), 0,
 					   SPA_PARAM_Format, flags,
@@ -408,6 +454,104 @@ static int configure_format(struct impl *this, uint32_t flags, const struct spa_
 	return res;
 }
 
+static int configure_convert(struct impl *this, uint32_t mode)
+{
+	struct spa_pod_builder b = { 0 };
+	uint8_t buffer[1024];
+	struct spa_pod *param;
+
+	spa_pod_builder_init(&b, buffer, sizeof(buffer));
+
+	spa_log_debug(this->log, "%p: configure convert %p", this, this->target);
+
+	param = spa_pod_builder_add_object(&b,
+		SPA_TYPE_OBJECT_ParamPortConfig, SPA_PARAM_PortConfig,
+		SPA_PARAM_PORT_CONFIG_direction,	SPA_POD_Id(this->direction),
+		SPA_PARAM_PORT_CONFIG_mode,		SPA_POD_Id(mode));
+
+	return spa_node_set_param(this->convert, SPA_PARAM_PortConfig, 0, param);
+}
+
+extern const struct spa_handle_factory spa_audioconvert_factory;
+
+static const struct spa_node_events follower_node_events;
+
+static int reconfigure_mode(struct impl *this, bool passthrough,
+                enum spa_direction direction, struct spa_pod *format)
+{
+	int res = 0;
+	struct spa_hook l;
+
+	spa_log_debug(this->log, "%p: passthrough mode %d", this, passthrough);
+
+	if (this->passthrough != passthrough) {
+		if (passthrough) {
+			/* remove converter split/merge ports */
+			configure_convert(this, SPA_PARAM_PORT_CONFIG_MODE_none);
+		} else {
+			/* remove follower ports */
+			this->follower_removing = true;
+			spa_zero(l);
+			spa_node_add_listener(this->follower, &l, &follower_node_events, this);
+			spa_hook_remove(&l);
+			this->follower_removing = false;
+		}
+	}
+
+	/* set new target */
+	this->target = passthrough ? this->follower : this->convert;
+
+	if ((res = configure_format(this, 0, format)) < 0)
+		return res;
+
+	if (this->passthrough != passthrough) {
+		this->passthrough = passthrough;
+		if (passthrough) {
+			/* add follower ports */
+			spa_zero(l);
+			spa_node_add_listener(this->follower, &l, &follower_node_events, this);
+			spa_hook_remove(&l);
+		} else {
+			/* add converter ports */
+			configure_convert(this, SPA_PARAM_PORT_CONFIG_MODE_dsp);
+			link_io(this);
+		}
+	}
+
+	this->info.change_mask |= SPA_NODE_CHANGE_MASK_FLAGS | SPA_NODE_CHANGE_MASK_PARAMS;
+	this->info.flags &= ~SPA_NODE_FLAG_NEED_CONFIGURE;
+	this->params[IDX_Props].user++;
+
+	emit_node_info(this, false);
+
+	return 0;
+}
+
+static int format_audio_raw_parse_opt(const struct spa_pod *format, struct spa_audio_info_raw *info)
+{
+	struct spa_pod *position = NULL;
+	uint32_t media_type, media_subtype;
+	int res;
+	if ((res = spa_format_parse(format, &media_type, &media_subtype)) < 0)
+		return res;
+	if (media_type != SPA_MEDIA_TYPE_audio ||
+	    media_subtype != SPA_MEDIA_SUBTYPE_raw)
+		return -ENOTSUP;
+
+	spa_zero(*info);
+	res = spa_pod_parse_object(format,
+			SPA_TYPE_OBJECT_Format, NULL,
+			SPA_FORMAT_AUDIO_format, SPA_POD_OPT_Id(&info->format),
+			SPA_FORMAT_AUDIO_rate, SPA_POD_OPT_Int(&info->rate),
+			SPA_FORMAT_AUDIO_channels, SPA_POD_OPT_Int(&info->channels),
+			SPA_FORMAT_AUDIO_position, SPA_POD_OPT_Pod(&position));
+	if (position == NULL ||
+	    !spa_pod_copy_array(position, SPA_TYPE_Id, info->position, SPA_AUDIO_MAX_CHANNELS))
+		SPA_FLAG_SET(info->flags, SPA_AUDIO_FLAG_UNPOSITIONED);
+
+	return res;
+}
+
 static int impl_node_set_param(void *object, uint32_t id, uint32_t flags,
 			       const struct spa_pod *param)
 {
@@ -415,7 +559,7 @@ static int impl_node_set_param(void *object, uint32_t id, uint32_t flags,
 	struct impl *this = object;
 	struct spa_audio_info info = { 0 };
 
-	spa_log_debug(this->log, NAME" %p: set param %d", this, id);
+	spa_log_debug(this->log, "%p: set param %d", this, id);
 
 	switch (id) {
 	case SPA_PARAM_Format:
@@ -436,22 +580,65 @@ static int impl_node_set_param(void *object, uint32_t id, uint32_t flags,
 		break;
 
 	case SPA_PARAM_PortConfig:
-		if (this->started)
+	{
+		enum spa_direction dir;
+		enum spa_param_port_config_mode mode;
+		struct spa_pod *format = NULL;
+		int monitor = false;
+
+		if (this->started) {
+			spa_log_error(this->log, "was started");
 			return -EIO;
+		}
+
+		if (spa_pod_parse_object(param,
+				SPA_TYPE_OBJECT_ParamPortConfig, NULL,
+				SPA_PARAM_PORT_CONFIG_direction,	SPA_POD_Id(&dir),
+				SPA_PARAM_PORT_CONFIG_mode,		SPA_POD_Id(&mode),
+				SPA_PARAM_PORT_CONFIG_monitor,		SPA_POD_OPT_Bool(&monitor),
+				SPA_PARAM_PORT_CONFIG_format,		SPA_POD_OPT_Pod(&format)) < 0)
+			return -EINVAL;
+
+		if (format) {
+			struct spa_audio_info info;
+			if (format_audio_raw_parse_opt(format, &info.info.raw) >= 0)
+				this->default_format = info;
+		}
+
+		switch (mode) {
+		case SPA_PARAM_PORT_CONFIG_MODE_none:
+			return -ENOTSUP;
+		case SPA_PARAM_PORT_CONFIG_MODE_passthrough:
+			if ((res = reconfigure_mode(this, true, dir, format)) < 0)
+				return res;
+			break;
+		case SPA_PARAM_PORT_CONFIG_MODE_convert:
+		case SPA_PARAM_PORT_CONFIG_MODE_dsp:
+			if ((res = reconfigure_mode(this, false, dir, NULL)) < 0)
+				return res;
+			break;
+		default:
+			return -EINVAL;
+		}
+
 		if (this->target != this->follower) {
 			if ((res = spa_node_set_param(this->target, id, flags, param)) < 0)
 				return res;
 		}
 		break;
+	}
 
 	case SPA_PARAM_Props:
 		if (this->target != this->follower)
 			res = spa_node_set_param(this->target, id, flags, param);
-		res = spa_node_set_param(this->follower, id, flags, param);
+		res2 = spa_node_set_param(this->follower, id, flags, param);
 		if (res < 0 && res2 < 0)
 			return res;
 		res = 0;
 		break;
+	case SPA_PARAM_ProcessLatency:
+		res = spa_node_set_param(this->follower, id, flags, param);
+		break;
 	default:
 		res = -ENOTSUP;
 		break;
@@ -466,6 +653,14 @@ static int impl_node_set_io(void *object, uint32_t id, void *data, size_t size)
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
+	switch (id) {
+	case SPA_IO_Position:
+		this->io_position = data;
+		break;
+	default:
+		break;
+	}
+
 	if (this->target)
 		res = spa_node_set_io(this->target, id, data, size);
 
@@ -475,10 +670,44 @@ static int impl_node_set_io(void *object, uint32_t id, void *data, size_t size)
 	return res;
 }
 
+static struct spa_pod *merge_objects(struct impl *this, struct spa_pod_builder *b, uint32_t id,
+			struct spa_pod_object *o1, struct spa_pod_object *o2)
+{
+	const struct spa_pod_prop *p1, *p2;
+	struct spa_pod_frame f;
+	struct spa_pod_builder_state state;
+	int res = 0;
+
+	if (o2 == NULL || SPA_POD_TYPE(o1) != SPA_POD_TYPE(o2))
+		return (struct spa_pod*)o1;
+
+	spa_pod_builder_push_object(b, &f, o1->body.type, o1->body.id);
+	p2 = NULL;
+	SPA_POD_OBJECT_FOREACH(o1, p1) {
+		p2 = spa_pod_object_find_prop(o2, p2, p1->key);
+		if (p2 != NULL) {
+			spa_pod_builder_get_state(b, &state);
+			res = spa_pod_filter_prop(b, p1, p2);
+			if (res < 0)
+		                spa_pod_builder_reset(b, &state);
+		}
+		if (p2 == NULL || res < 0)
+			spa_pod_builder_raw_padded(b, p1, SPA_POD_PROP_SIZE(p1));
+	}
+	p1 = NULL;
+	SPA_POD_OBJECT_FOREACH(o2, p2) {
+		p1 = spa_pod_object_find_prop(o1, p1, p2->key);
+		if (p1 != NULL)
+			continue;
+		spa_pod_builder_raw_padded(b, p2, SPA_POD_PROP_SIZE(p2));
+	}
+	return spa_pod_builder_pop(b, &f);
+}
+
 static int negotiate_format(struct impl *this)
 {
 	uint32_t state;
-	struct spa_pod *format;
+	struct spa_pod *format, *def;
 	uint8_t buffer[4096];
 	struct spa_pod_builder b = { 0 };
 	int res;
@@ -486,9 +715,12 @@ static int negotiate_format(struct impl *this)
 	if (this->have_format)
 		return 0;
 
+	if (this->target == this->follower)
+		return 0;
+
 	spa_pod_builder_init(&b, buffer, sizeof(buffer));
 
-	spa_log_debug(this->log, NAME " %p: negiotiate", this);
+	spa_log_debug(this->log, "%p: negiotiate", this);
 
 	spa_node_send_command(this->follower,
 			&SPA_NODE_COMMAND_INIT(SPA_NODE_COMMAND_ParamBegin));
@@ -507,7 +739,6 @@ static int negotiate_format(struct impl *this)
 			goto done;
 		}
 	}
-
 	if (this->convert) {
 		state = 0;
 		if ((res = spa_node_port_enum_params_sync(this->convert,
@@ -526,6 +757,13 @@ static int negotiate_format(struct impl *this)
 		goto done;
 	}
 
+	def = spa_format_audio_raw_build(&b,
+			SPA_PARAM_Format, &this->default_format.info.raw);
+
+	format = merge_objects(this, &b, SPA_PARAM_Format,
+			(struct spa_pod_object*)format,
+			(struct spa_pod_object*)def);
+
 	spa_pod_fixate(format);
 
 	res = configure_format(this, 0, format);
@@ -545,7 +783,7 @@ static int impl_node_send_command(void *object, const struct spa_command *comman
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
-	spa_log_debug(this->log, NAME " %p: command %d", this, SPA_NODE_COMMAND_ID(command));
+	spa_log_debug(this->log, "%p: command %d", this, SPA_NODE_COMMAND_ID(command));
 
 	switch (SPA_NODE_COMMAND_ID(command)) {
 	case SPA_NODE_COMMAND_Start:
@@ -568,15 +806,17 @@ static int impl_node_send_command(void *object, const struct spa_command *comman
 	}
 
 	if ((res = spa_node_send_command(this->target, command)) < 0) {
-		spa_log_error(this->log, NAME " %p: can't send command: %s",
-				this, spa_strerror(res));
+		spa_log_error(this->log, "%p: can't send command %d: %s",
+				this, SPA_NODE_COMMAND_ID(command),
+				spa_strerror(res));
 		return res;
 	}
 
 	if (this->target != this->follower) {
 		if ((res = spa_node_send_command(this->follower, command)) < 0) {
-			spa_log_error(this->log, NAME " %p: can't send command: %s",
-					this, spa_strerror(res));
+			spa_log_error(this->log, "%p: can't send command %d: %s",
+					this, SPA_NODE_COMMAND_ID(command),
+					spa_strerror(res));
 			return res;
 		}
 	}
@@ -593,10 +833,6 @@ static void convert_node_info(void *data, const struct spa_node_info *info)
 	struct impl *this = data;
 	uint32_t i;
 
-	if (info->change_mask & SPA_NODE_CHANGE_MASK_FLAGS) {
-		this->info.change_mask |= SPA_NODE_CHANGE_MASK_FLAGS;
-		this->info.flags = info->flags;
-	}
 	if (info->change_mask & SPA_NODE_CHANGE_MASK_PARAMS) {
 		for (i = 0; i < info->n_params; i++) {
 			uint32_t idx;
@@ -641,16 +877,21 @@ static void convert_port_info(void *data,
 			port_id--;
 	}
 
-	spa_log_trace(this->log, NAME" %p: port info %d:%d", this,
+	spa_log_trace(this->log, "%p: port info %d:%d", this,
 			direction, port_id);
 
-	spa_node_emit_port_info(&this->hooks, direction, port_id, info);
+	if (this->target != this->follower)
+		spa_node_emit_port_info(&this->hooks, direction, port_id, info);
 }
 
 static void convert_result(void *data, int seq, int res, uint32_t type, const void *result)
 {
 	struct impl *this = data;
-	spa_log_trace(this->log, NAME" %p: result %d %d", this, seq, res);
+
+	if (this->target == this->follower)
+		return;
+
+	spa_log_trace(this->log, "%p: result %d %d", this, seq, res);
 	spa_node_emit_result(&this->hooks, seq, res, type, result);
 }
 
@@ -666,6 +907,9 @@ static void follower_info(void *data, const struct spa_node_info *info)
 	struct impl *this = data;
 	uint32_t i;
 
+	if (this->follower_removing)
+		return;
+
 	this->async = (info->flags & SPA_NODE_FLAG_ASYNC) != 0;
 
 	if (info->max_input_ports > 0)
@@ -673,10 +917,15 @@ static void follower_info(void *data, const struct spa_node_info *info)
         else
 		this->direction = SPA_DIRECTION_OUTPUT;
 
-	this->info.max_input_ports = this->direction == SPA_DIRECTION_INPUT ? MAX_PORTS : 0;
-	this->info.max_output_ports = this->direction == SPA_DIRECTION_OUTPUT ? MAX_PORTS : 0;
+	if (this->direction == SPA_DIRECTION_INPUT) {
+		this->info.flags |= SPA_NODE_FLAG_IN_PORT_CONFIG;
+		this->info.max_input_ports = MAX_PORTS;
+	} else {
+		this->info.flags |= SPA_NODE_FLAG_OUT_PORT_CONFIG;
+		this->info.max_output_ports = MAX_PORTS;
+	}
 
-	spa_log_debug(this->log, NAME" %p: follower info %s", this,
+	spa_log_debug(this->log, "%p: follower info %s", this,
 			this->direction == SPA_DIRECTION_INPUT ?
 				"Input" : "Output");
 
@@ -689,9 +938,15 @@ static void follower_info(void *data, const struct spa_node_info *info)
 			uint32_t idx;
 
 			switch (info->params[i].id) {
+			case SPA_PARAM_PropInfo:
+				idx = IDX_PropInfo;
+				break;
 			case SPA_PARAM_Props:
 				idx = IDX_Props;
 				break;
+			case SPA_PARAM_ProcessLatency:
+				idx = IDX_ProcessLatency;
+				break;
 			default:
 				continue;
 			}
@@ -712,21 +967,70 @@ static void follower_info(void *data, const struct spa_node_info *info)
 	emit_node_info(this, false);
 }
 
+static int recalc_latency(struct impl *this, enum spa_direction direction, uint32_t port_id)
+{
+	struct spa_pod_builder b = { 0 };
+	uint8_t buffer[1024];
+	struct spa_pod *param;
+	uint32_t index = 0;
+	struct spa_latency_info latency;
+	int res;
+
+	spa_log_debug(this->log, "%p: ", this);
+
+	if (this->target == this->follower)
+		return 0;
+
+	while (true) {
+		spa_pod_builder_init(&b, buffer, sizeof(buffer));
+		if ((res = spa_node_port_enum_params_sync(this->follower,
+						direction, port_id, SPA_PARAM_Latency,
+						&index, NULL, &param, &b)) != 1)
+			return res;
+		if ((res = spa_latency_parse(param, &latency)) < 0)
+			return res;
+		if (latency.direction == direction)
+			break;
+	}
+	if ((res = spa_node_port_set_param(this->target,
+					SPA_DIRECTION_REVERSE(direction), 0,
+					SPA_PARAM_Latency, 0, param)) < 0)
+		return res;
+
+	return 0;
+}
+
 static void follower_port_info(void *data,
 		enum spa_direction direction, uint32_t port_id,
 		const struct spa_port_info *info)
 {
 	struct impl *this = data;
 	uint32_t i;
+	int res;
+
+	if (this->follower_removing) {
+	      spa_node_emit_port_info(&this->hooks, direction, port_id, NULL);
+	      return;
+	}
+
+	spa_log_debug(this->log, "%p: follower port info %s %p %08"PRIx64, this,
+			this->direction == SPA_DIRECTION_INPUT ?
+				"Input" : "Output", info, info->change_mask);
 
 	if (info->change_mask & SPA_PORT_CHANGE_MASK_PARAMS) {
 		for (i = 0; i < info->n_params; i++) {
 			uint32_t idx;
 
 			switch (info->params[i].id) {
+			case SPA_PARAM_EnumFormat:
+				idx = IDX_EnumFormat;
+				break;
 			case SPA_PARAM_Format:
 				idx = IDX_Format;
 				break;
+			case SPA_PARAM_Latency:
+				idx = IDX_Latency;
+				break;
 			default:
 				continue;
 			}
@@ -734,35 +1038,57 @@ static void follower_port_info(void *data,
 			    this->follower_params_flags[idx] == info->params[i].flags)
 				continue;
 
-			this->info.change_mask |= SPA_NODE_CHANGE_MASK_PARAMS;
 			this->follower_params_flags[idx] = info->params[i].flags;
 			this->params[idx].flags =
 				(this->params[idx].flags & SPA_PARAM_INFO_SERIAL) |
 				(info->params[i].flags & SPA_PARAM_INFO_READWRITE);
 
+			if (idx == IDX_Latency) {
+				res = recalc_latency(this, direction, port_id);
+				spa_log_debug(this->log, "latency: %d", res);
+			}
+
+			this->info.change_mask |= SPA_NODE_CHANGE_MASK_PARAMS;
 			if (!this->add_listener)
 				this->params[idx].user++;
 		}
 	}
 	emit_node_info(this, false);
+
+	if (this->target == this->follower)
+	      spa_node_emit_port_info(&this->hooks, direction, port_id, info);
+}
+
+static void follower_result(void *data, int seq, int res, uint32_t type, const void *result)
+{
+	struct impl *this = data;
+
+	if (this->target != this->follower)
+		return;
+
+	spa_log_trace(this->log, "%p: result %d %d", this, seq, res);
+	spa_node_emit_result(&this->hooks, seq, res, type, result);
 }
 
 static const struct spa_node_events follower_node_events = {
 	SPA_VERSION_NODE_EVENTS,
 	.info = follower_info,
 	.port_info = follower_port_info,
+	.result = follower_result,
 };
 
 static int follower_ready(void *data, int status)
 {
 	struct impl *this = data;
 
-	spa_log_trace_fp(this->log, NAME " %p: ready %d", this, status);
+	spa_log_trace_fp(this->log, "%p: ready %d", this, status);
 
-	this->driver = true;
+	if (this->target != this->follower) {
+		this->driver = true;
 
-	if (this->direction == SPA_DIRECTION_OUTPUT)
-		status = spa_node_process(this->convert);
+		if (this->direction == SPA_DIRECTION_OUTPUT)
+			status = spa_node_process(this->convert);
+	}
 
 	return spa_node_call_ready(&this->callbacks, status);
 }
@@ -772,7 +1098,7 @@ static int follower_reuse_buffer(void *data, uint32_t port_id, uint32_t buffer_i
 	int res;
 	struct impl *this = data;
 
-	if (this->convert)
+	if (this->target != this->follower && this->convert)
 		res = spa_node_port_reuse_buffer(this->convert, port_id, buffer_id);
 	else
 		res = spa_node_call_reuse_buffer(&this->callbacks, port_id, buffer_id);
@@ -804,7 +1130,7 @@ static int impl_node_add_listener(void *object,
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
-	spa_log_trace(this->log, NAME" %p: add listener %p", this, listener);
+	spa_log_trace(this->log, "%p: add listener %p", this, listener);
 	spa_hook_list_isolate(&this->hooks, &save, listener, events, data);
 
 
@@ -894,7 +1220,7 @@ impl_node_port_enum_params(void *object, int seq,
 	if (direction != this->direction)
 		port_id++;
 
-	spa_log_debug(this->log, NAME" %p: %d %u", this, seq, id);
+	spa_log_debug(this->log, "%p: %d %u", this, seq, id);
 
 	return spa_node_port_enum_params(this->target, seq, direction, port_id, id,
 			start, num, filter);
@@ -920,6 +1246,13 @@ impl_node_port_set_param(void *object,
 			flags, param)) < 0)
 		return res;
 
+	if ((id == SPA_PARAM_Latency) &&
+	    direction == this->direction) {
+		if ((res = spa_node_port_set_param(this->follower, direction, 0, id,
+				flags, param)) < 0)
+			return res;
+	}
+
 	return res;
 }
 
@@ -958,7 +1291,7 @@ impl_node_port_use_buffers(void *object,
 	if (direction != this->direction)
 		port_id++;
 
-	spa_log_debug(this->log, NAME" %p: %d %d:%d", this,
+	spa_log_debug(this->log, "%p: %d %d:%d", this,
 			n_buffers, direction, port_id);
 
 	if ((res = spa_node_port_use_buffers(this->target,
@@ -986,6 +1319,12 @@ static int impl_node_process(void *object)
 	spa_log_trace_fp(this->log, "%p: process convert:%p driver:%d",
 			this, this->convert, this->driver);
 
+	if (this->target == this->follower) {
+		if (this->io_position)
+			this->io_rate_match.size = this->io_position->clock.duration;
+		return spa_node_process(this->follower);
+	}
+
 	if (this->direction == SPA_DIRECTION_INPUT) {
 		/* an input node (sink).
 		 * First we run the converter to process the input for the follower
@@ -1096,7 +1435,7 @@ static int impl_get_interface(struct spa_handle *handle, const char *type, void
 
 	this = (struct impl *) handle;
 
-	if (strcmp(type, SPA_TYPE_INTERFACE_Node) == 0)
+	if (spa_streq(type, SPA_TYPE_INTERFACE_Node))
 		*interface = &this->node;
 	else
 		return -ENOENT;
@@ -1124,25 +1463,6 @@ static int impl_clear(struct spa_handle *handle)
 	return 0;
 }
 
-static int configure_adapt(struct impl *this)
-{
-	struct spa_pod_builder b = { 0 };
-	uint8_t buffer[1024];
-	struct spa_pod *param;
-
-	spa_pod_builder_init(&b, buffer, sizeof(buffer));
-
-	spa_log_debug(this->log, "%p: configure convert %p", this, this->target);
-
-	param = spa_pod_builder_add_object(&b,
-		SPA_TYPE_OBJECT_ParamPortConfig, SPA_PARAM_PortConfig,
-		SPA_PARAM_PORT_CONFIG_direction,	SPA_POD_Id(this->direction),
-		SPA_PARAM_PORT_CONFIG_mode,		SPA_POD_Id(SPA_PARAM_PORT_CONFIG_MODE_dsp));
-
-	return spa_node_set_param(this->target, SPA_PARAM_PortConfig, 0, param);
-}
-
-extern const struct spa_handle_factory spa_audioconvert_factory;
 
 static size_t
 impl_get_size(const struct spa_handle_factory *factory,
@@ -1176,6 +1496,8 @@ impl_init(const struct spa_handle_factory *factory,
 	this = (struct impl *) handle;
 
 	this->log = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_Log);
+	spa_log_topic_init(this->log, log_topic);
+
 	this->cpu = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_CPU);
 
 	if (info == NULL ||
@@ -1196,7 +1518,7 @@ impl_init(const struct spa_handle_factory *factory,
 			SPA_VERSION_NODE,
 			&impl_node, this);
 
-	this->hnd_convert = SPA_MEMBER(this, sizeof(struct impl), struct spa_handle);
+	this->hnd_convert = SPA_PTROFF(this, sizeof(struct impl), struct spa_handle);
 	spa_handle_factory_init(&spa_audioconvert_factory,
 				this->hnd_convert,
 				info, support, n_support);
@@ -1209,8 +1531,6 @@ impl_init(const struct spa_handle_factory *factory,
 		SPA_NODE_CHANGE_MASK_PARAMS;
 	this->info = SPA_NODE_INFO_INIT();
 	this->info.flags = SPA_NODE_FLAG_RT |
-		SPA_NODE_FLAG_IN_PORT_CONFIG |
-		SPA_NODE_FLAG_OUT_PORT_CONFIG |
 		SPA_NODE_FLAG_NEED_CONFIGURE;
 	this->params[IDX_EnumFormat] = SPA_PARAM_INFO(SPA_PARAM_EnumFormat, SPA_PARAM_INFO_READ);
 	this->params[IDX_PropInfo] = SPA_PARAM_INFO(SPA_PARAM_PropInfo, SPA_PARAM_INFO_READ);
@@ -1218,8 +1538,10 @@ impl_init(const struct spa_handle_factory *factory,
 	this->params[IDX_Format] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
 	this->params[IDX_EnumPortConfig] = SPA_PARAM_INFO(SPA_PARAM_EnumPortConfig, SPA_PARAM_INFO_READ);
 	this->params[IDX_PortConfig] = SPA_PARAM_INFO(SPA_PARAM_PortConfig, SPA_PARAM_INFO_READWRITE);
+	this->params[IDX_Latency] = SPA_PARAM_INFO(SPA_PARAM_Latency, SPA_PARAM_INFO_READWRITE);
+	this->params[IDX_ProcessLatency] = SPA_PARAM_INFO(SPA_PARAM_ProcessLatency, SPA_PARAM_INFO_READWRITE);
 	this->info.params = this->params;
-	this->info.n_params = 6;
+	this->info.n_params = N_NODE_PARAMS;
 
 	spa_node_add_listener(this->follower,
 			&this->follower_listener, &follower_node_events, this);
@@ -1228,7 +1550,7 @@ impl_init(const struct spa_handle_factory *factory,
 	spa_node_add_listener(this->convert,
 			&this->convert_listener, &convert_node_events, this);
 
-	configure_adapt(this);
+	configure_convert(this, SPA_PARAM_PORT_CONFIG_MODE_dsp);
 
 	link_io(this);
 
