@@ -30,6 +30,7 @@
 #include <spa/support/log.h>
 #include <spa/utils/list.h>
 #include <spa/utils/names.h>
+#include <spa/utils/string.h>
 #include <spa/node/node.h>
 #include <spa/node/io.h>
 #include <spa/node/utils.h>
@@ -40,13 +41,13 @@
 
 #include "resample.h"
 
-#define NAME "resample"
+#undef SPA_LOG_TOPIC_DEFAULT
+#define SPA_LOG_TOPIC_DEFAULT log_topic
+static struct spa_log_topic *log_topic = &SPA_LOG_TOPIC(0, "spa.resample");
 
 #define DEFAULT_RATE		48000
 #define DEFAULT_CHANNELS	2
 
-#define MAX_SAMPLES	8192
-#define MAX_ALIGN	16
 #define MAX_BUFFERS	32
 
 struct impl;
@@ -54,12 +55,14 @@ struct impl;
 struct props {
 	double rate;
 	int quality;
+	bool disabled;
 };
 
 static void props_reset(struct props *props)
 {
 	props->rate = 1.0;
 	props->quality = RESAMPLE_DEFAULT_QUALITY;
+	props->disabled = false;
 }
 
 struct buffer {
@@ -101,6 +104,8 @@ struct impl {
 	struct spa_log *log;
 	struct spa_cpu *cpu;
 
+	uint32_t quantum_limit;
+
 	struct spa_io_position *io_position;
 	struct spa_io_rate_match *io_rate_match;
 
@@ -123,7 +128,7 @@ struct impl {
 
 	struct resample resample;
 
-	float empty[MAX_SAMPLES + MAX_ALIGN];
+	double rate_scale;
 };
 
 #define CHECK_PORT(this,d,id)		(id == 0)
@@ -146,7 +151,7 @@ static int setup_convert(struct impl *this,
 		dst_info = info;
 	}
 
-	spa_log_info(this->log, NAME " %p: %s/%d@%d->%s/%d@%d", this,
+	spa_log_info(this->log, "%p: %s/%d@%d->%s/%d@%d", this,
 			spa_debug_type_find_name(spa_type_audio_format, src_info->info.raw.format),
 			src_info->info.raw.channels,
 			src_info->info.raw.rate,
@@ -178,7 +183,147 @@ static int impl_node_enum_params(void *object, int seq,
 				 uint32_t id, uint32_t start, uint32_t num,
 				 const struct spa_pod *filter)
 {
-	return -ENOTSUP;
+	struct impl *this = object;
+	struct spa_pod *param;
+	struct spa_pod_builder b = { 0 };
+	uint8_t buffer[4096];
+	struct spa_result_node_params result;
+	uint32_t count = 0;
+
+	spa_return_val_if_fail(this != NULL, -EINVAL);
+	spa_return_val_if_fail(num != 0, -EINVAL);
+
+	result.id = id;
+	result.next = start;
+      next:
+	result.index = result.next++;
+
+	spa_pod_builder_init(&b, buffer, sizeof(buffer));
+
+	switch (id) {
+	case SPA_PARAM_PropInfo:
+	{
+		struct props *p = &this->props;
+
+		switch (result.index) {
+		case 0:
+			param = spa_pod_builder_add_object(&b,
+				SPA_TYPE_OBJECT_PropInfo, id,
+				SPA_PROP_INFO_id, SPA_POD_Id(SPA_PROP_rate),
+				SPA_PROP_INFO_description, SPA_POD_String("Rate scaler"),
+				SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Double(p->rate, 0.0, 10.0));
+			break;
+		case 1:
+			param = spa_pod_builder_add_object(&b,
+				SPA_TYPE_OBJECT_PropInfo, id,
+				SPA_PROP_INFO_id, SPA_POD_Id(SPA_PROP_quality),
+				SPA_PROP_INFO_name, SPA_POD_String("resample.quality"),
+				SPA_PROP_INFO_description, SPA_POD_String("Resample Quality"),
+				SPA_PROP_INFO_type, SPA_POD_CHOICE_RANGE_Int(p->quality, 0, 14),
+				SPA_PROP_INFO_params, SPA_POD_Bool(true));
+			break;
+		case 2:
+			param = spa_pod_builder_add_object(&b,
+				SPA_TYPE_OBJECT_PropInfo, id,
+				SPA_PROP_INFO_name, SPA_POD_String("resample.disable"),
+				SPA_PROP_INFO_description, SPA_POD_String("Disable Resampling"),
+				SPA_PROP_INFO_type, SPA_POD_CHOICE_Bool(p->disabled),
+				SPA_PROP_INFO_params, SPA_POD_Bool(true));
+			break;
+		default:
+			return 0;
+		}
+		break;
+	}
+	case SPA_PARAM_Props:
+	{
+		struct props *p = &this->props;
+		struct spa_pod_frame f[2];
+
+		switch (result.index) {
+		case 0:
+			spa_pod_builder_push_object(&b, &f[0], SPA_TYPE_OBJECT_Props, id);
+			spa_pod_builder_add(&b,
+				SPA_PROP_rate,			SPA_POD_Double(p->rate),
+				SPA_PROP_quality,		SPA_POD_Int(p->quality),
+				0);
+			spa_pod_builder_prop(&b, SPA_PROP_params, 0);
+			spa_pod_builder_push_struct(&b, &f[1]);
+			spa_pod_builder_string(&b, "resample.quality");
+			spa_pod_builder_int(&b, p->quality);
+			spa_pod_builder_string(&b, "resample.disable");
+			spa_pod_builder_bool(&b, p->disabled);
+			spa_pod_builder_pop(&b, &f[1]);
+			param = spa_pod_builder_pop(&b, &f[0]);
+			break;
+		default:
+			return 0;
+		}
+		break;
+	}
+	default:
+		return -ENOENT;
+	}
+
+	if (spa_pod_filter(&b, &result.param, param, filter) < 0)
+		goto next;
+
+	spa_node_emit_result(&this->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
+
+	if (++count != num)
+		goto next;
+
+	return 0;
+}
+
+static int resample_set_param(struct impl *this, const char *k, const char *s)
+{
+	if (spa_streq(k, "resample.quality"))
+		this->props.quality = atoi(s);
+	else if (spa_streq(k, "resample.disable"))
+		this->props.disabled = spa_atob(s);
+	return 0;
+}
+
+static int parse_prop_params(struct impl *this, struct spa_pod *params)
+{
+	struct spa_pod_parser prs;
+	struct spa_pod_frame f;
+
+	spa_pod_parser_pod(&prs, params);
+	if (spa_pod_parser_push_struct(&prs, &f) < 0)
+		return 0;
+
+	while (true) {
+		const char *name;
+		struct spa_pod *pod;
+		char value[512];
+
+		if (spa_pod_parser_get_string(&prs, &name) < 0)
+			break;
+
+		if (spa_pod_parser_get_pod(&prs, &pod) < 0)
+			break;
+
+		if (spa_pod_is_string(pod)) {
+			spa_pod_copy_string(pod, sizeof(value), value);
+		} else if (spa_pod_is_float(pod)) {
+			snprintf(value, sizeof(value), "%f",
+					SPA_POD_VALUE(struct spa_pod_float, pod));
+		} else if (spa_pod_is_int(pod)) {
+			snprintf(value, sizeof(value), "%d",
+					SPA_POD_VALUE(struct spa_pod_int, pod));
+		} else if (spa_pod_is_bool(pod)) {
+			snprintf(value, sizeof(value), "%s",
+					SPA_POD_VALUE(struct spa_pod_bool, pod) ?
+					"true" : "false");
+		} else
+			continue;
+
+		spa_log_info(this->log, "key:'%s' val:'%s'", name, value);
+		resample_set_param(this, name, value);
+	}
+	return 0;
 }
 
 static int apply_props(struct impl *this, const struct spa_pod *param)
@@ -186,22 +331,28 @@ static int apply_props(struct impl *this, const struct spa_pod *param)
 	struct spa_pod_prop *prop;
 	struct spa_pod_object *obj = (struct spa_pod_object *) param;
 	struct props *p = &this->props;
+	int changed = 0;
 
 	SPA_POD_OBJECT_FOREACH(obj, prop) {
 		switch (prop->key) {
 		case SPA_PROP_rate:
 			if (spa_pod_get_double(&prop->value, &p->rate) == 0) {
 				resample_update_rate(&this->resample, p->rate);
+				changed++;
 			}
 			break;
 		case SPA_PROP_quality:
-			spa_pod_get_int(&prop->value, &p->quality);
+			if (spa_pod_get_int(&prop->value, &p->quality) == 0)
+				changed++;
+			break;
+		case SPA_PROP_params:
+			changed += parse_prop_params(this, &prop->value);
 			break;
 		default:
 			break;
 		}
 	}
-	return 0;
+	return changed;
 }
 
 static int impl_node_set_param(void *object, uint32_t id, uint32_t flags,
@@ -229,7 +380,7 @@ static int impl_node_set_io(void *object, uint32_t id, void *data, size_t size)
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
-	spa_log_debug(this->log, NAME " %p: io %d %p/%zd", this, id, data, size);
+	spa_log_debug(this->log, "%p: io %d %p/%zd", this, id, data, size);
 
 	switch (id) {
 	case SPA_IO_Position:
@@ -251,20 +402,33 @@ static void update_rate_match(struct impl *this, bool passthrough, uint32_t out_
 			match_size = out_size;
 		} else {
 			if (SPA_FLAG_IS_SET(this->io_rate_match->flags, SPA_IO_RATE_MATCH_FLAG_ACTIVE))
-				resample_update_rate(&this->resample, this->io_rate_match->rate);
+				resample_update_rate(&this->resample, this->rate_scale * this->io_rate_match->rate);
 			else
-				resample_update_rate(&this->resample, 1.0);
+				resample_update_rate(&this->resample, this->rate_scale);
 
 			this->io_rate_match->delay = resample_delay(&this->resample);
 			match_size = resample_in_len(&this->resample, out_size);
 		}
 		match_size -= SPA_MIN(match_size, in_queued);
 		this->io_rate_match->size = match_size;
-		spa_log_trace_fp(this->log, NAME " %p: next match %u", this, match_size);
+		spa_log_trace_fp(this->log, "%p: next match %u", this, match_size);
 	} else {
-		resample_update_rate(&this->resample, this->props.rate);
+		resample_update_rate(&this->resample, this->rate_scale * this->props.rate);
 	}
 }
+static inline bool is_passthrough(struct impl *this)
+{
+	return this->resample.i_rate == this->resample.o_rate && this->rate_scale == 1.0 &&
+		(this->io_rate_match == NULL || this->props.disabled ||
+		 !SPA_FLAG_IS_SET(this->io_rate_match->flags, SPA_IO_RATE_MATCH_FLAG_ACTIVE));
+}
+
+static void recalc_rate_match(struct impl *this)
+{
+	bool passthrough = is_passthrough(this);
+	uint32_t out_size = this->io_position ? this->io_position->clock.duration : 1024;
+	update_rate_match(this, passthrough, out_size, 0);
+}
 
 static void reset_node(struct impl *this)
 {
@@ -272,7 +436,8 @@ static void reset_node(struct impl *this)
 	outport = GET_OUT_PORT(this, 0);
 	inport = GET_IN_PORT(this, 0);
 
-	resample_reset(&this->resample);
+	if (this->resample.reset)
+		resample_reset(&this->resample);
 	outport->offset = 0;
 	inport->offset = 0;
 }
@@ -286,15 +451,9 @@ static int impl_node_send_command(void *object, const struct spa_command *comman
 
 	switch (SPA_NODE_COMMAND_ID(command)) {
 	case SPA_NODE_COMMAND_Start:
-	{
-		bool passthrough = this->resample.i_rate == this->resample.o_rate &&
-			(this->io_rate_match == NULL ||
-			 !SPA_FLAG_IS_SET(this->io_rate_match->flags, SPA_IO_RATE_MATCH_FLAG_ACTIVE));
-		uint32_t out_size = this->io_position ? this->io_position->clock.duration : 1024;
+		recalc_rate_match(this);
 		this->started = true;
-		update_rate_match(this, passthrough, out_size, 0);
 		break;
-	}
 	case SPA_NODE_COMMAND_Suspend:
 	case SPA_NODE_COMMAND_Flush:
 		reset_node(this);
@@ -310,23 +469,24 @@ static int impl_node_send_command(void *object, const struct spa_command *comman
 
 static void emit_node_info(struct impl *this, bool full)
 {
+	uint64_t old = full ? this->info.change_mask : 0;
 	if (full)
 		this->info.change_mask = this->info_all;
-
 	if (this->info.change_mask) {
 		spa_node_emit_info(&this->hooks, &this->info);
-		this->info.change_mask = 0;
+		this->info.change_mask = old;
 	}
 }
 
 static void emit_port_info(struct impl *this, struct port *port, bool full)
 {
+	uint64_t old = full ? port->info.change_mask : 0;
 	if (full)
 		port->info.change_mask = port->info_all;
 	if (port->info.change_mask) {
 		spa_node_emit_port_info(&this->hooks,
 				port->direction, port->id, &port->info);
-		port->info.change_mask = 0;
+		port->info.change_mask = old;
 	}
 }
 
@@ -381,20 +541,24 @@ static int port_enum_formats(void *object,
 	struct impl *this = object;
 	struct port *other;
 	struct spa_pod_frame f;
+	uint32_t rate, min = 1, max = INT32_MAX;
 
 	other = GET_PORT(this, SPA_DIRECTION_REVERSE(direction), 0);
 
 	switch (index) {
 	case 0:
 		if (other->have_format) {
+			rate = other->format.info.raw.rate;
+			if (this->props.disabled)
+				min = max = rate;
+
 			spa_pod_builder_push_object(builder, &f,
 				SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat);
 			spa_pod_builder_add(builder,
 				SPA_FORMAT_mediaType,      SPA_POD_Id(SPA_MEDIA_TYPE_audio),
 				SPA_FORMAT_mediaSubtype,   SPA_POD_Id(SPA_MEDIA_SUBTYPE_raw),
 				SPA_FORMAT_AUDIO_format,   SPA_POD_Id(SPA_AUDIO_FORMAT_F32P),
-				SPA_FORMAT_AUDIO_rate,     SPA_POD_CHOICE_RANGE_Int(
-								other->format.info.raw.rate, 1, INT32_MAX),
+				SPA_FORMAT_AUDIO_rate,     SPA_POD_CHOICE_RANGE_Int(rate, min, max),
 				SPA_FORMAT_AUDIO_channels, SPA_POD_Int(other->format.info.raw.channels),
 				0);
 			spa_pod_builder_prop(builder, SPA_FORMAT_AUDIO_position, 0);
@@ -402,12 +566,17 @@ static int port_enum_formats(void *object,
 					other->format.info.raw.channels, other->format.info.raw.position);
 			*param = spa_pod_builder_pop(builder, &f);
 		} else {
+			rate = this->io_position ?
+				this->io_position->clock.rate.denom : DEFAULT_RATE;
+			if (this->props.disabled)
+				min = max = rate;
+
 			*param = spa_pod_builder_add_object(builder,
 				SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat,
 				SPA_FORMAT_mediaType,      SPA_POD_Id(SPA_MEDIA_TYPE_audio),
 				SPA_FORMAT_mediaSubtype,   SPA_POD_Id(SPA_MEDIA_SUBTYPE_raw),
 				SPA_FORMAT_AUDIO_format,   SPA_POD_Id(SPA_AUDIO_FORMAT_F32P),
-				SPA_FORMAT_AUDIO_rate,     SPA_POD_CHOICE_RANGE_Int(DEFAULT_RATE, 1, INT32_MAX),
+				SPA_FORMAT_AUDIO_rate,     SPA_POD_CHOICE_RANGE_Int(rate, min, max),
 				SPA_FORMAT_AUDIO_channels, SPA_POD_CHOICE_RANGE_Int(DEFAULT_CHANNELS, 1, INT32_MAX));
 		}
 		break;
@@ -467,20 +636,26 @@ impl_node_port_enum_params(void *object, int seq,
 	case SPA_PARAM_Buffers:
 	{
 		uint32_t buffers, size;
+		uint32_t rate;
 
-		if (!port->have_format)
+		if (!port->have_format || !other->have_format)
 			return -EIO;
 		if (result.index > 0)
 			return 0;
 
+		if (direction == SPA_DIRECTION_OUTPUT) {
+			rate = (this->resample.o_rate + this->resample.i_rate - 1) / this->resample.i_rate;
+		} else {
+			rate = (this->resample.i_rate + this->resample.o_rate - 1) / this->resample.o_rate;
+		}
 		if (other->n_buffers > 0) {
 			buffers = other->n_buffers;
-			size = other->size / other->stride * 2;
-		}
-		else {
+			size = (other->size / other->stride) * rate;
+		} else {
 			buffers = 1;
-			size = MAX_SAMPLES*2 * other->stride;
+			size = this->quantum_limit * rate;
 		}
+		size = SPA_MAX(size, this->quantum_limit) * 2;
 
 		param = spa_pod_builder_add_object(&b,
 			SPA_TYPE_OBJECT_ParamBuffers, id,
@@ -490,8 +665,7 @@ impl_node_port_enum_params(void *object, int seq,
 							size * port->stride,
 							16 * port->stride,
 							INT32_MAX),
-			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(port->stride),
-			SPA_PARAM_BUFFERS_align,   SPA_POD_Int(16));
+			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(port->stride));
 		break;
 	}
 	case SPA_PARAM_Meta:
@@ -536,7 +710,7 @@ impl_node_port_enum_params(void *object, int seq,
 static int clear_buffers(struct impl *this, struct port *port)
 {
 	if (port->n_buffers > 0) {
-		spa_log_debug(this->log, NAME " %p: clear buffers %p", this, port);
+		spa_log_debug(this->log, "%p: clear buffers %p", this, port);
 		port->n_buffers = 0;
 		spa_list_init(&port->queue);
 	}
@@ -587,7 +761,7 @@ static int port_set_format(void *object,
 		port->format = info;
 		port->have_format = true;
 
-		spa_log_debug(this->log, NAME " %p: set format on port %d %d", this, port_id, res);
+		spa_log_debug(this->log, "%p: set format on port %d %d", this, port_id, res);
 	}
 
 	port->info.change_mask |= SPA_PORT_CHANGE_MASK_PARAMS;
@@ -640,7 +814,7 @@ impl_node_port_use_buffers(void *object,
 
 	spa_return_val_if_fail(port->have_format, -EIO);
 
-	spa_log_debug(this->log, NAME " %p: use buffers %d on port %d:%d", this,
+	spa_log_debug(this->log, "%p: use buffers %d on port %d:%d", this,
 			n_buffers, direction, port_id);
 
 	clear_buffers(this, port);
@@ -660,13 +834,13 @@ impl_node_port_use_buffers(void *object,
 				size = d[j].maxsize;
 			else
 				if (size != d[j].maxsize) {
-					spa_log_error(this->log, NAME " %p: invalid size %d on buffer %p", this,
+					spa_log_error(this->log, "%p: invalid size %d on buffer %p", this,
 						      size, buffers[i]);
 					return -EINVAL;
 				}
 
 			if (d[j].data == NULL) {
-				spa_log_error(this->log, NAME " %p: invalid memory on buffer %p", this,
+				spa_log_error(this->log, "%p: invalid memory on buffer %p", this,
 					      buffers[i]);
 				return -EINVAL;
 			}
@@ -697,7 +871,7 @@ impl_node_port_set_io(void *object,
 
 	spa_return_val_if_fail(CHECK_PORT(this, direction, port_id), -EINVAL);
 
-	spa_log_trace_fp(this->log, NAME " %p: %d:%d io %d", this, direction, port_id, id);
+	spa_log_trace_fp(this->log, "%p: %d:%d io %d", this, direction, port_id, id);
 
 	port = GET_PORT(this, direction, port_id);
 
@@ -722,7 +896,7 @@ static void recycle_buffer(struct impl *this, uint32_t id)
 	if (SPA_FLAG_IS_SET(b->flags, BUFFER_FLAG_OUT)) {
 		spa_list_append(&port->queue, &b->link);
 		SPA_FLAG_CLEAR(b->flags, BUFFER_FLAG_OUT);
-		spa_log_trace_fp(this->log, NAME " %p: recycle buffer %d", this, id);
+		spa_log_trace_fp(this->log, "%p: recycle buffer %d", this, id);
 	}
 }
 
@@ -786,7 +960,7 @@ static int impl_node_process(void *object)
 	spa_return_val_if_fail(outio != NULL, -EIO);
 	spa_return_val_if_fail(inio != NULL, -EIO);
 
-	spa_log_trace_fp(this->log, NAME " %p: status %p %d %d -> %p %d %d", this,
+	spa_log_trace_fp(this->log, "%p: status %p %d %d -> %p %d %d", this,
 			inio, inio->status, inio->buffer_id,
 			outio, outio->status, outio->buffer_id);
 
@@ -798,9 +972,10 @@ static int impl_node_process(void *object)
 		outio->buffer_id = SPA_ID_INVALID;
 	}
 	if (SPA_UNLIKELY(inio->status != SPA_STATUS_HAVE_DATA)) {
-		if (inio->status != SPA_STATUS_DRAINED || this->drained)
+		if (inio->status != SPA_STATUS_DRAINED || this->drained) {
+			recalc_rate_match(this);
 			return outio->status = inio->status;
-
+		}
 		inio->buffer_id = 0;
 		inport->buffers[0].outbuf->datas[0].chunk->size = 0;
 	}
@@ -819,8 +994,26 @@ static int impl_node_process(void *object)
 	size = sb->datas[0].chunk->size;
 	maxsize = db->datas[0].maxsize;
 
-	if (SPA_LIKELY(this->io_position))
+	if (SPA_LIKELY(this->io_position)) {
+		double r =  this->rate_scale;
+
 		max = this->io_position->clock.duration;
+		if (this->mode == MODE_SPLIT) {
+			if (this->io_position->clock.rate.denom != this->resample.o_rate)
+				r = (double) this->io_position->clock.rate.denom / this->resample.o_rate;
+			else
+				r = 1.0;
+		} else {
+			if (this->io_position->clock.rate.denom != this->resample.i_rate)
+				r = (double) this->resample.i_rate / this->io_position->clock.rate.denom;
+			else
+				r = 1.0;
+		}
+		if (this->rate_scale != r) {
+			spa_log_info(this->log, "scale %f->%f", this->rate_scale, r);
+			this->rate_scale = r;
+		}
+	}
 	else
 		max = maxsize / sizeof(float);
 
@@ -841,18 +1034,24 @@ static int impl_node_process(void *object)
 	src_datas = alloca(sizeof(void*) * this->resample.channels);
 	dst_datas = alloca(sizeof(void*) * this->resample.channels);
 
+	if (inport->offset > size)
+		inport->offset = size;
+	if (outport->offset > maxsize)
+		outport->offset = maxsize;
+
 	if (size == 0) {
-		size = MAX_SAMPLES * sizeof(float);
+		size = sb->datas[0].maxsize;
+		memset(sb->datas[0].data, 0, size);
 		for (i = 0; i < sb->n_datas; i++)
-			src_datas[i] = SPA_PTR_ALIGN(this->empty, MAX_ALIGN, void);
+			src_datas[i] = sb->datas[0].data;
 		inport->offset = 0;
 		flush_in = draining = true;
 	} else {
 		for (i = 0; i < sb->n_datas; i++)
-			src_datas[i] = SPA_MEMBER(sb->datas[i].data, inport->offset, void);
+			src_datas[i] = SPA_PTROFF(sb->datas[i].data, inport->offset, void);
 	}
 	for (i = 0; i < db->n_datas; i++)
-		dst_datas[i] = SPA_MEMBER(db->datas[i].data, outport->offset, void);
+		dst_datas[i] = SPA_PTROFF(db->datas[i].data, outport->offset, void);
 
 	in_len = (size - inport->offset) / sizeof(float);
 	out_len = (maxsize - outport->offset) / sizeof(float);
@@ -862,9 +1061,7 @@ static int impl_node_process(void *object)
 	pin_len = in_len;
 	pout_len = out_len;
 #endif
-	passthrough = this->resample.i_rate == this->resample.o_rate &&
-		(this->io_rate_match == NULL ||
-		 !SPA_FLAG_IS_SET(this->io_rate_match->flags, SPA_IO_RATE_MATCH_FLAG_ACTIVE));
+	passthrough = is_passthrough(this);
 
 	if (passthrough) {
 		uint32_t len = SPA_MIN(in_len, out_len);
@@ -876,7 +1073,7 @@ static int impl_node_process(void *object)
 	}
 
 #ifndef FASTPATH
-	spa_log_trace_fp(this->log, NAME " %p: in %d/%d %zd %d out %d/%d %zd %d max:%d",
+	spa_log_trace_fp(this->log, "%p: in %d/%d %zd %d out %d/%d %zd %d max:%d",
 			this, pin_len, in_len, size / sizeof(float), inport->offset,
 			pout_len, out_len, maxsize / sizeof(float), outport->offset,
 			max);
@@ -890,7 +1087,7 @@ static int impl_node_process(void *object)
 	inport->offset += in_len * sizeof(float);
 	if (inport->offset >= size || flush_in) {
 		inio->status = SPA_STATUS_NEED_DATA;
-		spa_log_trace_fp(this->log, NAME" %p: return input buffer of %zd samples",
+		spa_log_trace_fp(this->log, "%p: return input buffer of %zd samples",
 				this, size / sizeof(float));
 		inport->offset = 0;
 		size = 0;
@@ -901,7 +1098,7 @@ static int impl_node_process(void *object)
 	if (outport->offset > 0 && (outport->offset >= maxsize || flush_out)) {
 		outio->status = SPA_STATUS_HAVE_DATA;
 		outio->buffer_id = dbuf->id;
-		spa_log_trace_fp(this->log, NAME" %p: have output buffer of %zd samples",
+		spa_log_trace_fp(this->log, "%p: have output buffer of %zd samples",
 				this, outport->offset / sizeof(float));
 		dequeue_buffer(this, dbuf);
 		outport->offset = 0;
@@ -912,7 +1109,7 @@ static int impl_node_process(void *object)
 		outio->status = SPA_STATUS_HAVE_DATA;
 		outio->buffer_id = SPA_ID_INVALID;
 		SPA_FLAG_SET(res, SPA_STATUS_HAVE_DATA);
-		spa_log_trace_fp(this->log, NAME " %p: no output buffer", this);
+		spa_log_trace_fp(this->log, "%p: no output buffer", this);
 	}
 
 	update_rate_match(this, passthrough, max - outport->offset / sizeof(float),
@@ -947,7 +1144,7 @@ static int impl_get_interface(struct spa_handle *handle, const char *type, void
 
 	this = (struct impl *) handle;
 
-	if (strcmp(type, SPA_TYPE_INTERFACE_Node) == 0)
+	if (spa_streq(type, SPA_TYPE_INTERFACE_Node))
 		*interface = &this->node;
 	else
 		return -ENOENT;
@@ -984,7 +1181,7 @@ impl_init(const struct spa_handle_factory *factory,
 {
 	struct impl *this;
 	struct port *port;
-	const char *str;
+	uint32_t i;
 
 	spa_return_val_if_fail(factory != NULL, -EINVAL);
 	spa_return_val_if_fail(handle != NULL, -EINVAL);
@@ -995,27 +1192,33 @@ impl_init(const struct spa_handle_factory *factory,
 	this = (struct impl *) handle;
 
 	this->log = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_Log);
-	this->cpu = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_CPU);
+	spa_log_topic_init(this->log, log_topic);
 
+	this->cpu = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_CPU);
 	if (this->cpu)
 		this->resample.cpu_flags = spa_cpu_get_flags(this->cpu);
 
 	props_reset(&this->props);
 
-	if (info != NULL) {
-		if ((str = spa_dict_lookup(info, "resample.quality")) != NULL)
-			this->props.quality = atoi(str);
-		if ((str = spa_dict_lookup(info, "resample.peaks")) != NULL)
-			this->peaks = strcmp(str, "true") == 0 || atoi(str) == 1;
-		if ((str = spa_dict_lookup(info, "factory.mode")) != NULL) {
-			if (strcmp(str, "split") == 0)
+	for (i = 0; info && i < info->n_items; i++) {
+		const char *k = info->items[i].key;
+		const char *s = info->items[i].value;
+		if (spa_streq(k, "clock.quantum-limit"))
+			spa_atou32(s, &this->quantum_limit, 0);
+		else if (spa_streq(k, "resample.peaks"))
+			this->peaks = spa_atob(s);
+		else if (spa_streq(k, "factory.mode")) {
+			if (spa_streq(s, "split"))
 				this->mode = MODE_SPLIT;
-			else if (strcmp(str, "merge") == 0)
+			else if (spa_streq(s, "merge"))
 				this->mode = MODE_MERGE;
 			else
 				this->mode = MODE_CONVERT;
-		}
+		} else
+			resample_set_param(this, k, s);
+
 	}
+
 	spa_log_debug(this->log, "mode:%d", this->mode);
 
 	this->node.iface = SPA_INTERFACE_INIT(
@@ -1025,6 +1228,8 @@ impl_init(const struct spa_handle_factory *factory,
 
 	spa_hook_list_init(&this->hooks);
 
+	this->rate_scale = 1.0;
+
 	this->info = SPA_NODE_INFO_INIT();
 	this->info_all = SPA_NODE_CHANGE_MASK_FLAGS;
 	this->info.max_input_ports = 1;
