@@ -34,6 +34,7 @@
 #include <spa/utils/type.h>
 #include <spa/utils/keys.h>
 #include <spa/utils/names.h>
+#include <spa/utils/string.h>
 #include <spa/node/node.h>
 #include <spa/support/loop.h>
 #include <spa/support/plugin.h>
@@ -52,7 +53,9 @@
 #include "defs.h"
 #include "a2dp-codecs.h"
 
-#define NAME  "bluez5-device"
+static struct spa_log_topic log_topic = SPA_LOG_TOPIC(0, "spa.bluez5.device");
+#undef SPA_LOG_TOPIC_DEFAULT
+#define SPA_LOG_TOPIC_DEFAULT &log_topic
 
 #define MAX_DEVICES	64
 
@@ -91,6 +94,7 @@ struct node {
 	unsigned int active:1;
 	unsigned int mute:1;
 	unsigned int save:1;
+	unsigned int a2dp_duplex:1;
 	uint32_t n_channels;
 	int64_t latency_offset;
 	uint32_t channels[SPA_AUDIO_MAX_CHANNELS];
@@ -105,6 +109,7 @@ struct dynamic_node
 	struct spa_hook transport_listener;
 	uint32_t id;
 	const char *factory_name;
+	bool a2dp_duplex;
 };
 
 struct impl {
@@ -132,12 +137,14 @@ struct impl {
 
 	uint32_t profile;
 	unsigned int switching_codec:1;
+	unsigned int save_profile:1;
 	uint32_t prev_bt_connected_profiles;
 
 	const struct a2dp_codec **supported_codecs;
 	size_t supported_codec_count;
 
 	struct dynamic_node dyn_a2dp_source;
+	struct dynamic_node dyn_a2dp_sink;
 	struct dynamic_node dyn_sco_source;
 	struct dynamic_node dyn_sco_sink;
 
@@ -154,28 +161,40 @@ static void init_node(struct impl *this, struct node *node, uint32_t id)
 
 	spa_zero(*node);
 	node->id = id;
-	for (i = 0; i < SPA_AUDIO_MAX_CHANNELS; i++)
-		node->volumes[i] = 1.0;
+	for (i = 0; i < SPA_AUDIO_MAX_CHANNELS; i++) {
+		node->volumes[i] = 1.0f;
+		node->soft_volumes[i] = 1.0f;
+	}
 }
 
-static const struct a2dp_codec *get_a2dp_codec(enum spa_bluetooth_audio_codec id)
+static void get_a2dp_codecs(struct impl *this, enum spa_bluetooth_audio_codec id, const struct a2dp_codec **codecs, size_t size)
 {
-	const struct a2dp_codec **c;
+	const struct a2dp_codec * const *c;
 
-	for (c = a2dp_codecs; *c; ++c)
-		if ((*c)->id == id)
-			return *c;
+	spa_assert(size > 0);
+	spa_assert(this->supported_codecs);
 
-	return NULL;
+	for (c = this->supported_codecs; *c && size > 1; ++c) {
+		if ((*c)->id == id || id == 0) {
+			*codecs++ = *c;
+			--size;
+		}
+	}
+
+	*codecs = NULL;
 }
 
-static const struct a2dp_codec *get_supported_a2dp_codec(struct impl *this, enum spa_bluetooth_audio_codec id)
+static const struct a2dp_codec *get_supported_a2dp_codec(struct impl *this, enum spa_bluetooth_audio_codec id, size_t *idx)
 {
 	const struct a2dp_codec *a2dp_codec = NULL;
 	size_t i;
-	for (i = 0; i < this->supported_codec_count; ++i)
-		if (this->supported_codecs[i]->id == id)
+	for (i = 0; i < this->supported_codec_count; ++i) {
+		if (this->supported_codecs[i]->id == id) {
 			a2dp_codec = this->supported_codecs[i];
+			if (idx)
+				*idx = i;
+		}
+	}
 	return a2dp_codec;
 }
 
@@ -224,10 +243,13 @@ static const char *get_hfp_codec_name(unsigned int codec)
 	return "unknown";
 }
 
-static const char *get_codec_name(struct spa_bt_transport *t)
+static const char *get_codec_name(struct spa_bt_transport *t, bool a2dp_duplex)
 {
-	if (t->a2dp_codec != NULL)
+	if (t->a2dp_codec != NULL) {
+		if (a2dp_duplex && t->a2dp_codec->duplex_codec)
+			return t->a2dp_codec->duplex_codec->name;
 		return t->a2dp_codec->name;
+	}
 	return get_hfp_codec_name(t->codec);
 }
 
@@ -237,7 +259,7 @@ static void transport_destroy(void *userdata)
 	node->transport = NULL;
 }
 
-static void emit_volume(struct impl *this, struct node *node)
+static void emit_node_props(struct impl *this, struct node *node, bool full)
 {
 	struct spa_event *event;
 	uint8_t buffer[4096];
@@ -253,16 +275,50 @@ static void emit_volume(struct impl *this, struct node *node)
 	spa_pod_builder_add_object(&b,
 			SPA_TYPE_OBJECT_Props, SPA_EVENT_DEVICE_Props,
 			SPA_PROP_channelVolumes, SPA_POD_Array(sizeof(float),
+				SPA_TYPE_Float, node->n_channels, node->volumes),
+			SPA_PROP_softVolumes, SPA_POD_Array(sizeof(float),
 				SPA_TYPE_Float, node->n_channels, node->soft_volumes),
 			SPA_PROP_channelMap, SPA_POD_Array(sizeof(uint32_t),
 				SPA_TYPE_Id, node->n_channels, node->channels));
+	if (full) {
+		spa_pod_builder_add(&b,
+			SPA_PROP_mute, SPA_POD_Bool(node->mute),
+			SPA_PROP_softMute, SPA_POD_Bool(node->mute),
+			SPA_PROP_latencyOffsetNsec, SPA_POD_Long(node->latency_offset),
+			0);
+	}
 	event = spa_pod_builder_pop(&b, &f[0]);
 
 	spa_device_emit_event(&this->hooks, event);
 }
 
+static void emit_volume(struct impl *this, struct node *node)
+{
+	emit_node_props(this, node, false);
+}
+
 static void emit_info(struct impl *this, bool full);
 
+static float get_soft_volume_boost(struct node *node)
+{
+	/*
+	 * For A2DP duplex, the duplex microphone channel sometimes does not appear
+	 * to have hardware gain, and input volume is very low.
+	 *
+	 * Work around this by boosting the software volume level, i.e. adjust
+	 * the scale on the user-visible volume control to something more sensible.
+	 * If this causes clipping, the user can just reduce the mic volume to
+	 * bring SW gain below 1.
+	 */
+	if (node->a2dp_duplex && node->transport &&
+			node->id == DEVICE_ID_SOURCE &&
+			!node->transport->volumes[SPA_BT_VOLUME_ID_RX].active)
+		return 10.0f;	/* 20 dB boost */
+
+	/* In all other cases, no boost */
+	return 1.0f;
+}
+
 static float node_get_hw_volume(struct node *node)
 {
 	uint32_t i;
@@ -310,14 +366,11 @@ static void volume_changed(void *userdata)
 
 	node_update_soft_volumes(node, t_volume->volume);
 
+	emit_volume(impl, node);
+
 	impl->info.change_mask |= SPA_DEVICE_CHANGE_MASK_PARAMS;
 	impl->params[IDX_Route].flags ^= SPA_PARAM_INFO_SERIAL;
 	emit_info(impl, false);
-
-	/* It sometimes flips volume to over 100% in pavucontrol silder
-	 * if volume is emited before route info emitting while node
-	 * volumes are not identical to route volumes. Not sure why. */
-	emit_volume(impl, node);
 }
 
 static const struct spa_bt_transport_events transport_events = {
@@ -326,12 +379,39 @@ static const struct spa_bt_transport_events transport_events = {
 	.volume_changed = volume_changed,
 };
 
+static void get_channels(struct spa_bt_transport *t, bool a2dp_duplex, uint32_t *n_channels, uint32_t *channels)
+{
+	const struct a2dp_codec *codec;
+	struct spa_audio_info info = { 0 };
+
+	if (!a2dp_duplex || !t->a2dp_codec || !t->a2dp_codec->duplex_codec) {
+		*n_channels = t->n_channels;
+		memcpy(channels, t->channels, t->n_channels * sizeof(uint32_t));
+		return;
+	}
+
+	codec = t->a2dp_codec->duplex_codec;
+
+	if (!codec->validate_config ||
+			codec->validate_config(codec, 0,
+					t->configuration, t->configuration_len,
+					&info) < 0) {
+		*n_channels = 1;
+		channels[0] = SPA_AUDIO_CHANNEL_MONO;
+		return;
+	}
+
+	*n_channels = info.info.raw.channels;
+	memcpy(channels, info.info.raw.position,
+			info.info.raw.channels * sizeof(uint32_t));
+}
+
 static void emit_node(struct impl *this, struct spa_bt_transport *t,
-		uint32_t id, const char *factory_name)
+		uint32_t id, const char *factory_name, bool a2dp_duplex)
 {
 	struct spa_bt_device *device = this->bt_dev;
 	struct spa_device_object_info info;
-	struct spa_dict_item items[6];
+	struct spa_dict_item items[8];
 	uint32_t n_items = 0;
 	char transport[32], str_id[32];
 	bool is_dyn_node = SPA_FLAG_IS_SET(id, DYNAMIC_NODE_ID_FLAG);
@@ -339,7 +419,7 @@ static void emit_node(struct impl *this, struct spa_bt_transport *t,
 	snprintf(transport, sizeof(transport), "pointer:%p", t);
 	items[0] = SPA_DICT_ITEM_INIT(SPA_KEY_API_BLUEZ5_TRANSPORT, transport);
 	items[1] = SPA_DICT_ITEM_INIT(SPA_KEY_API_BLUEZ5_PROFILE, spa_bt_profile_name(t->profile));
-	items[2] = SPA_DICT_ITEM_INIT(SPA_KEY_API_BLUEZ5_CODEC, get_codec_name(t));
+	items[2] = SPA_DICT_ITEM_INIT(SPA_KEY_API_BLUEZ5_CODEC, get_codec_name(t, a2dp_duplex));
 	items[3] = SPA_DICT_ITEM_INIT(SPA_KEY_API_BLUEZ5_ADDRESS, device->address);
 	items[4] = SPA_DICT_ITEM_INIT("device.routes", "1");
 	n_items = 5;
@@ -348,6 +428,14 @@ static void emit_node(struct impl *this, struct spa_bt_transport *t,
 		items[5] = SPA_DICT_ITEM_INIT("card.profile.device", str_id);
 		n_items++;
 	}
+	if (spa_streq(spa_bt_profile_name(t->profile), "headset-head-unit")) {
+		items[n_items] = SPA_DICT_ITEM_INIT("device.intended-roles", "Communication");
+		n_items++;
+	}
+	if (a2dp_duplex) {
+		items[n_items] = SPA_DICT_ITEM_INIT("api.bluez5.a2dp-duplex", "true");
+		n_items++;
+	}
 
 	info = SPA_DEVICE_OBJECT_INFO_INIT();
 	info.type = SPA_TYPE_INTERFACE_Node;
@@ -359,26 +447,36 @@ static void emit_node(struct impl *this, struct spa_bt_transport *t,
 	spa_device_emit_object_info(&this->hooks, id, &info);
 
 	if (!is_dyn_node) {
-		if (this->nodes[id].n_channels > 0) {
-			size_t i;
-
-			/*
-			* Spread mono volume to all channels, if we had switched HFP -> A2DP.
-			* XXX: we should also use different route for hfp and a2dp
-			*/
-			for (i = this->nodes[id].n_channels; i < t->n_channels; ++i)
-				this->nodes[id].volumes[i] = this->nodes[id].volumes[i % this->nodes[id].n_channels];
-		}
+		uint32_t prev_channels = this->nodes[id].n_channels;
+		float boost;
 
 		this->nodes[id].impl = this;
 		this->nodes[id].active = true;
-		this->nodes[id].n_channels = t->n_channels;
-		memcpy(this->nodes[id].channels, t->channels,
-				t->n_channels * sizeof(uint32_t));
+		this->nodes[id].a2dp_duplex = a2dp_duplex;
+		get_channels(t, a2dp_duplex, &this->nodes[id].n_channels, this->nodes[id].channels);
 		if (this->nodes[id].transport)
 			spa_hook_remove(&this->nodes[id].transport_listener);
 		this->nodes[id].transport = t;
 		spa_bt_transport_add_listener(t, &this->nodes[id].transport_listener, &transport_events, &this->nodes[id]);
+
+		if (prev_channels > 0) {
+			size_t i;
+			/*
+			 * Spread mono volume to all channels, if we had switched HFP -> A2DP.
+			 * XXX: we should also use different route for hfp and a2dp
+			 */
+			for (i = prev_channels; i < this->nodes[id].n_channels; ++i)
+				this->nodes[id].volumes[i] = this->nodes[id].volumes[i % prev_channels];
+		}
+
+		boost = get_soft_volume_boost(&this->nodes[id]);
+		if (boost != 1.0f) {
+			size_t i;
+			for (i = 0; i < this->nodes[id].n_channels; ++i)
+				this->nodes[id].soft_volumes[i] = this->nodes[id].volumes[i] * boost;
+		}
+
+		emit_node_props(this, &this->nodes[id], true);
 	}
 }
 
@@ -386,17 +484,15 @@ static struct spa_bt_transport *find_transport(struct impl *this, int profile, e
 {
 	struct spa_bt_device *device = this->bt_dev;
 	struct spa_bt_transport *t;
-	const struct a2dp_codec *a2dp_codec;
-	unsigned int hfp_codec;
-
-	a2dp_codec = get_a2dp_codec(codec);
-	hfp_codec = get_hfp_codec(codec);
 
 	spa_list_for_each(t, &device->transport_list, device_link) {
+		bool codec_ok = codec == 0 ||
+			(t->a2dp_codec != NULL && t->a2dp_codec->id == codec) ||
+			get_hfp_codec_id(t->codec) == codec;
+
 		if ((t->profile & device->connected_profiles) &&
 				(t->profile & profile) == t->profile &&
-				(a2dp_codec == NULL || t->a2dp_codec == a2dp_codec) &&
-				(hfp_codec == 0 || t->codec == hfp_codec))
+				codec_ok)
 			return t;
 	}
 
@@ -423,11 +519,13 @@ static void dynamic_node_transport_state_changed(void *data,
 	if (state >= SPA_BT_TRANSPORT_STATE_PENDING && old < SPA_BT_TRANSPORT_STATE_PENDING) {
 		if (!SPA_FLAG_IS_SET(this->id, DYNAMIC_NODE_ID_FLAG)) {
 			SPA_FLAG_SET(this->id, DYNAMIC_NODE_ID_FLAG);
-			emit_node(impl, t, this->id, this->factory_name);
+			spa_bt_transport_keepalive(t, true);
+			emit_node(impl, t, this->id, this->factory_name, this->a2dp_duplex);
 		}
 	} else if (state < SPA_BT_TRANSPORT_STATE_PENDING && old >= SPA_BT_TRANSPORT_STATE_PENDING) {
 		if (SPA_FLAG_IS_SET(this->id, DYNAMIC_NODE_ID_FLAG)) {
 			SPA_FLAG_CLEAR(this->id, DYNAMIC_NODE_ID_FLAG);
+			spa_bt_transport_keepalive(t, false);
 			spa_device_emit_object_info(&impl->hooks, this->id, NULL);
 		}
 	}
@@ -488,15 +586,22 @@ static const struct spa_bt_transport_events dynamic_node_transport_events = {
 };
 
 static void emit_dynamic_node(struct dynamic_node *this, struct impl *impl,
-	struct spa_bt_transport *t, uint32_t id, const char *factory_name)
+	struct spa_bt_transport *t, uint32_t id, const char *factory_name, bool a2dp_duplex)
 {
-	if (this->transport != NULL)
-		return;
+	spa_log_debug(impl->log, "dynamic node, transport: %p->%p id: %08x->%08x",
+		this->transport, t, this->id, id);
+
+	if (this->transport) {
+		/* Session manager don't really handles transport ptr changing. */
+		spa_assert(this->transport == t);
+		spa_hook_remove(&this->transport_listener);
+	}
 
 	this->impl = impl;
 	this->transport = t;
 	this->id = id;
 	this->factory_name = factory_name;
+	this->a2dp_duplex = a2dp_duplex;
 
 	spa_bt_transport_add_listener(this->transport,
 		&this->transport_listener, &dynamic_node_transport_events, this);
@@ -539,9 +644,9 @@ static int emit_nodes(struct impl *this)
 				else
 					this->props.codec = get_hfp_codec_id(t->codec);
 				emit_dynamic_node(&this->dyn_sco_source, this, t,
-						0, SPA_NAME_API_BLUEZ5_SCO_SOURCE);
+						0, SPA_NAME_API_BLUEZ5_SCO_SOURCE, false);
 				emit_dynamic_node(&this->dyn_sco_sink, this, t,
-						1, SPA_NAME_API_BLUEZ5_SCO_SINK);
+						1, SPA_NAME_API_BLUEZ5_SCO_SINK, false);
 			}
 		}
 		if (this->bt_dev->connected_profiles & SPA_BT_PROFILE_A2DP_SOURCE) {
@@ -549,7 +654,12 @@ static int emit_nodes(struct impl *this)
 			if (t) {
 				this->props.codec = t->a2dp_codec->id;
 				emit_dynamic_node(&this->dyn_a2dp_source, this, t,
-						2, SPA_NAME_API_BLUEZ5_A2DP_SOURCE);
+						2, SPA_NAME_API_BLUEZ5_A2DP_SOURCE, false);
+
+				if (t->a2dp_codec->duplex_codec) {
+					emit_dynamic_node(&this->dyn_a2dp_sink, this, t,
+						3, SPA_NAME_API_BLUEZ5_A2DP_SINK, true);
+				}
 			}
 		}
 		break;
@@ -559,7 +669,12 @@ static int emit_nodes(struct impl *this)
 			if (t) {
 				this->props.codec = t->a2dp_codec->id;
 				emit_dynamic_node(&this->dyn_a2dp_source, this, t,
-					DEVICE_ID_SOURCE, SPA_NAME_API_BLUEZ5_A2DP_SOURCE);
+					DEVICE_ID_SOURCE, SPA_NAME_API_BLUEZ5_A2DP_SOURCE, false);
+
+				if (t->a2dp_codec->duplex_codec) {
+					emit_node(this, t,
+						DEVICE_ID_SINK, SPA_NAME_API_BLUEZ5_A2DP_SINK, true);
+				}
 			}
 		}
 
@@ -567,9 +682,17 @@ static int emit_nodes(struct impl *this)
 			t = find_transport(this, SPA_BT_PROFILE_A2DP_SINK, this->props.codec);
 			if (t) {
 				this->props.codec = t->a2dp_codec->id;
-				emit_node(this, t, DEVICE_ID_SINK, SPA_NAME_API_BLUEZ5_A2DP_SINK);
+				emit_node(this, t, DEVICE_ID_SINK, SPA_NAME_API_BLUEZ5_A2DP_SINK, false);
+
+				if (t->a2dp_codec->duplex_codec) {
+					emit_node(this, t,
+						DEVICE_ID_SOURCE, SPA_NAME_API_BLUEZ5_A2DP_SOURCE, true);
+				}
 			}
 		}
+
+		if (get_supported_a2dp_codec(this, this->props.codec, NULL) == NULL)
+			this->props.codec = 0;
 		break;
 	case DEVICE_PROFILE_HSP_HFP:
 		if (this->bt_dev->connected_profiles & SPA_BT_PROFILE_HEADSET_HEAD_UNIT) {
@@ -581,10 +704,13 @@ static int emit_nodes(struct impl *this)
 					this->props.codec = 0;
 				else
 					this->props.codec = get_hfp_codec_id(t->codec);
-				emit_node(this, t, DEVICE_ID_SOURCE, SPA_NAME_API_BLUEZ5_SCO_SOURCE);
-				emit_node(this, t, DEVICE_ID_SINK, SPA_NAME_API_BLUEZ5_SCO_SINK);
+				emit_node(this, t, DEVICE_ID_SOURCE, SPA_NAME_API_BLUEZ5_SCO_SOURCE, false);
+				emit_node(this, t, DEVICE_ID_SINK, SPA_NAME_API_BLUEZ5_SCO_SINK, false);
 			}
 		}
+
+		if (spa_bt_device_supports_hfp_codec(this->bt_dev, get_hfp_codec(this->props.codec)) != 1)
+			this->props.codec = 0;
 		break;
 	default:
 		return -EINVAL;
@@ -600,19 +726,21 @@ static const struct spa_dict_item info_items[] = {
 
 static void emit_info(struct impl *this, bool full)
 {
+	uint64_t old = full ? this->info.change_mask : 0;
 	if (full)
 		this->info.change_mask = this->info_all;
 	if (this->info.change_mask) {
 		this->info.props = &SPA_DICT_INIT_ARRAY(info_items);
 
 		spa_device_emit_info(&this->hooks, &this->info);
-		this->info.change_mask = 0;
+		this->info.change_mask = old;
 	}
 }
 
 static void emit_remove_nodes(struct impl *this)
 {
 	remove_dynamic_node (&this->dyn_a2dp_source);
+	remove_dynamic_node (&this->dyn_a2dp_sink);
 	remove_dynamic_node (&this->dyn_sco_source);
 	remove_dynamic_node (&this->dyn_sco_sink);
 
@@ -629,8 +757,20 @@ static void emit_remove_nodes(struct impl *this)
 	}
 }
 
-static int set_profile(struct impl *this, uint32_t profile, enum spa_bluetooth_audio_codec codec)
+static bool validate_profile(struct impl *this, uint32_t profile,
+		enum spa_bluetooth_audio_codec codec);
+
+static int set_profile(struct impl *this, uint32_t profile, enum spa_bluetooth_audio_codec codec, bool save)
 {
+	if (!validate_profile(this, profile, codec)) {
+		spa_log_warn(this->log, "trying to set invalid profile %d, codec %d, %08x %08x",
+			    profile, codec,
+			    this->bt_dev->profiles, this->bt_dev->connected_profiles);
+		return -EINVAL;
+	}
+
+	this->save_profile = save;
+
 	if (this->profile == profile &&
 	    (this->profile != DEVICE_PROFILE_A2DP || codec == this->props.codec) &&
 	    (this->profile != DEVICE_PROFILE_HSP_HFP || codec == this->props.codec))
@@ -645,29 +785,22 @@ static int set_profile(struct impl *this, uint32_t profile, enum spa_bluetooth_a
 	this->props.codec = codec;
 
 	/*
-	 * A2DP: ensure there's a transport with the selected codec (NULL means any).
+	 * A2DP: ensure there's a transport with the selected codec (0 means any).
 	 * Don't try to switch codecs when the device is in the A2DP source role, since
 	 * devices do not appear to like that.
 	 */
 	if (profile == DEVICE_PROFILE_A2DP && !(this->bt_dev->connected_profiles & SPA_BT_PROFILE_A2DP_SOURCE)) {
 		int ret;
-		const struct a2dp_codec *codec_list[2], **codecs, *a2dp_codec;
+		const struct a2dp_codec *codecs[64];
 
-		a2dp_codec = get_a2dp_codec(codec);
-		if (a2dp_codec == NULL) {
-			codecs = a2dp_codecs;
-		} else {
-			codec_list[0] = a2dp_codec;
-			codec_list[1] = NULL;
-			codecs = codec_list;
-		}
+		get_a2dp_codecs(this, codec, codecs, SPA_N_ELEMENTS(codecs));
 
 		this->switching_codec = true;
 
 		ret = spa_bt_device_ensure_a2dp_codec(this->bt_dev, codecs);
 		if (ret < 0) {
 			if (ret != -ENOTSUP)
-				spa_log_error(this->log, NAME": failed to switch codec (%d), setting basic profile", ret);
+				spa_log_error(this->log, "failed to switch codec (%d), setting basic profile", ret);
 		} else {
 			return 0;
 		}
@@ -679,7 +812,7 @@ static int set_profile(struct impl *this, uint32_t profile, enum spa_bluetooth_a
 		ret = spa_bt_device_ensure_hfp_codec(this->bt_dev, get_hfp_codec(codec));
 		if (ret < 0) {
 			if (ret != -ENOTSUP)
-				spa_log_error(this->log, NAME": failed to switch codec (%d), setting basic profile", ret);
+				spa_log_error(this->log, "failed to switch codec (%d), setting basic profile", ret);
 		} else {
 			return 0;
 		}
@@ -704,13 +837,13 @@ static void codec_switched(void *userdata, int status)
 {
 	struct impl *this = userdata;
 
-	spa_log_debug(this->log, NAME": codec switched (status %d)", status);
+	spa_log_debug(this->log, "codec switched (status %d)", status);
 
 	this->switching_codec = false;
 
 	if (status < 0) {
 		/* Failed to switch: return to a fallback profile */
-		spa_log_error(this->log, NAME": failed to switch codec (%d), setting fallback profile", status);
+		spa_log_error(this->log, "failed to switch codec (%d), setting fallback profile", status);
 		if (this->profile == DEVICE_PROFILE_A2DP && this->props.codec != 0) {
 			this->props.codec = 0;
 		} else if (this->profile == DEVICE_PROFILE_HSP_HFP && this->props.codec != 0) {
@@ -743,7 +876,7 @@ static void profiles_changed(void *userdata, uint32_t prev_profiles, uint32_t pr
 	connected_change = (this->bt_dev->connected_profiles ^ prev_connected_profiles);
 
 	/* Profiles changed. We have to re-emit device information. */
-	spa_log_info(this->log, NAME": profiles changed to  %08x %08x (prev %08x %08x, change %08x)"
+	spa_log_info(this->log, "profiles changed to  %08x %08x (prev %08x %08x, change %08x)"
 		     " switching_codec:%d",
 		     this->bt_dev->profiles, this->bt_dev->connected_profiles,
 		     prev_profiles, prev_connected_profiles, connected_change,
@@ -766,22 +899,22 @@ static void profiles_changed(void *userdata, uint32_t prev_profiles, uint32_t pr
 	case DEVICE_PROFILE_AG:
 		nodes_changed = (connected_change & (SPA_BT_PROFILE_HEADSET_AUDIO_GATEWAY |
 						     SPA_BT_PROFILE_A2DP_SOURCE));
-		spa_log_debug(this->log, NAME": profiles changed: AG nodes changed: %d",
+		spa_log_debug(this->log, "profiles changed: AG nodes changed: %d",
 			      nodes_changed);
 		break;
 	case DEVICE_PROFILE_A2DP:
-		if (get_supported_a2dp_codec(this, this->props.codec) == NULL)
+		if (get_supported_a2dp_codec(this, this->props.codec, NULL) == NULL)
 			this->props.codec = 0;
 		nodes_changed = (connected_change & (SPA_BT_PROFILE_A2DP_SINK |
 						     SPA_BT_PROFILE_A2DP_SOURCE));
-		spa_log_debug(this->log, NAME": profiles changed: A2DP nodes changed: %d",
+		spa_log_debug(this->log, "profiles changed: A2DP nodes changed: %d",
 			      nodes_changed);
 		break;
 	case DEVICE_PROFILE_HSP_HFP:
 		if (spa_bt_device_supports_hfp_codec(this->bt_dev, get_hfp_codec(this->props.codec)) != 1)
 			this->props.codec = 0;
 		nodes_changed = (connected_change & SPA_BT_PROFILE_HEADSET_HEAD_UNIT);
-		spa_log_debug(this->log, NAME": profiles changed: HSP/HFP nodes changed: %d",
+		spa_log_debug(this->log, "profiles changed: HSP/HFP nodes changed: %d",
 			      nodes_changed);
 		break;
 	}
@@ -854,16 +987,21 @@ static int impl_sync(void *object, int seq)
 	return 0;
 }
 
-static uint32_t profile_direction_mask(struct impl *this, uint32_t index)
+static uint32_t profile_direction_mask(struct impl *this, uint32_t index, enum spa_bluetooth_audio_codec codec)
 {
 	struct spa_bt_device *device = this->bt_dev;
 	uint32_t mask;
 	bool have_output = false, have_input = false;
+	const struct a2dp_codec *a2dp_codec;
 
 	switch (index) {
 	case DEVICE_PROFILE_A2DP:
 		if (device->connected_profiles & SPA_BT_PROFILE_A2DP_SINK)
 			have_output = true;
+
+		a2dp_codec = get_supported_a2dp_codec(this, codec, NULL);
+		if (a2dp_codec && a2dp_codec->duplex_codec)
+			have_input = true;
 		break;
 	case DEVICE_PROFILE_HSP_HFP:
 		if (device->connected_profiles & SPA_BT_PROFILE_HEADSET_HEAD_UNIT)
@@ -932,7 +1070,8 @@ static uint32_t get_index_from_profile(struct impl *this, uint32_t profile, enum
 	return SPA_ID_INVALID;
 }
 
-static bool find_hsp_hfp_profile(struct impl *this) {
+static bool set_initial_hsp_hfp_profile(struct impl *this)
+{
 	struct spa_bt_transport *t;
 	int i;
 
@@ -945,6 +1084,9 @@ static bool find_hsp_hfp_profile(struct impl *this) {
 			this->profile = (i & SPA_BT_PROFILE_HEADSET_AUDIO_GATEWAY) ?
 				DEVICE_PROFILE_AG : DEVICE_PROFILE_HSP_HFP;
 			this->props.codec = get_hfp_codec_id(t->codec);
+
+			spa_log_debug(this->log, "initial profile HSP/HFP profile:%d codec:%d",
+					this->profile, this->props.codec);
 			return true;
 		}
 	}
@@ -964,11 +1106,12 @@ static void set_initial_profile(struct impl *this)
 	/* Prefer A2DP, then HFP, then null, but select AG if the device
 	   appears not to have A2DP_SINK or any HEAD_UNIT profile */
 
-	// If default profile is set to HSP/HFP, first try those and exit if found
-	const char *str;
+	/* If default profile is set to HSP/HFP, first try those and exit if found. */
 	if (this->bt_dev->settings != NULL) {
-		str = spa_dict_lookup(this->bt_dev->settings, "device.profile");
-		if (str != NULL && strcmp(str, "headset-head-unit") == 0 && find_hsp_hfp_profile(this))
+		const char *str = spa_dict_lookup(this->bt_dev->settings, "bluez5.profile");
+		if (spa_streq(str, "off"))
+			goto off;
+		if (spa_streq(str, "headset-head-unit") && set_initial_hsp_hfp_profile(this))
 			return;
 	}
 
@@ -981,18 +1124,25 @@ static void set_initial_profile(struct impl *this)
 			this->profile = (i == SPA_BT_PROFILE_A2DP_SOURCE) ?
 				DEVICE_PROFILE_AG : DEVICE_PROFILE_A2DP;
 			this->props.codec = t->a2dp_codec->id;
+			spa_log_debug(this->log, "initial profile A2DP profile:%d codec:%d",
+					this->profile, this->props.codec);
 			return;
 		}
 	}
 
-	if (find_hsp_hfp_profile(this)) return;
+	if (set_initial_hsp_hfp_profile(this))
+		return;
+
+off:
+	spa_log_debug(this->log, "initial profile off");
 
 	this->profile = DEVICE_PROFILE_OFF;
 	this->props.codec = 0;
 }
 
 static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *b,
-		uint32_t id, uint32_t index, uint32_t profile_index, enum spa_bluetooth_audio_codec codec)
+		uint32_t id, uint32_t index, uint32_t profile_index, enum spa_bluetooth_audio_codec codec,
+		bool current)
 {
 	struct spa_bt_device *device = this->bt_dev;
 	struct spa_pod_frame f[2];
@@ -1001,11 +1151,13 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 	char *desc_and_codec = NULL;
 	uint32_t n_source = 0, n_sink = 0;
 	uint32_t capture[1] = { DEVICE_ID_SOURCE }, playback[1] = { DEVICE_ID_SINK };
+	int priority;
 
 	switch (profile_index) {
 	case DEVICE_PROFILE_OFF:
 		name = "off";
 		desc = _("Off");
+		priority = 0;
 		break;
 	case DEVICE_PROFILE_AG:
 	{
@@ -1017,6 +1169,7 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 			name = "audio-gateway";
 			desc = _("Audio Gateway (A2DP Source & HSP/HFP AG)");
 		}
+		priority = 256;
 		break;
 	}
 	case DEVICE_PROFILE_A2DP:
@@ -1030,26 +1183,31 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 		name = spa_bt_profile_name(profile);
 		n_sink++;
 		if (codec) {
-			const struct a2dp_codec *a2dp_codec = get_supported_a2dp_codec(this, codec);
+			size_t idx;
+			const struct a2dp_codec *a2dp_codec = get_supported_a2dp_codec(this, codec, &idx);
 			if (a2dp_codec == NULL) {
 				errno = EINVAL;
 				return NULL;
 			}
 			name_and_codec = spa_aprintf("%s-%s", name, a2dp_codec->name);
 			name = name_and_codec;
-			if (profile == SPA_BT_PROFILE_A2DP_SINK) {
-				desc = _("High Fidelity Playback (A2DP Sink, codec %s)");
+			if (profile == SPA_BT_PROFILE_A2DP_SINK && !a2dp_codec->duplex_codec) {
+				desc_and_codec = spa_aprintf(_("High Fidelity Playback (A2DP Sink, codec %s)"),
+							     a2dp_codec->description);
 			} else {
-				desc = _("High Fidelity Duplex (A2DP Source/Sink, codec %s)");
+				desc_and_codec = spa_aprintf(_("High Fidelity Duplex (A2DP Source/Sink, codec %s)"),
+							     a2dp_codec->description);
+
 			}
-			desc_and_codec = spa_aprintf(desc, a2dp_codec->description);
 			desc = desc_and_codec;
+			priority = 16 + this->supported_codec_count - idx;  /* order as in codec list */
 		} else {
 			if (profile == SPA_BT_PROFILE_A2DP_SINK) {
 				desc = _("High Fidelity Playback (A2DP Sink)");
 			} else {
 				desc = _("High Fidelity Duplex (A2DP Source/Sink)");
 			}
+			priority = 16;
 		}
 		break;
 	}
@@ -1078,8 +1236,10 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 			desc_and_codec = spa_aprintf(_("Headset Head Unit (HSP/HFP, codec %s)"),
 						get_hfp_codec_description(hfp_codec));
 			desc = desc_and_codec;
+			priority = 1 + hfp_codec;  /* prefer msbc over cvsd */
 		} else {
 			desc = _("Headset Head Unit (HSP/HFP)");
+			priority = 1;
 		}
 		break;
 	}
@@ -1094,6 +1254,7 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 		SPA_PARAM_PROFILE_name, SPA_POD_String(name),
 		SPA_PARAM_PROFILE_description, SPA_POD_String(desc),
 		SPA_PARAM_PROFILE_available, SPA_POD_Id(SPA_PARAM_AVAILABILITY_yes),
+		SPA_PARAM_PROFILE_priority, SPA_POD_Int(priority),
 		0);
 	if (n_source > 0 || n_sink > 0) {
 		spa_pod_builder_prop(b, SPA_PARAM_PROFILE_classes, 0);
@@ -1114,6 +1275,10 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 		}
 		spa_pod_builder_pop(b, &f[1]);
 	}
+	if (current) {
+                spa_pod_builder_prop(b, SPA_PARAM_PROFILE_save, 0);
+                spa_pod_builder_bool(b, this->save_profile);
+	}
 
 	if (name_and_codec)
 		free(name_and_codec);
@@ -1123,6 +1288,16 @@ static struct spa_pod *build_profile(struct impl *this, struct spa_pod_builder *
 	return spa_pod_builder_pop(b, &f[0]);
 }
 
+static bool validate_profile(struct impl *this, uint32_t profile,
+		enum spa_bluetooth_audio_codec codec)
+{
+	struct spa_pod_builder b = { 0 };
+	uint8_t buffer[1024];
+
+	spa_pod_builder_init(&b, buffer, sizeof(buffer));
+	return (build_profile(this, &b, 0, 0, profile, codec, false) != NULL);
+}
+
 static struct spa_pod *build_route(struct impl *this, struct spa_pod_builder *b,
 		uint32_t id, uint32_t port, uint32_t dev, uint32_t profile)
 {
@@ -1205,15 +1380,6 @@ static struct spa_pod *build_route(struct impl *this, struct spa_pod_builder *b,
 		return NULL;
 	}
 
-	if (dev != SPA_ID_INVALID && !(profile_direction_mask(this, this->profile) & (1 << direction)))
-		return NULL;
-
-	mask = 0;
-	for (i = 1; i < 4; i++)
-		mask |= profile_direction_mask(this, i);
-	if ((mask & (1 << direction)) == 0)
-		return NULL;
-
 	spa_pod_builder_push_object(b, &f[0], SPA_TYPE_OBJECT_ParamRoute, id);
 	spa_pod_builder_add(b,
 		SPA_PARAM_ROUTE_index, SPA_POD_Int(port),
@@ -1233,28 +1399,37 @@ static struct spa_pod *build_route(struct impl *this, struct spa_pod_builder *b,
 	spa_pod_builder_pop(b, &f[1]);
 	spa_pod_builder_prop(b, SPA_PARAM_ROUTE_profiles, 0);
 	spa_pod_builder_push_array(b, &f[1]);
-	for (i = 1; (j = get_profile_from_index(this, i, &next, &codec)) != SPA_ID_INVALID; i = next) {
-		struct spa_pod_builder b2 = { 0 };
-		uint8_t buffer[1024];
-		struct spa_pod *param;
 
-		if (!(profile_direction_mask(this, j) & (1 << direction)))
+	mask = 0;
+	for (i = 1; (j = get_profile_from_index(this, i, &next, &codec)) != SPA_ID_INVALID; i = next) {
+		uint32_t profile_mask;
+
+		profile_mask = profile_direction_mask(this, j, codec);
+		if (!(profile_mask & (1 << direction)))
 			continue;
 
 		/* Check the profile actually exists */
-		spa_pod_builder_init(&b2, buffer, sizeof(buffer));
-		param = build_profile(this, &b2, 0, i, j, codec);
-		if (param == NULL)
+		if (!validate_profile(this, j, codec))
 			continue;
 
+		mask |= profile_mask;
 		spa_pod_builder_int(b, i);
 	}
 	spa_pod_builder_pop(b, &f[1]);
 
+	if (!(mask & (1 << direction))) {
+		/* No profile has route direction */
+		return NULL;
+	}
+
 	if (dev != SPA_ID_INVALID) {
 		struct node *node = &this->nodes[dev];
 		struct spa_bt_transport_volume *t_volume;
 
+		mask = profile_direction_mask(this, this->profile, this->props.codec);
+		if (!(mask & (1 << direction)))
+			return NULL;
+
 		t_volume = node->transport
 			? &node->transport->volumes[node->id]
 			: NULL;
@@ -1306,15 +1481,34 @@ static struct spa_pod *build_route(struct impl *this, struct spa_pod_builder *b,
 	return spa_pod_builder_pop(b, &f[0]);
 }
 
+static bool iterate_supported_a2dp_codecs(struct impl *this, int *j, const struct a2dp_codec **codec)
+{
+	int i;
+
+next:
+	*j = *j + 1;
+	spa_assert(*j >= 0);
+	if ((size_t)*j >= this->supported_codec_count)
+		return false;
+
+	for (i = 0; i < *j; ++i)
+		if (this->supported_codecs[i]->id == this->supported_codecs[*j]->id)
+			goto next;
+
+	*codec = this->supported_codecs[*j];
+	return true;
+}
+
 static struct spa_pod *build_prop_info(struct impl *this, struct spa_pod_builder *b, uint32_t id)
 {
 	struct spa_pod_frame f[2];
 	struct spa_pod_choice *choice;
 	const struct a2dp_codec *codec;
-	size_t n, j;
+	size_t n;
+	int j;
 
 #define FOR_EACH_A2DP_CODEC(j, codec) \
-		for (j = 0; (j < this->supported_codec_count) ? (codec = this->supported_codecs[j]) : NULL; ++j)
+		for (j = -1; iterate_supported_a2dp_codecs(this, &j, &codec);)
 #define FOR_EACH_HFP_CODEC(j) \
 		for (j = HFP_AUDIO_CODEC_MSBC; j >= HFP_AUDIO_CODEC_CVSD; --j) \
 			if (spa_bt_device_supports_hfp_codec(this->bt_dev, j) == 1)
@@ -1416,7 +1610,7 @@ static int impl_enum_params(void *object, int seq,
 		case DEVICE_PROFILE_AG:
 		case DEVICE_PROFILE_A2DP:
 		case DEVICE_PROFILE_HSP_HFP:
-			param = build_profile(this, &b, id, result.index, profile, codec);
+			param = build_profile(this, &b, id, result.index, profile, codec, false);
 			if (param == NULL)
 				goto next;
 			break;
@@ -1432,7 +1626,7 @@ static int impl_enum_params(void *object, int seq,
 		switch (result.index) {
 		case 0:
 			index = get_index_from_profile(this, this->profile, this->props.codec);
-			param = build_profile(this, &b, id, index, this->profile, this->props.codec);
+			param = build_profile(this, &b, id, index, this->profile, this->props.codec, true);
 			if (param == NULL)
 				return 0;
 			break;
@@ -1516,7 +1710,7 @@ static int node_set_volume(struct impl *this, struct node *node, float volumes[]
 	if (n_volumes == 0)
 		return -EINVAL;
 
-	spa_log_debug(this->log, "node %p volume %f", node, volumes[0]);
+	spa_log_info(this->log, "node %p volume %f", node, volumes[0]);
 
 	for (i = 0; i < node->n_channels; i++) {
 		if (node->volumes[i] == volumes[i % n_volumes])
@@ -1535,10 +1729,13 @@ static int node_set_volume(struct impl *this, struct node *node, float volumes[]
 		node_update_soft_volumes(node, hw_volume);
 		spa_bt_transport_set_volume(node->transport, node->id, hw_volume);
 	} else {
+		float boost = get_soft_volume_boost(node);
 		for (uint32_t i = 0; i < node->n_channels; ++i)
-			node->soft_volumes[i] = node->volumes[i];
+			node->soft_volumes[i] = node->volumes[i] * boost;
 	}
 
+	emit_volume(this, node);
+
 	return changed;
 }
 
@@ -1564,7 +1761,8 @@ static int node_set_mute(struct impl *this, struct node *node, bool mute)
 
 	spa_pod_builder_add_object(&b,
 			SPA_TYPE_OBJECT_Props, SPA_EVENT_DEVICE_Props,
-			SPA_PROP_mute, SPA_POD_Bool(mute));
+			SPA_PROP_mute, SPA_POD_Bool(mute),
+			SPA_PROP_softMute, SPA_POD_Bool(mute));
 	event = spa_pod_builder_pop(&b, &f[0]);
 
 	spa_device_emit_event(&this->hooks, event);
@@ -1670,32 +1868,33 @@ static int impl_set_param(void *object,
 	switch (id) {
 	case SPA_PARAM_Profile:
 	{
-		uint32_t id, next;
+		uint32_t idx, next;
 		uint32_t profile;
 		enum spa_bluetooth_audio_codec codec;
+		bool save = false;
 
 		if (param == NULL)
 			return -EINVAL;
 
 		if ((res = spa_pod_parse_object(param,
 				SPA_TYPE_OBJECT_ParamProfile, NULL,
-				SPA_PARAM_PROFILE_index, SPA_POD_Int(&id))) < 0) {
+				SPA_PARAM_PROFILE_index, SPA_POD_Int(&idx),
+				SPA_PARAM_PROFILE_save, SPA_POD_OPT_Bool(&save))) < 0) {
 			spa_log_warn(this->log, "can't parse profile");
 			spa_debug_pod(0, NULL, param);
 			return res;
 		}
 
-		profile = get_profile_from_index(this, id, &next, &codec);
+		profile = get_profile_from_index(this, idx, &next, &codec);
 		if (profile == SPA_ID_INVALID)
 			return -EINVAL;
 
-		spa_log_debug(this->log, NAME": setting profile %d codec:%d", profile, codec);
-		set_profile(this, profile, codec);
-		break;
+		spa_log_debug(this->log, "setting profile %d codec:%d save:%d", profile, codec, (int)save);
+		return set_profile(this, profile, codec, save);
 	}
 	case SPA_PARAM_Route:
 	{
-		uint32_t id, device;
+		uint32_t idx, device;
 		struct spa_pod *props = NULL;
 		struct node *node;
 		bool save = false;
@@ -1705,7 +1904,7 @@ static int impl_set_param(void *object,
 
 		if ((res = spa_pod_parse_object(param,
 				SPA_TYPE_OBJECT_ParamRoute, NULL,
-				SPA_PARAM_ROUTE_index, SPA_POD_Int(&id),
+				SPA_PARAM_ROUTE_index, SPA_POD_Int(&idx),
 				SPA_PARAM_ROUTE_device, SPA_POD_Int(&device),
 				SPA_PARAM_ROUTE_props, SPA_POD_OPT_Pod(&props),
 				SPA_PARAM_ROUTE_save, SPA_POD_OPT_Bool(&save))) < 0) {
@@ -1725,8 +1924,6 @@ static int impl_set_param(void *object,
 				this->params[IDX_Route].flags ^= SPA_PARAM_INFO_SERIAL;
 			}
 			emit_info(this, false);
-			/* See volume_changed(void *) */
-			emit_volume(this, node);
 		}
 		break;
 	}
@@ -1745,6 +1942,8 @@ static int impl_set_param(void *object,
 			return res;
 		}
 
+		spa_log_debug(this->log, "setting props codec:%d", codec_id);
+
 		if (codec_id == SPA_ID_INVALID)
 			return 0;
 
@@ -1752,19 +1951,16 @@ static int impl_set_param(void *object,
 			size_t j;
 			for (j = 0; j < this->supported_codec_count; ++j) {
 				if (this->supported_codecs[j]->id == codec_id) {
-					set_profile(this, this->profile, codec_id);
-					return 0;
+					return set_profile(this, this->profile, codec_id, true);
 				}
 			}
 		} else if (this->profile == DEVICE_PROFILE_HSP_HFP) {
 			if (codec_id == SPA_BLUETOOTH_AUDIO_CODEC_CVSD &&
 					spa_bt_device_supports_hfp_codec(this->bt_dev, HFP_AUDIO_CODEC_CVSD) == 1) {
-				set_profile(this, this->profile, codec_id);
-				return 0;
+				return set_profile(this, this->profile, codec_id, true);
 			} else if (codec_id == SPA_BLUETOOTH_AUDIO_CODEC_MSBC &&
 					spa_bt_device_supports_hfp_codec(this->bt_dev, HFP_AUDIO_CODEC_MSBC) == 1) {
-				set_profile(this, this->profile, codec_id);
-				return 0;
+				return set_profile(this, this->profile, codec_id, true);
 			}
 		}
 		return -EINVAL;
@@ -1792,7 +1988,7 @@ static int impl_get_interface(struct spa_handle *handle, const char *type, void
 
 	this = (struct impl *) handle;
 
-	if (strcmp(type, SPA_TYPE_INTERFACE_Device) == 0)
+	if (spa_streq(type, SPA_TYPE_INTERFACE_Device))
 		*interface = &this->device;
 	else
 		return -ENOENT;
@@ -1869,6 +2065,8 @@ impl_init(const struct spa_handle_factory *factory,
 	this->log = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_Log);
 	_i18n = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_I18N);
 
+	spa_log_topic_init(this->log, &log_topic);
+
 	if (info && (str = spa_dict_lookup(info, SPA_KEY_API_BLUEZ5_DEVICE)))
 		sscanf(str, "pointer:%p", &this->bt_dev);
 
