@@ -32,9 +32,11 @@
 #include <spa/support/loop.h>
 #include <spa/support/log.h>
 #include <spa/support/system.h>
+#include <spa/utils/result.h>
 #include <spa/utils/list.h>
 #include <spa/utils/keys.h>
 #include <spa/utils/names.h>
+#include <spa/utils/string.h>
 #include <spa/monitor/device.h>
 
 #include <spa/node/node.h>
@@ -42,6 +44,7 @@
 #include <spa/node/io.h>
 #include <spa/node/keys.h>
 #include <spa/param/param.h>
+#include <spa/param/latency-utils.h>
 #include <spa/param/audio/format.h>
 #include <spa/param/audio/format-utils.h>
 #include <spa/pod/filter.h>
@@ -50,9 +53,16 @@
 
 #include "defs.h"
 
+static struct spa_log_topic log_topic = SPA_LOG_TOPIC(0, "spa.bluez5.sink.sco");
+#undef SPA_LOG_TOPIC_DEFAULT
+#define SPA_LOG_TOPIC_DEFAULT &log_topic
+
+#define DEFAULT_CLOCK_NAME	"clock.system.monotonic"
+
 struct props {
 	uint32_t min_latency;
 	uint32_t max_latency;
+	char clock_name[64];
 };
 
 #define MAX_BUFFERS 32
@@ -74,7 +84,15 @@ struct port {
 	struct spa_port_info info;
 	struct spa_io_buffers *io;
 	struct spa_io_rate_match *rate_match;
-	struct spa_param_info params[8];
+	struct spa_latency_info latency;
+#define IDX_EnumFormat	0
+#define IDX_Meta	1
+#define IDX_IO		2
+#define IDX_Format	3
+#define IDX_Buffers	4
+#define IDX_Latency	5
+#define N_PORT_PARAMS	6
+	struct spa_param_info params[N_PORT_PARAMS];
 
 	struct buffer buffers[MAX_BUFFERS];
 	uint32_t n_buffers;
@@ -104,7 +122,10 @@ struct impl {
 	/* Info */
 	uint64_t info_all;
 	struct spa_node_info info;
-	struct spa_param_info params[8];
+#define IDX_PropInfo	0
+#define IDX_Props	1
+#define N_NODE_PARAMS	2
+	struct spa_param_info params[N_NODE_PARAMS];
 	struct props props;
 
 	/* Transport */
@@ -116,6 +137,7 @@ struct impl {
 
 	/* Flags */
 	unsigned int started:1;
+	unsigned int following:1;
 
 	/* Sources */
 	struct spa_source source;
@@ -138,8 +160,6 @@ struct impl {
 	uint64_t total_samples;
 };
 
-#define NAME "sco-sink"
-
 #define CHECK_PORT(this,d,p)	((d) == SPA_DIRECTION_INPUT && (p) == 0)
 
 static const uint32_t default_min_latency = MIN_LATENCY;
@@ -150,6 +170,7 @@ static void reset_props(struct props *props)
 {
 	props->min_latency = default_min_latency;
 	props->max_latency = default_max_latency;
+	strncpy(props->clock_name, DEFAULT_CLOCK_NAME, sizeof(props->clock_name));
 }
 
 static int impl_node_enum_params(void *object, int seq,
@@ -229,24 +250,26 @@ static int impl_node_enum_params(void *object, int seq,
 	return 0;
 }
 
-static void set_timeout(struct impl *this, uint64_t timeout)
+static int set_timeout(struct impl *this, uint64_t time)
 {
 	struct itimerspec ts;
-
-	ts.it_value.tv_sec = timeout / SPA_NSEC_PER_SEC;
-	ts.it_value.tv_nsec = timeout % SPA_NSEC_PER_SEC;
+	ts.it_value.tv_sec = time / SPA_NSEC_PER_SEC;
+	ts.it_value.tv_nsec = time % SPA_NSEC_PER_SEC;
 	ts.it_interval.tv_sec = 0;
 	ts.it_interval.tv_nsec = 0;
+	return spa_system_timerfd_settime(this->data_system,
+			this->timerfd, 0, &ts, NULL);
+}
 
-	spa_system_timerfd_settime(this->data_system, this->timerfd, 0, &ts, NULL);
-	this->source.mask = SPA_IO_IN;
-	spa_loop_update_source(this->data_loop, &this->source);
+static int set_timers(struct impl *this)
+{
+	return set_timeout(this, this->following ? 0 : 1);
 }
 
 static uint64_t get_next_timeout(struct impl *this, uint64_t now_time, uint64_t processed_samples)
 {
 	struct port *port = &this->port;
-	uint64_t playback_time = 0, elapsed_time = 0, next_time = 1;
+	uint64_t playback_time = 0, elapsed_time = 0, next_time = 0;
 
 	this->total_samples += processed_samples;
 
@@ -259,15 +282,38 @@ static uint64_t get_next_timeout(struct impl *this, uint64_t now_time, uint64_t
 	return next_time;
 }
 
+static int do_reassign_follower(struct spa_loop *loop,
+			bool async,
+			uint32_t seq,
+			const void *data,
+			size_t size,
+			void *user_data)
+{
+	struct impl *this = user_data;
+	set_timers(this);
+	return 0;
+}
+
+static inline bool is_following(struct impl *this)
+{
+	return this->position && this->clock && this->position->clock.id != this->clock->id;
+}
+
 static int impl_node_set_io(void *object, uint32_t id, void *data, size_t size)
 {
 	struct impl *this = object;
+	bool following;
 
-	spa_return_val_if_fail(object != NULL, -EINVAL);
+	spa_return_val_if_fail(this != NULL, -EINVAL);
 
 	switch (id) {
 	case SPA_IO_Clock:
 		this->clock = data;
+		if (this->clock != NULL) {
+			spa_scnprintf(this->clock->name,
+					sizeof(this->clock->name),
+					"%s", this->props.clock_name);
+		}
 		break;
 	case SPA_IO_Position:
 		this->position = data;
@@ -276,6 +322,12 @@ static int impl_node_set_io(void *object, uint32_t id, void *data, size_t size)
 		return -ENOENT;
 	}
 
+	following = is_following(this);
+	if (this->started && following != this->following) {
+		spa_log_debug(this->log, "%p: reassign follower %d->%d", this, this->following, following);
+		this->following = following;
+		spa_loop_invoke(this->data_loop, do_reassign_follower, 0, NULL, 0, true, this);
+	}
 	return 0;
 }
 
@@ -312,7 +364,7 @@ static int impl_node_set_param(void *object, uint32_t id, uint32_t flags,
 	{
 		if (apply_props(this, param) > 0) {
 			this->info.change_mask |= SPA_NODE_CHANGE_MASK_PARAMS;
-			this->params[1].flags ^= SPA_PARAM_INFO_SERIAL;
+			this->params[IDX_Props].flags ^= SPA_PARAM_INFO_SERIAL;
 			emit_node_info(this, false);
 		}
 		break;
@@ -328,62 +380,51 @@ static void flush_data(struct impl *this)
 {
 	struct port *port = &this->port;
 	struct spa_data *datas;
-	uint64_t next_timeout = 1;
-	uint32_t min_in_size;
-	uint8_t *packet;
+	const uint32_t min_in_size =
+		(this->transport->codec == HFP_AUDIO_CODEC_MSBC) ?
+		MSBC_DECODED_SIZE : this->transport->write_mtu;
+	uint8_t * const packet =
+		(this->transport->codec == HFP_AUDIO_CODEC_MSBC) ?
+		this->buffer_head : port->write_buffer;
 
 	if (this->transport == NULL || this->transport->sco_io == NULL)
 		return;
 
-	/* get buffer */
-	if (!port->current_buffer) {
-		spa_return_if_fail(!spa_list_is_empty(&port->ready));
-		port->current_buffer = spa_list_first(&port->ready, struct buffer, link);
-		port->ready_offset = 0;
-	}
-	datas = port->current_buffer->buf->datas;
+again:
+	while (!spa_list_is_empty(&port->ready) && port->write_buffer_size < min_in_size) {
+		/* get buffer */
+		if (!port->current_buffer) {
+			spa_return_if_fail(!spa_list_is_empty(&port->ready));
+			port->current_buffer = spa_list_first(&port->ready, struct buffer, link);
+			port->ready_offset = 0;
+		}
+		datas = port->current_buffer->buf->datas;
 
-	if (this->transport->codec == HFP_AUDIO_CODEC_MSBC) {
-		min_in_size = MSBC_DECODED_SIZE;
-		packet = this->buffer_head;
-	} else {
-		min_in_size = this->transport->write_mtu;
-		packet = port->write_buffer;
-	}
+		/* if buffer has data, copy it into the write buffer */
+		if (datas[0].chunk->size - port->ready_offset > 0) {
+			const uint32_t avail =
+				SPA_MIN(min_in_size, datas[0].chunk->size - port->ready_offset);
+			const uint32_t size =
+				(avail + port->write_buffer_size) > min_in_size ?
+				min_in_size - port->write_buffer_size : avail;
+			memcpy(port->write_buffer + port->write_buffer_size,
+				(uint8_t *)datas[0].data + port->ready_offset,
+				size);
+			port->write_buffer_size += size;
+			port->ready_offset += size;
+		} else {
+			struct buffer *b;
 
-	/* if buffer has data, copy it into the write buffer */
-	if (datas[0].chunk->size - port->ready_offset > 0) {
-		uint32_t avail = SPA_MIN(min_in_size, datas[0].chunk->size - port->ready_offset);
-		uint32_t size = (avail + port->write_buffer_size) > min_in_size ? min_in_size - port->write_buffer_size : avail;
-		memcpy(port->write_buffer + port->write_buffer_size,
-			(uint8_t *)datas[0].data + port->ready_offset,
-			size);
-		port->write_buffer_size += size;
-		port->ready_offset += size;
-	}
+			b = port->current_buffer;
+			port->current_buffer = NULL;
 
-	/* otherwise request a new buffer */
-	else {
-		struct buffer *b;
-
-		b = port->current_buffer;
-		port->current_buffer = NULL;
-
-		/* reuse buffer */
-		spa_list_remove(&b->link);
-		b->outstanding = true;
-		spa_log_trace(this->log, "sco-sink %p: reuse buffer %u", this, b->id);
-		port->io->buffer_id = b->id;
-		spa_node_call_reuse_buffer(&this->callbacks, 0, b->id);
-
-		/* notify we need more data */
-		port->io->status = SPA_STATUS_NEED_DATA;
-		spa_node_call_ready(&this->callbacks, SPA_STATUS_NEED_DATA);
-
-		next_timeout = (this->transport->write_mtu / port->frame_size
-				* SPA_NSEC_PER_SEC / port->current_format.info.raw.rate);
-		set_timeout(this, next_timeout);
-		return;
+			/* reuse buffer */
+			spa_list_remove(&b->link);
+			b->outstanding = true;
+			spa_log_trace(this->log, "sco-sink %p: reuse buffer %u", this, b->id);
+			port->io->buffer_id = b->id;
+			spa_node_call_reuse_buffer(&this->callbacks, 0, b->id);
+		}
 	}
 
 	/* send the data if the write buffer is full */
@@ -393,6 +434,7 @@ static void flush_data(struct impl *this)
 		int processed = 0;
 		ssize_t out_encoded;
 		int written;
+		uint64_t next_timeout;
 
 		spa_system_clock_gettime(this->data_system, CLOCK_MONOTONIC, &this->now);
 		now_time = SPA_TIMESPEC_TO_NSEC(&this->now);
@@ -420,12 +462,13 @@ static void flush_data(struct impl *this)
 			port->write_buffer_size = 0;
 
 			/* Write */
-			written = spa_bt_sco_io_write(this->transport->sco_io, packet, this->buffer_next - this->buffer_head);
+			written = spa_bt_sco_io_write(this->transport->sco_io, packet,
+					this->buffer_next - this->buffer_head);
 			if (written < 0) {
-				spa_log_warn(this->log, "failed to write data");
+				spa_log_warn(this->log, "failed to write data: %d (%s)",
+						written, spa_strerror(written));
 				goto stop;
 			}
-			spa_log_trace(this->log, "wrote socket data %d", written);
 
 			this->buffer_head += written;
 
@@ -441,10 +484,15 @@ static void flush_data(struct impl *this)
 				this->buffer_head = this->buffer;
 			}
 		} else {
-			written = spa_bt_sco_io_write(this->transport->sco_io, packet, port->write_buffer_size);
+			written = spa_bt_sco_io_write(this->transport->sco_io, packet,
+					port->write_buffer_size);
 			if (written < 0) {
-				spa_log_warn(this->log, "sco-sink: write failure: %d", written);
+				spa_log_warn(this->log, "sco-sink: write failure: %d (%s)",
+						written, spa_strerror(written));
 				goto stop;
+			} else if (written == 0) {
+				/* EAGAIN or similar, just skip ahead */
+				written = SPA_MIN(port->write_buffer_size, (uint32_t)48);
 			}
 
 			processed = written;
@@ -455,19 +503,33 @@ static void flush_data(struct impl *this)
 			}
 		}
 
+		spa_log_trace(this->log, "write socket data %d", written);
+
 		next_timeout = get_next_timeout(this, now_time, processed / port->frame_size);
 
-		if (this->clock) {
+		if (!this->following && this->clock) {
 			this->clock->nsec = now_time;
 			this->clock->position = this->total_samples;
 			this->clock->delay = processed / port->frame_size;
 			this->clock->rate_diff = 1.0f;
-			this->clock->next_nsec = next_timeout;
+			this->clock->next_nsec = now_time + next_timeout;
 		}
+
+		if (next_timeout == 0)
+			goto again;
+
+		spa_log_trace(this->log, "timeout %"PRIu64" ns", next_timeout);
+		set_timeout(this, next_timeout);
+	} else {
+		/* As follower, driver will wake us up when there is data */
+		if (this->following)
+			return;
+
+		/* As driver, run timeout now to schedule data */
+		spa_log_trace(this->log, "timeout 1 ns (driver: schedule now)");
+		set_timeout(this, 1);
 	}
 
-	/* schedule next timeout */
-	set_timeout(this, next_timeout);
 	return;
 
 stop:
@@ -489,7 +551,7 @@ static void sco_on_timeout(struct spa_source *source)
 		spa_log_warn(this->log, "error reading timerfd: %s", strerror(errno));
 
 	/* delay if no buffers available */
-	if (spa_list_is_empty(&port->ready)) {
+	if (!this->following && spa_list_is_empty(&port->ready)) {
 		set_timeout(this, this->transport->write_mtu / port->frame_size * SPA_NSEC_PER_SEC / port->current_format.info.raw.rate);
 		port->io->status = SPA_STATUS_NEED_DATA;
 		spa_node_call_ready(&this->callbacks, SPA_STATUS_NEED_DATA);
@@ -526,17 +588,21 @@ static int do_start(struct impl *this)
 	/* Make sure the transport is valid */
 	spa_return_val_if_fail(this->transport != NULL, -EIO);
 
+	this->following = is_following(this);
+
+	spa_log_debug(this->log, "%p: start following:%d", this, this->following);
+
 	/* Do accept if Gateway; otherwise do connect for Head Unit */
 	do_accept = this->transport->profile & SPA_BT_PROFILE_HEADSET_AUDIO_GATEWAY;
 
-	/* acquire the socked fd (false -> connect | true -> accept) */
+	/* acquire the socket fd (false -> connect | true -> accept) */
 	if ((res = spa_bt_transport_acquire(this->transport, do_accept)) < 0)
 		return res;
 
 	/* Init mSBC if needed */
 	if (this->transport->codec == HFP_AUDIO_CODEC_MSBC) {
 		sbc_init_msbc(&this->msbc, 0);
-		/* Libsbc expects audio samples by default in host endianity, mSBC requires little endian */
+		/* Libsbc expects audio samples by default in host endianness, mSBC requires little endian */
 		this->msbc.endian = SBC_LE;
 
 		/* write_mtu might not be correct at this point, so we'll throw
@@ -565,7 +631,7 @@ static int do_start(struct impl *this)
 	spa_loop_add_source(this->data_loop, &this->source);
 
 	/* start processing */
-	set_timeout(this, 1);
+	set_timers(this);
 
 	/* Set the started flag */
 	this->started = true;
@@ -680,18 +746,21 @@ static int impl_node_send_command(void *object, const struct spa_command *comman
 
 static void emit_node_info(struct impl *this, bool full)
 {
-	bool is_ag = (this->transport->profile & SPA_BT_PROFILE_HEADSET_AUDIO_GATEWAY);
-	struct spa_dict_item ag_node_info_items[] = {
+	static const struct spa_dict_item hu_node_info_items[] = {
+		{ SPA_KEY_DEVICE_API, "bluez5" },
+		{ SPA_KEY_MEDIA_CLASS, "Audio/Sink" },
+		{ SPA_KEY_NODE_DRIVER, "true" },
+	};
+
+	const struct spa_dict_item ag_node_info_items[] = {
 		{ SPA_KEY_DEVICE_API, "bluez5" },
 		{ SPA_KEY_MEDIA_CLASS, "Stream/Input/Audio" },
 		{ "media.name", ((this->transport && this->transport->device->name) ?
 		                 this->transport->device->name : "HSP/HFP") },
 	};
-	struct spa_dict_item hu_node_info_items[] = {
-		{ SPA_KEY_DEVICE_API, "bluez5" },
-		{ SPA_KEY_MEDIA_CLASS, "Audio/Sink" },
-		{ SPA_KEY_NODE_DRIVER, "true" },
-	};
+	bool is_ag = this->transport &&
+		(this->transport->profile & SPA_BT_PROFILE_HEADSET_AUDIO_GATEWAY);
+	uint64_t old = full ? this->info.change_mask : 0;
 
 	if (full)
 		this->info.change_mask = this->info_all;
@@ -700,18 +769,19 @@ static void emit_node_info(struct impl *this, bool full)
 			&SPA_DICT_INIT_ARRAY(ag_node_info_items) :
 			&SPA_DICT_INIT_ARRAY(hu_node_info_items);
 		spa_node_emit_info(&this->hooks, &this->info);
-		this->info.change_mask = 0;
+		this->info.change_mask = old;
 	}
 }
 
 static void emit_port_info(struct impl *this, struct port *port, bool full)
 {
+	uint64_t old = full ? port->info.change_mask : 0;
 	if (full)
 		port->info.change_mask = port->info_all;
 	if (port->info.change_mask) {
 		spa_node_emit_port_info(&this->hooks,
 				SPA_DIRECTION_INPUT, 0, &port->info);
-		port->info.change_mask = 0;
+		port->info.change_mask = old;
 	}
 }
 
@@ -848,8 +918,7 @@ impl_node_port_enum_params(void *object, int seq,
 							this->props.max_latency * port->frame_size,
 							this->props.min_latency * port->frame_size,
 							INT32_MAX),
-			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(port->frame_size),
-			SPA_PARAM_BUFFERS_align,   SPA_POD_Int(16));
+			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(port->frame_size));
 		break;
 
 	case SPA_PARAM_Meta:
@@ -884,6 +953,16 @@ impl_node_port_enum_params(void *object, int seq,
 		}
 		break;
 
+	case SPA_PARAM_Latency:
+		switch (result.index) {
+		case 0:
+			param = spa_latency_build(&b, id, &port->latency);
+			break;
+		default:
+			return 0;
+		}
+		break;
+
 	default:
 		return -ENOENT;
 	}
@@ -943,11 +1022,12 @@ static int port_set_format(struct impl *this, struct port *port,
 		port->info.flags = SPA_PORT_FLAG_LIVE;
 		port->info.change_mask |= SPA_PORT_CHANGE_MASK_RATE;
 		port->info.rate = SPA_FRACTION(1, port->current_format.info.raw.rate);
-		port->params[3] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_READWRITE);
-		port->params[4] = SPA_PARAM_INFO(SPA_PARAM_Buffers, SPA_PARAM_INFO_READ);
+		port->params[IDX_Format] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_READWRITE);
+		port->params[IDX_Buffers] = SPA_PARAM_INFO(SPA_PARAM_Buffers, SPA_PARAM_INFO_READ);
+		port->params[IDX_Latency].flags ^= SPA_PARAM_INFO_SERIAL;
 	} else {
-		port->params[3] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
-		port->params[4] = SPA_PARAM_INFO(SPA_PARAM_Buffers, 0);
+		port->params[IDX_Format] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
+		port->params[IDX_Buffers] = SPA_PARAM_INFO(SPA_PARAM_Buffers, 0);
 	}
 	emit_port_info(this, port, false);
 
@@ -972,6 +1052,9 @@ impl_node_port_set_param(void *object,
 	case SPA_PARAM_Format:
 		res = port_set_format(this, port, flags, param);
 		break;
+	case SPA_PARAM_Latency:
+		res = 0;
+		break;
 	default:
 		res = -ENOENT;
 		break;
@@ -1010,7 +1093,7 @@ impl_node_port_use_buffers(void *object,
 		b->h = spa_buffer_find_meta_data(buffers[i], SPA_META_Header, sizeof(*b->h));
 
 		if (buffers[i]->datas[0].data == NULL) {
-			spa_log_error(this->log, NAME " %p: need mapped memory", this);
+			spa_log_error(this->log, "%p: need mapped memory", this);
 			return -EINVAL;
 		}
 	}
@@ -1068,12 +1151,12 @@ static int impl_node_process(void *object)
 		struct buffer *b = &port->buffers[io->buffer_id];
 
 		if (!b->outstanding) {
-			spa_log_warn(this->log, NAME " %p: buffer %u in use", this, io->buffer_id);
+			spa_log_warn(this->log, "%p: buffer %u in use", this, io->buffer_id);
 			io->status = -EINVAL;
 			return -EINVAL;
 		}
 
-		spa_log_trace(this->log, NAME " %p: queue buffer %u", this, io->buffer_id);
+		spa_log_trace(this->log, "%p: queue buffer %u", this, io->buffer_id);
 
 		spa_list_append(&port->ready, &b->link);
 		b->outstanding = false;
@@ -1139,7 +1222,7 @@ static int impl_get_interface(struct spa_handle *handle, const char *type, void
 
 	this = (struct impl *) handle;
 
-	if (strcmp(type, SPA_TYPE_INTERFACE_Node) == 0)
+	if (spa_streq(type, SPA_TYPE_INTERFACE_Node))
 		*interface = &this->node;
 	else
 		return -ENOENT;
@@ -1186,6 +1269,8 @@ impl_init(const struct spa_handle_factory *factory,
 	this->data_loop = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_DataLoop);
 	this->data_system = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_DataSystem);
 
+	spa_log_topic_init(this->log, &log_topic);
+
 	if (this->data_loop == NULL) {
 		spa_log_error(this->log, "a data loop is needed");
 		return -EINVAL;
@@ -1210,23 +1295,29 @@ impl_init(const struct spa_handle_factory *factory,
 	this->info.max_input_ports = 1;
 	this->info.max_output_ports = 0;
 	this->info.flags = SPA_NODE_FLAG_RT;
-	this->params[0] = SPA_PARAM_INFO(SPA_PARAM_PropInfo, SPA_PARAM_INFO_READ);
-	this->params[1] = SPA_PARAM_INFO(SPA_PARAM_Props, SPA_PARAM_INFO_READWRITE);
+	this->params[IDX_PropInfo] = SPA_PARAM_INFO(SPA_PARAM_PropInfo, SPA_PARAM_INFO_READ);
+	this->params[IDX_Props] = SPA_PARAM_INFO(SPA_PARAM_Props, SPA_PARAM_INFO_READWRITE);
 	this->info.params = this->params;
-	this->info.n_params = 2;
+	this->info.n_params = N_NODE_PARAMS;
 
 	port = &this->port;
 	port->info_all = SPA_PORT_CHANGE_MASK_FLAGS |
 			SPA_PORT_CHANGE_MASK_PARAMS;
 	port->info = SPA_PORT_INFO_INIT();
 	port->info.flags = 0;
-	port->params[0] = SPA_PARAM_INFO(SPA_PARAM_EnumFormat, SPA_PARAM_INFO_READ);
-	port->params[1] = SPA_PARAM_INFO(SPA_PARAM_Meta, SPA_PARAM_INFO_READ);
-	port->params[2] = SPA_PARAM_INFO(SPA_PARAM_IO, SPA_PARAM_INFO_READ);
-	port->params[3] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
-	port->params[4] = SPA_PARAM_INFO(SPA_PARAM_Buffers, 0);
+	port->params[IDX_EnumFormat] = SPA_PARAM_INFO(SPA_PARAM_EnumFormat, SPA_PARAM_INFO_READ);
+	port->params[IDX_Meta] = SPA_PARAM_INFO(SPA_PARAM_Meta, SPA_PARAM_INFO_READ);
+	port->params[IDX_IO] = SPA_PARAM_INFO(SPA_PARAM_IO, SPA_PARAM_INFO_READ);
+	port->params[IDX_Format] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
+	port->params[IDX_Buffers] = SPA_PARAM_INFO(SPA_PARAM_Buffers, 0);
+	port->params[IDX_Latency] = SPA_PARAM_INFO(SPA_PARAM_Latency, SPA_PARAM_INFO_READWRITE);
 	port->info.params = port->params;
-	port->info.n_params = 5;
+	port->info.n_params = N_PORT_PARAMS;
+
+	port->latency = SPA_LATENCY_INFO(SPA_DIRECTION_INPUT);
+	port->latency.min_quantum = 1.0f;
+	port->latency.max_quantum = 1.0f;
+
 	spa_list_init(&port->ready);
 
 	if (info && (str = spa_dict_lookup(info, SPA_KEY_API_BLUEZ5_TRANSPORT)))
@@ -1276,7 +1367,7 @@ static const struct spa_dict_item info_items[] = {
 
 static const struct spa_dict info = SPA_DICT_INIT_ARRAY(info_items);
 
-struct spa_handle_factory spa_sco_sink_factory = {
+const struct spa_handle_factory spa_sco_sink_factory = {
 	SPA_VERSION_HANDLE_FACTORY,
 	SPA_NAME_API_BLUEZ5_SCO_SINK,
 	&info,
