@@ -37,6 +37,7 @@
 #include <spa/utils/list.h>
 #include <spa/utils/keys.h>
 #include <spa/utils/names.h>
+#include <spa/utils/string.h>
 #include <spa/monitor/device.h>
 
 #include <spa/node/node.h>
@@ -44,6 +45,7 @@
 #include <spa/node/io.h>
 #include <spa/node/keys.h>
 #include <spa/param/param.h>
+#include <spa/param/latency-utils.h>
 #include <spa/param/audio/format.h>
 #include <spa/param/audio/format-utils.h>
 #include <spa/pod/filter.h>
@@ -52,9 +54,16 @@
 
 #include "defs.h"
 
+static struct spa_log_topic log_topic = SPA_LOG_TOPIC(0, "spa.bluez5.source.sco");
+#undef SPA_LOG_TOPIC_DEFAULT
+#define SPA_LOG_TOPIC_DEFAULT &log_topic
+
+#define DEFAULT_CLOCK_NAME	"clock.system.monotonic"
+
 struct props {
 	uint32_t min_latency;
 	uint32_t max_latency;
+	char clock_name[64];
 };
 
 #define MAX_BUFFERS 32
@@ -76,7 +85,15 @@ struct port {
 	struct spa_port_info info;
 	struct spa_io_buffers *io;
 	struct spa_io_rate_match *rate_match;
-	struct spa_param_info params[8];
+	struct spa_latency_info latency;
+#define IDX_EnumFormat	0
+#define IDX_Meta	1
+#define IDX_IO		2
+#define IDX_Format	3
+#define IDX_Buffers	4
+#define IDX_Latency	5
+#define N_PORT_PARAMS	6
+	struct spa_param_info params[N_PORT_PARAMS];
 
 	struct buffer buffers[MAX_BUFFERS];
 	uint32_t n_buffers;
@@ -101,7 +118,11 @@ struct impl {
 
 	uint64_t info_all;
 	struct spa_node_info info;
-	struct spa_param_info params[8];
+#define IDX_PropInfo	0
+#define IDX_Props	1
+#define IDX_NODE_IO	2
+#define N_NODE_PARAMS	3
+	struct spa_param_info params[N_NODE_PARAMS];
 	struct props props;
 
 	struct spa_bt_transport *transport;
@@ -127,8 +148,6 @@ struct impl {
 	struct timespec now;
 };
 
-#define NAME "sco-source"
-
 #define CHECK_PORT(this,d,p)	((d) == SPA_DIRECTION_OUTPUT && (p) == 0)
 
 static const uint32_t default_min_latency = 128;
@@ -138,6 +157,7 @@ static void reset_props(struct props *props)
 {
 	props->min_latency = default_min_latency;
 	props->max_latency = default_max_latency;
+	strncpy(props->clock_name, DEFAULT_CLOCK_NAME, sizeof(props->clock_name));
 }
 
 static int impl_node_enum_params(void *object, int seq,
@@ -232,6 +252,11 @@ static int impl_node_set_io(void *object, uint32_t id, void *data, size_t size)
 	switch (id) {
 	case SPA_IO_Clock:
 		this->clock = data;
+		if (this->clock != NULL) {
+			spa_scnprintf(this->clock->name,
+					sizeof(this->clock->name),
+					"%s", this->props.clock_name);
+		}
 		break;
 	case SPA_IO_Position:
 		this->position = data;
@@ -242,7 +267,7 @@ static int impl_node_set_io(void *object, uint32_t id, void *data, size_t size)
 
 	following = is_following(this);
 	if (this->started && following != this->following) {
-		spa_log_debug(this->log, NAME" %p: reassign follower %d->%d", this, this->following, following);
+		spa_log_debug(this->log, "%p: reassign follower %d->%d", this, this->following, following);
 		this->following = following;
 	}
 
@@ -282,7 +307,7 @@ static int impl_node_set_param(void *object, uint32_t id, uint32_t flags,
 	{
 		if (apply_props(this, param) > 0) {
 			this->info.change_mask |= SPA_NODE_CHANGE_MASK_PARAMS;
-			this->params[1].flags ^= SPA_PARAM_INFO_SERIAL;
+			this->params[IDX_Props].flags ^= SPA_PARAM_INFO_SERIAL;
 			emit_node_info(this, false);
 		}
 		break;
@@ -315,7 +340,7 @@ static void recycle_buffer(struct impl *this, struct port *port, uint32_t buffer
 	struct buffer *b = &port->buffers[buffer_id];
 
 	if (b->outstanding) {
-		spa_log_trace(this->log, NAME " %p: recycle buffer %u", this, buffer_id);
+		spa_log_trace(this->log, "%p: recycle buffer %u", this, buffer_id);
 		spa_list_append(&port->free, &b->link);
 		b->outstanding = false;
 	}
@@ -368,26 +393,22 @@ static void msbc_buffer_append_byte(struct impl *this, uint8_t byte)
         ++this->msbc_buffer_pos;
 }
 
-/*
-   Helper function for easier debugging
-   Caveat: If size_read is not a multiple of 16, then the last bytes
-   will not be printed / logged
-*/
-static void hexdump_to_log(void *log, uint8_t *read_data, int size_read)
+/* Helper function for debugging */
+static SPA_UNUSED void hexdump_to_log(struct impl *this, uint8_t *data, size_t size)
 {
-
-	int rowsize = 16*3;
-	int line_idx = 0;
-	char hexline[16*3] = "";
-	int i;
-	for (i = 0; i < size_read; ++i) {
-		snprintf(&hexline[line_idx], 4, "%02X ", read_data[i]);
-		line_idx += 3;
-		if (line_idx == rowsize) {
-			spa_log_trace(log, "Processing read data: read_data %02i: %s", i-15, hexline);
-			line_idx = 0;
-		}
+	char buf[2048];
+	size_t i, col = 0, pos = 0;
+	buf[0] = '\0';
+	for (i = 0; i < size; ++i) {
+		int res;
+		res = spa_scnprintf(buf + pos, sizeof(buf) - pos, "%s%02x",
+				(col == 0) ? "\n\t" : " ", data[i]);
+		if (res < 0)
+			break;
+		pos += res;
+		col = (col + 1) % 16;
 	}
+	spa_log_trace(this->log, "hexdump (%d bytes):%s", (int)size, buf);
 }
 
 /* helper function to detect if a packet consists only of zeros */
@@ -409,12 +430,6 @@ static void preprocess_and_decode_msbc_data(void *userdata, uint8_t *read_data,
 
 	spa_log_trace(this->log, "handling mSBC data");
 
-	/* print hexdump of package  */
-	bool flag_hexdump_to_log = false;
-	if (flag_hexdump_to_log) {
-		hexdump_to_log(this->log, read_data, size_read);
-	}
-
 	/* check if the packet contains only zeros - if so ignore the packet.
 	   This is necessary, because some kernels insert bogus "all-zero" packets
 	   into the datastream.
@@ -508,12 +523,22 @@ static int sco_source_cb(void *userdata, uint8_t *read_data, int size_read)
 
 	/* handle data read from socket */
 	spa_log_trace(this->log, "read socket data %d", size_read);
+#if 0
+	hexdump_to_log(this, read_data, size_read);
+#endif
 
 	if (this->transport->codec == HFP_AUDIO_CODEC_MSBC) {
 		preprocess_and_decode_msbc_data(userdata, read_data, size_read);
 
 	} else {
 		uint8_t *packet;
+		if (size_read != 48 && is_zero_packet(read_data, size_read)) {
+			/* Adapter is returning non-standard CVSD stream. For example
+			 * Intel 8087:0029 at Firmware revision 0.0 build 191 week 21 2021
+			 * on kernel 5.13.19 produces such data.
+			 */
+			return 0;
+		}
 		packet = (uint8_t *)datas[0].data + port->ready_offset;
 		spa_memmove(packet, read_data, size_read);
 		port->ready_offset += size_read;
@@ -597,7 +622,7 @@ static int do_start(struct impl *this)
 
 	this->following = is_following(this);
 
-	spa_log_debug(this->log, NAME" %p: start following:%d",
+	spa_log_debug(this->log, "%p: start following:%d",
 			this, this->following);
 
 	/* Make sure the transport is valid */
@@ -606,7 +631,7 @@ static int do_start(struct impl *this)
 	/* Do accept if Gateway; otherwise do connect for Head Unit */
 	do_accept = this->transport->profile & SPA_BT_PROFILE_HEADSET_AUDIO_GATEWAY;
 
-	/* acquire the socked fd (false -> connect | true -> accept) */
+	/* acquire the socket fd (false -> connect | true -> accept) */
 	if ((res = spa_bt_transport_acquire(this->transport, do_accept)) < 0)
 		return res;
 
@@ -616,7 +641,7 @@ static int do_start(struct impl *this)
 	/* Init mSBC if needed */
 	if (this->transport->codec == HFP_AUDIO_CODEC_MSBC) {
 		sbc_init_msbc(&this->msbc, 0);
-		/* Libsbc expects audio samples by default in host endianity, mSBC requires little endian */
+		/* Libsbc expects audio samples by default in host endianness, mSBC requires little endian */
 		this->msbc.endian = SBC_LE;
 		this->msbc_seq_initialized = false;
 
@@ -708,21 +733,23 @@ static int impl_node_send_command(void *object, const struct spa_command *comman
 
 static void emit_node_info(struct impl *this, bool full)
 {
-	bool is_ag = this->transport && (this->transport->profile & SPA_BT_PROFILE_HEADSET_AUDIO_GATEWAY);
-	char latency[64] = "128/8000";
-	struct spa_dict_item ag_node_info_items[] = {
-		{ SPA_KEY_DEVICE_API, "bluez5" },
-		{ SPA_KEY_MEDIA_CLASS, "Stream/Output/Audio" },
-		{ SPA_KEY_NODE_LATENCY, latency },
-		{ "media.name", ((this->transport && this->transport->device->name) ?
-		                 this->transport->device->name : "HSP/HFP") },
-	};
-	struct spa_dict_item hu_node_info_items[] = {
+	static const struct spa_dict_item hu_node_info_items[] = {
 		{ SPA_KEY_DEVICE_API, "bluez5" },
 		{ SPA_KEY_MEDIA_CLASS, "Audio/Source" },
 		{ SPA_KEY_NODE_DRIVER, "true" },
 	};
 
+	char latency[64] = "128/8000";
+	const struct spa_dict_item ag_node_info_items[] = {
+		{ SPA_KEY_DEVICE_API, "bluez5" },
+		{ SPA_KEY_MEDIA_CLASS, "Stream/Output/Audio" },
+		{ SPA_KEY_NODE_LATENCY, latency },
+		{ "media.name", ((this->transport && this->transport->device->name) ?
+					this->transport->device->name : "HSP/HFP") },
+	};
+	bool is_ag = this->transport && (this->transport->profile & SPA_BT_PROFILE_HEADSET_AUDIO_GATEWAY);
+	uint64_t old = full ? this->info.change_mask : 0;
+
 	if (full)
 		this->info.change_mask = this->info_all;
 	if (this->info.change_mask) {
@@ -733,18 +760,19 @@ static void emit_node_info(struct impl *this, bool full)
 			&SPA_DICT_INIT_ARRAY(ag_node_info_items) :
 			&SPA_DICT_INIT_ARRAY(hu_node_info_items);
 		spa_node_emit_info(&this->hooks, &this->info);
-		this->info.change_mask = 0;
+		this->info.change_mask = old;
 	}
 }
 
 static void emit_port_info(struct impl *this, struct port *port, bool full)
 {
+	uint64_t old = full ? port->info.change_mask : 0;
 	if (full)
 		port->info.change_mask = port->info_all;
 	if (port->info.change_mask) {
 		spa_node_emit_port_info(&this->hooks,
 				SPA_DIRECTION_OUTPUT, 0, &port->info);
-		port->info.change_mask = 0;
+		port->info.change_mask = old;
 	}
 }
 
@@ -880,8 +908,7 @@ impl_node_port_enum_params(void *object, int seq,
 							this->props.max_latency * port->frame_size,
 							this->props.min_latency * port->frame_size,
 							INT32_MAX),
-			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(port->frame_size),
-			SPA_PARAM_BUFFERS_align,   SPA_POD_Int(16));
+			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(port->frame_size));
 		break;
 
 	case SPA_PARAM_Meta:
@@ -916,6 +943,16 @@ impl_node_port_enum_params(void *object, int seq,
 		}
 		break;
 
+	case SPA_PARAM_Latency:
+		switch (result.index) {
+		case 0:
+			param = spa_latency_build(&b, id, &port->latency);
+			break;
+		default:
+			return 0;
+		}
+		break;
+
 	default:
 		return -ENOENT;
 	}
@@ -977,11 +1014,12 @@ static int port_set_format(struct impl *this, struct port *port,
 		port->info.flags = SPA_PORT_FLAG_LIVE;
 		port->info.change_mask |= SPA_PORT_CHANGE_MASK_RATE;
 		port->info.rate = SPA_FRACTION(1, port->current_format.info.raw.rate);
-		port->params[3] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_READWRITE);
-		port->params[4] = SPA_PARAM_INFO(SPA_PARAM_Buffers, SPA_PARAM_INFO_READ);
+		port->params[IDX_Format] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_READWRITE);
+		port->params[IDX_Buffers] = SPA_PARAM_INFO(SPA_PARAM_Buffers, SPA_PARAM_INFO_READ);
+		port->params[IDX_Latency].flags ^= SPA_PARAM_INFO_SERIAL;
 	} else {
-		port->params[3] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
-		port->params[4] = SPA_PARAM_INFO(SPA_PARAM_Buffers, 0);
+		port->params[IDX_Format] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
+		port->params[IDX_Buffers] = SPA_PARAM_INFO(SPA_PARAM_Buffers, 0);
 	}
 	emit_port_info(this, port, false);
 
@@ -1007,6 +1045,9 @@ impl_node_port_set_param(void *object,
 	case SPA_PARAM_Format:
 		res = port_set_format(this, port, flags, param);
 		break;
+	case SPA_PARAM_Latency:
+		res = 0;
+		break;
 	default:
 		res = -ENOENT;
 		break;
@@ -1046,7 +1087,7 @@ impl_node_port_use_buffers(void *object,
 		b->h = spa_buffer_find_meta_data(buffers[i], SPA_META_Header, sizeof(*b->h));
 
 		if (d[0].data == NULL) {
-			spa_log_error(this->log, NAME " %p: need mapped memory", this);
+			spa_log_error(this->log, "%p: need mapped memory", this);
 			return -EINVAL;
 		}
 		spa_list_append(&port->free, &b->link);
@@ -1198,7 +1239,7 @@ static int impl_get_interface(struct spa_handle *handle, const char *type, void
 
 	this = (struct impl *) handle;
 
-	if (strcmp(type, SPA_TYPE_INTERFACE_Node) == 0)
+	if (spa_streq(type, SPA_TYPE_INTERFACE_Node))
 		*interface = &this->node;
 	else
 		return -ENOENT;
@@ -1244,6 +1285,8 @@ impl_init(const struct spa_handle_factory *factory,
 	this->data_loop = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_DataLoop);
 	this->data_system = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_DataSystem);
 
+	spa_log_topic_init(this->log, &log_topic);
+
 	if (this->data_loop == NULL) {
 		spa_log_error(this->log, "a data loop is needed");
 		return -EINVAL;
@@ -1267,11 +1310,11 @@ impl_init(const struct spa_handle_factory *factory,
 			SPA_NODE_CHANGE_MASK_PARAMS;
 	this->info = SPA_NODE_INFO_INIT();
 	this->info.flags = SPA_NODE_FLAG_RT;
-	this->params[0] = SPA_PARAM_INFO(SPA_PARAM_PropInfo, SPA_PARAM_INFO_READ);
-	this->params[1] = SPA_PARAM_INFO(SPA_PARAM_Props, SPA_PARAM_INFO_READWRITE);
-	this->params[2] = SPA_PARAM_INFO(SPA_PARAM_IO, SPA_PARAM_INFO_READ);
+	this->params[IDX_PropInfo] = SPA_PARAM_INFO(SPA_PARAM_PropInfo, SPA_PARAM_INFO_READ);
+	this->params[IDX_Props] = SPA_PARAM_INFO(SPA_PARAM_Props, SPA_PARAM_INFO_READWRITE);
+	this->params[IDX_NODE_IO] = SPA_PARAM_INFO(SPA_PARAM_IO, SPA_PARAM_INFO_READ);
 	this->info.params = this->params;
-	this->info.n_params = 3;
+	this->info.n_params = N_NODE_PARAMS;
 
 	/* set the port info */
 	port = &this->port;
@@ -1281,13 +1324,18 @@ impl_init(const struct spa_handle_factory *factory,
 	port->info.change_mask = SPA_PORT_CHANGE_MASK_FLAGS;
 	port->info.flags = SPA_PORT_FLAG_LIVE |
 			   SPA_PORT_FLAG_TERMINAL;
-	port->params[0] = SPA_PARAM_INFO(SPA_PARAM_EnumFormat, SPA_PARAM_INFO_READ);
-	port->params[1] = SPA_PARAM_INFO(SPA_PARAM_Meta, SPA_PARAM_INFO_READ);
-	port->params[2] = SPA_PARAM_INFO(SPA_PARAM_IO, SPA_PARAM_INFO_READ);
-	port->params[3] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
-	port->params[4] = SPA_PARAM_INFO(SPA_PARAM_Buffers, 0);
+	port->params[IDX_EnumFormat] = SPA_PARAM_INFO(SPA_PARAM_EnumFormat, SPA_PARAM_INFO_READ);
+	port->params[IDX_Meta] = SPA_PARAM_INFO(SPA_PARAM_Meta, SPA_PARAM_INFO_READ);
+	port->params[IDX_IO] = SPA_PARAM_INFO(SPA_PARAM_IO, SPA_PARAM_INFO_READ);
+	port->params[IDX_Format] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
+	port->params[IDX_Buffers] = SPA_PARAM_INFO(SPA_PARAM_Buffers, 0);
+	port->params[IDX_Latency] = SPA_PARAM_INFO(SPA_PARAM_Latency, SPA_PARAM_INFO_READWRITE);
 	port->info.params = port->params;
-	port->info.n_params = 5;
+	port->info.n_params = N_PORT_PARAMS;
+
+	port->latency = SPA_LATENCY_INFO(SPA_DIRECTION_OUTPUT);
+	port->latency.min_quantum = 1.0f;
+	port->latency.max_quantum = 1.0f;
 
 	/* Init the buffer lists */
 	spa_list_init(&port->ready);
@@ -1337,7 +1385,7 @@ static const struct spa_dict_item info_items[] = {
 
 static const struct spa_dict info = SPA_DICT_INIT_ARRAY(info_items);
 
-struct spa_handle_factory spa_sco_source_factory = {
+const struct spa_handle_factory spa_sco_source_factory = {
 	SPA_VERSION_HANDLE_FACTORY,
 	SPA_NAME_API_BLUEZ5_SCO_SOURCE,
 	&info,
