@@ -31,6 +31,7 @@
 #include <sys/mman.h>
 
 #include <spa/pod/parser.h>
+#include <spa/pod/dynamic.h>
 #include <spa/node/utils.h>
 #include <spa/utils/result.h>
 #include <spa/debug/types.h>
@@ -38,10 +39,13 @@
 #include "pipewire/pipewire.h"
 #include "pipewire/private.h"
 
-#include "extensions/protocol-native.h"
-#include "extensions/client-node.h"
+#include "pipewire/extensions/protocol-native.h"
+#include "pipewire/extensions/client-node.h"
 
-#define MAX_MIX	4096
+#define MAX_BUFFERS	64
+
+PW_LOG_TOPIC_EXTERN(mod_topic);
+#define PW_LOG_TOPIC_DEFAULT mod_topic
 
 /** \cond */
 static bool mlock_warned = false;
@@ -70,7 +74,6 @@ struct node_data {
 	int rtwritefd;
 	struct pw_memmap *activation;
 
-	struct mix mix_pool[MAX_MIX];
 	struct spa_list mix[2];
 	struct spa_list free_mix;
 
@@ -144,8 +147,12 @@ static void clean_transport(struct node_data *data)
 	spa_list_consume(l, &data->links, link)
 		clear_link(data, l);
 
-	while ((mm = pw_mempool_find_tag(data->pool, tag, sizeof(uint32_t))) != NULL)
+	while ((mm = pw_mempool_find_tag(data->pool, tag, sizeof(uint32_t))) != NULL) {
+		if (mm->tag[1] == SPA_ID_INVALID)
+			spa_node_set_io(data->node->node, mm->tag[2], NULL, 0);
+
 		pw_memmap_free(mm);
+	}
 
 	pw_memmap_free(data->activation);
 	data->node->rt.activation = data->node->activation->map->ptr;
@@ -156,6 +163,7 @@ static void clean_transport(struct node_data *data)
 
 static void mix_init(struct mix *mix, struct pw_impl_port *port, uint32_t mix_id)
 {
+	pw_log_debug("port %p: mix init %d.%d", port, port->port_id, mix_id);
 	mix->port = port;
 	mix->mix_id = mix_id;
 	pw_impl_port_init_mix(port, &mix->mix);
@@ -214,8 +222,11 @@ static struct mix *find_mix(struct node_data *data,
 
 	spa_list_for_each(mix, &data->mix[direction], link) {
 		if (mix->port->port_id == port_id &&
-		    mix->mix_id == mix_id)
+		    mix->mix_id == mix_id) {
+			pw_log_debug("port %p: found mix %d:%d.%d", mix->port,
+					direction, port_id, mix_id);
 			return mix;
+		}
 	}
 	return NULL;
 }
@@ -229,15 +240,17 @@ static struct mix *ensure_mix(struct node_data *data,
 	if ((mix = find_mix(data, direction, port_id, mix_id)))
 		return mix;
 
-	if (spa_list_is_empty(&data->free_mix))
-		return NULL;
-
 	port = pw_impl_node_find_port(data->node, direction, port_id);
 	if (port == NULL)
 		return NULL;
 
-	mix = spa_list_first(&data->free_mix, struct mix, link);
-	spa_list_remove(&mix->link);
+	if (spa_list_is_empty(&data->free_mix)) {
+		if ((mix = calloc(1, sizeof(*mix))) == NULL)
+			return NULL;
+	} else {
+		mix = spa_list_first(&data->free_mix, struct mix, link);
+		spa_list_remove(&mix->link);
+	}
 
 	mix_init(mix, port, mix_id);
 	spa_list_append(&data->mix[direction], &mix->link);
@@ -288,8 +301,8 @@ static int add_node_update(struct node_data *data, uint32_t change_mask, uint32_
 
 	if (change_mask & PW_CLIENT_NODE_UPDATE_PARAMS) {
 		uint32_t i, idx, id;
-		uint8_t buf[2048];
-		struct spa_pod_builder b = { 0 };
+		uint8_t buf[4096];
+		struct spa_pod_dynamic_builder b;
 
 		for (i = 0; i < node->info.n_params; i++) {
 			struct spa_pod *param;
@@ -299,14 +312,17 @@ static int add_node_update(struct node_data *data, uint32_t change_mask, uint32_
 				continue;
 
 			for (idx = 0;;) {
-				spa_pod_builder_init(&b, buf, sizeof(buf));
-	                        if (spa_node_enum_params_sync(node->node,
-							id, &idx,
-							NULL, &param, &b) != 1)
-	                                break;
+				spa_pod_dynamic_builder_init(&b, buf, sizeof(buf), 4096);
 
-				params = realloc(params, sizeof(struct spa_pod *) * (n_params + 1));
-				params[n_params++] = spa_pod_copy(param);
+	                        res = spa_node_enum_params_sync(node->node,
+							id, &idx, NULL, &param, &b.b);
+				if (res == 1) {
+					params = realloc(params, sizeof(struct spa_pod *) * (n_params + 1));
+					params[n_params++] = spa_pod_copy(param);
+				}
+				spa_pod_dynamic_builder_clean(&b);
+				if (res != 1)
+	                                break;
 			}
                 }
 	}
@@ -343,8 +359,8 @@ static int add_port_update(struct node_data *data, struct pw_impl_port *port, ui
 
 	if (change_mask & PW_CLIENT_NODE_PORT_UPDATE_PARAMS) {
 		uint32_t i, idx, id;
-		uint8_t buf[2048];
-		struct spa_pod_builder b = { 0 };
+		uint8_t buf[4096];
+		struct spa_pod_dynamic_builder b;
 
 		for (i = 0; i < port->info.n_params; i++) {
 			struct spa_pod *param;
@@ -354,15 +370,20 @@ static int add_port_update(struct node_data *data, struct pw_impl_port *port, ui
 				continue;
 
 			for (idx = 0;;) {
-				spa_pod_builder_init(&b, buf, sizeof(buf));
-	                        if (spa_node_port_enum_params_sync(port->node->node,
+				spa_pod_dynamic_builder_init(&b, buf, sizeof(buf), 4096);
+
+	                        res = spa_node_port_enum_params_sync(port->node->node,
 							port->direction, port->port_id,
-							id, &idx,
-							NULL, &param, &b) != 1)
+							id, &idx, NULL, &param, &b.b);
+				if (res == 1) {
+					params = realloc(params, sizeof(struct spa_pod *) * (n_params + 1));
+					params[n_params++] = spa_pod_copy(param);
+				}
+				spa_pod_dynamic_builder_clean(&b);
+
+				if (res != 1)
 	                                break;
 
-				params = realloc(params, sizeof(struct spa_pod *) * (n_params + 1));
-				params[n_params++] = spa_pod_copy(param);
 			}
                 }
 	}
@@ -429,7 +450,7 @@ client_node_set_io(void *object,
 	struct pw_proxy *proxy = (struct pw_proxy*)data->client_node;
 	struct pw_memmap *old, *mm;
 	void *ptr;
-	uint32_t tag[5] = { data->remote_id, id, };
+	uint32_t tag[5] = { data->remote_id, SPA_ID_INVALID, id, };
 	int res;
 
 	old = pw_mempool_find_tag(data->pool, tag, sizeof(tag));
@@ -453,8 +474,7 @@ client_node_set_io(void *object,
 
 	res =  spa_node_set_io(data->node->node, id, ptr, size);
 
-	if (old != NULL)
-		pw_memmap_free(old);
+	pw_memmap_free(old);
 exit:
 	if (res < 0) {
 		pw_log_error("node %p: set_io: %s", proxy, spa_strerror(res));
@@ -501,7 +521,9 @@ static int client_node_command(void *object, const struct spa_command *command)
 			pw_proxy_error(proxy, res, "suspend failed");
 		}
 		break;
-
+	case SPA_NODE_COMMAND_RequestProcess:
+		res = pw_impl_node_send_command(data->node, command);
+		break;
 	default:
 		pw_log_warn("unhandled node command %d", SPA_NODE_COMMAND_ID(command));
 		res = -ENOTSUP;
@@ -617,6 +639,9 @@ client_node_port_use_buffers(void *object,
 		goto error_exit;
 	}
 
+	if (n_buffers > MAX_BUFFERS)
+		return -ENOSPC;
+
 	prot = PW_MEMMAP_FLAG_READWRITE;
 
 	/* clear previous buffers */
@@ -668,8 +693,8 @@ client_node_port_use_buffers(void *object,
 		}
 		memcpy(b, buffers[i].buffer, sizeof(struct spa_buffer));
 
-		b->metas = SPA_MEMBER(b, sizeof(struct spa_buffer), struct spa_meta);
-		b->datas = SPA_MEMBER(b->metas, sizeof(struct spa_meta) * b->n_metas,
+		b->metas = SPA_PTROFF(b, sizeof(struct spa_buffer), struct spa_meta);
+		b->datas = SPA_PTROFF(b->metas, sizeof(struct spa_meta) * b->n_metas,
 				       struct spa_data);
 
 		pw_log_debug("add buffer mem:%d id:%d offset:%u size:%u %p", mm->block->id,
@@ -679,7 +704,7 @@ client_node_port_use_buffers(void *object,
 		for (j = 0; j < b->n_metas; j++) {
 			struct spa_meta *m = &b->metas[j];
 			memcpy(m, &buffers[i].buffer->metas[j], sizeof(struct spa_meta));
-			m->data = SPA_MEMBER(mm->ptr, offset, void);
+			m->data = SPA_PTROFF(mm->ptr, offset, void);
 			offset += SPA_ROUND_UP_N(m->size, 8);
 		}
 
@@ -688,7 +713,7 @@ client_node_port_use_buffers(void *object,
 
 			memcpy(d, &buffers[i].buffer->datas[j], sizeof(struct spa_data));
 			d->chunk =
-			    SPA_MEMBER(mm->ptr, offset + sizeof(struct spa_chunk) * j,
+			    SPA_PTROFF(mm->ptr, offset + sizeof(struct spa_chunk) * j,
 				       struct spa_chunk);
 
 			if (flags & SPA_NODE_BUFFERS_FLAG_ALLOC)
@@ -713,7 +738,7 @@ client_node_port_use_buffers(void *object,
 						j, bm->id, bm->fd, d->maxsize);
 			} else if (d->type == SPA_DATA_MemPtr) {
 				int offs = SPA_PTR_TO_INT(d->data);
-				d->data = SPA_MEMBER(mm->ptr, offs, void);
+				d->data = SPA_PTROFF(mm->ptr, offs, void);
 				d->fd = -1;
 				pw_log_debug(" data %d id:%u -> mem:%p offs:%d maxsize:%d",
 						j, bid->id, d->data, offs, d->maxsize);
@@ -793,7 +818,7 @@ client_node_port_set_io(void *object,
 	}
 
 	if ((res = spa_node_port_set_io(mix->port->mix,
-			     direction, mix_id, id, ptr, size)) < 0) {
+			     direction, mix->mix.port.port_id, id, ptr, size)) < 0) {
 		if (res == -ENOTSUP)
 			res = 0;
 		else
@@ -805,8 +830,7 @@ client_node_port_set_io(void *object,
 			activate_mix(data, mix);
 	}
 exit_free:
-	if (old != NULL)
-		pw_memmap_free(old);
+	pw_memmap_free(old);
 exit:
 	if (res < 0) {
 		pw_log_error("port %p: set_io: %s", mix, spa_strerror(res));
@@ -962,6 +986,8 @@ static void do_node_init(struct node_data *data)
 
 static void clear_mix(struct node_data *data, struct mix *mix)
 {
+	pw_log_debug("port %p: mix clear %d.%d", mix->port, mix->port->port_id, mix->mix_id);
+
 	deactivate_mix(data, mix);
 
 	spa_list_remove(&mix->link);
@@ -969,21 +995,24 @@ static void clear_mix(struct node_data *data, struct mix *mix)
 	clear_buffers(data, mix);
 	pw_array_clear(&mix->buffers);
 
-	spa_list_remove(&mix->mix.link);
 	spa_list_append(&data->free_mix, &mix->link);
 	pw_impl_port_release_mix(mix->port, &mix->mix);
 }
 
 static void clean_node(struct node_data *d)
 {
-	struct mix *mix, *tmp;
+	struct mix *mix;
 
 	if (d->have_transport) {
-		spa_list_for_each_safe(mix, tmp, &d->mix[SPA_DIRECTION_INPUT], link)
+		spa_list_consume(mix, &d->mix[SPA_DIRECTION_INPUT], link)
 			clear_mix(d, mix);
-		spa_list_for_each_safe(mix, tmp, &d->mix[SPA_DIRECTION_OUTPUT], link)
+		spa_list_consume(mix, &d->mix[SPA_DIRECTION_OUTPUT], link)
 			clear_mix(d, mix);
 	}
+	spa_list_consume(mix, &d->free_mix, link) {
+		spa_list_remove(&mix->link);
+		free(mix);
+	}
 	clean_transport(d);
 }
 
@@ -1077,6 +1106,16 @@ static void node_active_changed(void *data, bool active)
 	pw_client_node_set_active(d->client_node, active);
 }
 
+static void node_event(void *data, const struct spa_event *event)
+{
+	struct node_data *d = data;
+	pw_log_debug("%p", d);
+
+	if (d->client_node == NULL)
+		return;
+	pw_client_node_event(d->client_node, event);
+}
+
 static const struct pw_impl_node_events node_events = {
 	PW_VERSION_IMPL_NODE_EVENTS,
 	.destroy = node_destroy,
@@ -1085,6 +1124,7 @@ static const struct pw_impl_node_events node_events = {
 	.port_info_changed = node_port_info_changed,
 	.port_removed = node_port_removed,
 	.active_changed = node_active_changed,
+	.event = node_event,
 };
 
 static void client_node_removed(void *_data)
@@ -1093,6 +1133,7 @@ static void client_node_removed(void *_data)
 	pw_log_debug("%p: removed", data);
 
 	spa_hook_remove(&data->proxy_client_node_listener);
+	spa_hook_remove(&data->client_node_listener);
 
 	if (data->node) {
 		spa_hook_remove(&data->node_listener);
@@ -1192,8 +1233,6 @@ static struct pw_proxy *node_export(struct pw_core *core, void *object, bool do_
 	struct pw_impl_node *node = object;
 	struct pw_proxy *client_node;
 	struct node_data *data;
-	const char *str;
-	int i;
 
 	user_data_size = SPA_ROUND_UP_N(user_data_size, __alignof__(struct node_data));
 
@@ -1207,7 +1246,7 @@ static struct pw_proxy *node_export(struct pw_core *core, void *object, bool do_
 		goto error;
 
 	data = pw_proxy_get_user_data(client_node);
-	data = SPA_MEMBER(data, user_data_size, struct node_data);
+	data = SPA_PTROFF(data, user_data_size, struct node_data);
 	data->pool = pw_core_get_mempool(core);
 	data->node = node;
 	data->do_free = do_free;
@@ -1215,21 +1254,18 @@ static struct pw_proxy *node_export(struct pw_core *core, void *object, bool do_
 	data->client_node = (struct pw_client_node *)client_node;
 	data->remote_id = SPA_ID_INVALID;
 
-	data->allow_mlock = data->context->defaults.mem_allow_mlock;
-	if ((str = pw_properties_get(node->properties, "mem.allow-mlock")) != NULL)
-		data->allow_mlock = pw_properties_parse_bool(str);
 
-	data->warn_mlock = data->context->defaults.mem_warn_mlock;
-	if ((str = pw_properties_get(node->properties, "mem.warn-mlock")) != NULL)
-		data->warn_mlock = pw_properties_parse_bool(str);
+	data->allow_mlock = pw_properties_get_bool(node->properties, "mem.allow-mlock",
+						   data->context->settings.mem_allow_mlock);
+
+	data->warn_mlock = pw_properties_get_bool(node->properties, "mem.warn-mlock",
+						  data->context->settings.mem_warn_mlock);
 
 	node->exported = true;
 
 	spa_list_init(&data->free_mix);
 	spa_list_init(&data->mix[0]);
 	spa_list_init(&data->mix[1]);
-	for (i = 0; i < MAX_MIX; i++)
-		spa_list_append(&data->free_mix, &data->mix_pool[i].link);
 
 	spa_list_init(&data->links);
 
@@ -1271,6 +1307,11 @@ struct pw_proxy *pw_core_spa_node_export(struct pw_core *core,
 {
 	struct pw_impl_node *node;
 	struct pw_proxy *proxy;
+	const char *str;
+	bool do_register;
+
+	str = props ? spa_dict_lookup(props, PW_KEY_OBJECT_REGISTER) : NULL;
+	do_register = str ? pw_properties_parse_bool(str) : true;
 
 	node = pw_context_create_node(pw_core_get_context(core),
 			props ? pw_properties_new_dict(props) : NULL, 0);
@@ -1278,7 +1319,9 @@ struct pw_proxy *pw_core_spa_node_export(struct pw_core *core,
 		return NULL;
 
 	pw_impl_node_set_implementation(node, (struct spa_node*)object);
-	pw_impl_node_register(node, NULL);
+
+	if (do_register)
+		pw_impl_node_register(node, NULL);
 
 	proxy = node_export(core, node, true, user_data_size);
 	if (proxy)
