@@ -29,12 +29,13 @@
 #include <unistd.h>
 #include <errno.h>
 #include <sys/socket.h>
+#include <sys/types.h>
 #include <sys/un.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <sys/file.h>
 #include <ctype.h>
-#if HAVE_PWD_H
+#ifdef HAVE_PWD_H
 #include <pwd.h>
 #endif
 #if defined(__FreeBSD__)
@@ -43,26 +44,35 @@
 
 #include <spa/pod/iter.h>
 #include <spa/utils/result.h>
+#include <spa/utils/string.h>
 
 #ifdef HAVE_SYSTEMD
 #include <systemd/sd-daemon.h>
 #endif
 
 #include <pipewire/impl.h>
-#include <extensions/protocol-native.h>
+#include <pipewire/extensions/protocol-native.h>
 
 #include "pipewire/private.h"
 
 #include "modules/module-protocol-native/connection.h"
 #include "modules/module-protocol-native/defs.h"
+#include "modules/module-protocol-native/protocol-footer.h"
+
+
+#define NAME "protocol-native"
+PW_LOG_TOPIC(mod_topic, "mod." NAME);
+#define PW_LOG_TOPIC_DEFAULT mod_topic
+PW_LOG_TOPIC(mod_topic_connection, "conn." NAME);
 
 #undef spa_debug
-#define spa_debug pw_log_debug
+#define spa_debug(...) pw_logt_debug(mod_topic_connection, __VA_ARGS__)
 
 #include <spa/debug/pod.h>
 #include <spa/debug/types.h>
 
-#define NAME "protocol-native"
+/** \page page_module_protocol_native PipeWire Module: Protocol Native
+ */
 
 #ifndef UNIX_PATH_MAX
 #define UNIX_PATH_MAX   108
@@ -106,6 +116,9 @@ struct client {
 
 	int ref;
 
+	struct footer_core_global_state footer_state;
+
+	unsigned int connected:1;
 	unsigned int disconnecting:1;
 	unsigned int need_flush:1;
 	unsigned int paused:1;
@@ -141,6 +154,8 @@ struct client_data {
 	struct pw_protocol_native_connection *connection;
 	struct spa_hook conn_listener;
 
+	struct footer_client_global_state footer_state;
+
 	unsigned int busy:1;
 	unsigned int need_flush:1;
 
@@ -150,10 +165,11 @@ struct client_data {
 static void debug_msg(const char *prefix, const struct pw_protocol_native_message *msg, bool hex)
 {
 	struct spa_pod *pod;
-	pw_log_debug("%s: id:%d op:%d size:%d seq:%d", prefix,
-			msg->id, msg->opcode, msg->size, msg->seq);
+	pw_logt_debug(mod_topic_connection,
+		      "%s: id:%d op:%d size:%d seq:%d", prefix,
+		      msg->id, msg->opcode, msg->size, msg->seq);
 
-	if ((pod = spa_pod_from_data(msg->data, msg->size, 0, msg->size)) != NULL)
+	if ((pod = get_first_pod_from_data(msg->data, msg->size, 0)) != NULL)
 		spa_debug_pod(0, NULL, pod);
 	else
 		hex = true;
@@ -161,6 +177,49 @@ static void debug_msg(const char *prefix, const struct pw_protocol_native_messag
 		spa_debug_mem(0, msg->data, msg->size);
 }
 
+static void pre_demarshal(struct pw_protocol_native_connection *conn,
+		const struct pw_protocol_native_message *msg,
+		void *object, const struct footer_demarshal *opcodes, size_t n_opcodes)
+{
+	struct spa_pod *footer = NULL;
+	struct spa_pod_parser parser;
+	struct spa_pod_frame f[2];
+	uint32_t opcode;
+	int ret;
+
+	footer = pw_protocol_native_connection_get_footer(conn, msg);
+	if (footer == NULL)
+		return;   /* No valid footer. Ignore silently. */
+
+	/*
+	 * Version 3 footer
+	 *
+	 * spa_pod Struct { [Id opcode, Struct { ... }]* }
+	 */
+
+	spa_pod_parser_pod(&parser, footer);
+	if (spa_pod_parser_push_struct(&parser, &f[0]) < 0) {
+		pw_log_error("malformed message footer");
+		return;
+	}
+
+	while (1) {
+		if (spa_pod_parser_get_id(&parser, &opcode) < 0)
+			break;
+		if (spa_pod_parser_push_struct(&parser, &f[1]) < 0)
+			break;
+		if (opcode < n_opcodes) {
+			if ((ret = opcodes[opcode].demarshal(object, &parser)) < 0)
+				pw_log_error("failed processing message footer (opcode %u): %d (%s)",
+						opcode, ret, spa_strerror(ret));
+		} else {
+			/* Ignore (don't log errors), in case we need to extend this later. */
+			pw_log_debug("unknown message footer opcode %u", opcode);
+		}
+		spa_pod_parser_pop(&parser, &f[1]);
+	}
+}
+
 static int
 process_messages(struct client_data *data)
 {
@@ -196,12 +255,15 @@ process_messages(struct client_data *data)
 
 		client->recv_seq = msg->seq;
 
-		pw_log_trace(NAME" %p: got message %d from %u", client->protocol,
+		pw_log_trace("%p: got message %d from %u", client->protocol,
 			     msg->opcode, msg->id);
 
 		if (debug_messages)
 			debug_msg("<<<<<< in", msg, false);
 
+		pre_demarshal(conn, msg, client, footer_client_demarshal,
+				SPA_N_ELEMENTS(footer_client_demarshal));
+
 		resource = pw_impl_client_find_resource(client, msg->id);
 		if (resource == NULL) {
 			pw_resource_errorf(client->core_resource,
@@ -211,14 +273,18 @@ process_messages(struct client_data *data)
 
 		marshal = pw_resource_get_marshal(resource);
 		if (marshal == NULL || msg->opcode >= marshal->n_client_methods) {
-			res = -ENOSYS;
-			goto invalid_method;
+			pw_resource_errorf_id(resource, msg->id,
+					-ENOSYS, "invalid method id:%u op:%u",
+					msg->id, msg->opcode);
+			continue;
 		}
 
 		demarshal = marshal->server_demarshal;
 		if (!demarshal[msg->opcode].func) {
-			res = -ENOTSUP;
-			goto invalid_message;
+			pw_resource_errorf_id(resource, msg->id,
+					-ENOTSUP, "function not supported id:%u op:%u",
+					msg->id, msg->opcode);
+			continue;
 		}
 
 		permissions = pw_resource_get_permissions(resource);
@@ -226,34 +292,30 @@ process_messages(struct client_data *data)
 
 		if ((required & permissions) != required) {
 			pw_resource_errorf_id(resource, msg->id,
-				-EACCES, "no permission to call method %u on %u (requires %08x, have %08x)",
-				msg->opcode, msg->id, required, permissions);
+				-EACCES, "no permission to call method %u on %u "
+				"(requires "PW_PERMISSION_FORMAT", have "PW_PERMISSION_FORMAT")",
+				msg->opcode, msg->id,
+				PW_PERMISSION_ARGS(required), PW_PERMISSION_ARGS(permissions));
 			continue;
 		}
 
 		pw_protocol_native_connection_enter(conn);
 		res = demarshal[msg->opcode].func(resource, msg);
 		pw_protocol_native_connection_leave(conn);
-		if (res < 0)
-			goto invalid_message;
+		if (res < 0) {
+			pw_resource_errorf_id(resource, msg->id,
+					res, "invalid message id:%u op:%u (%s)",
+					msg->id, msg->opcode, spa_strerror(res));
+			debug_msg("*invalid message*", msg, true);
+		}
 	}
 	res = 0;
 done:
 	context->current_client = NULL;
 	return res;
 
-invalid_method:
-	pw_resource_errorf_id(resource, msg->id, res, "invalid method id:%u op:%u",
-			msg->id, msg->opcode);
-	goto done;
-invalid_message:
-	pw_resource_errorf_id(resource, msg->id, res, "invalid message id:%u op:%u (%s)",
-			msg->id, msg->opcode, spa_strerror(res));
-	debug_msg("*invalid message*", msg, true);
-	goto done;
 error:
-	if (client->core_resource)
-		pw_resource_errorf(client->core_resource, res, "client error %d (%s)",
+	pw_resource_errorf(client->core_resource, res, "client error %d (%s)",
 				res, spa_strerror(res));
 	goto done;
 }
@@ -270,7 +332,7 @@ client_busy_changed(void *data, bool busy)
 
 	SPA_FLAG_UPDATE(mask, SPA_IO_IN, !busy);
 
-	pw_log_debug(NAME" %p: busy changed %d", client->protocol, busy);
+	pw_log_debug("%p: busy changed %d", client->protocol, busy);
 	pw_loop_update_io(client->context->main_loop, c->source, mask);
 
 	if (!busy)
@@ -279,10 +341,10 @@ client_busy_changed(void *data, bool busy)
 
 static void handle_client_error(struct pw_impl_client *client, int res)
 {
-	if (res == -EPIPE)
-		pw_log_info(NAME" %p: client %p disconnected", client->protocol, client);
+	if (res == -EPIPE || res == -ECONNRESET)
+		pw_log_info("%p: client %p disconnected", client->protocol, client);
 	else
-		pw_log_error(NAME" %p: client %p error %d (%s)", client->protocol,
+		pw_log_error("%p: client %p error %d (%s)", client->protocol,
 				client, res, spa_strerror(res));
 	pw_impl_client_destroy(client);
 }
@@ -325,7 +387,7 @@ static void client_free(void *data)
 	struct client_data *this = data;
 	struct pw_impl_client *client = this->client;
 
-	pw_log_debug(NAME" %p: free", this);
+	pw_log_debug("%p: free", this);
 	spa_list_remove(&this->protocol_link);
 
 	spa_hook_remove(&this->client_listener);
@@ -408,9 +470,10 @@ static struct client_data *client_new(struct server *s, int fd)
 	struct pw_impl_client *client;
 	struct pw_protocol *protocol = s->this.protocol;
 	socklen_t len;
-	struct ucred ucred;
 #if defined(__FreeBSD__)
 	struct xucred xucred;
+#else
+	struct ucred ucred;
 #endif
 	struct pw_context *context = protocol->context;
 	struct pw_properties *props;
@@ -466,6 +529,8 @@ static struct client_data *client_new(struct server *s, int fd)
 #endif
 		pw_properties_setf(props, PW_KEY_SEC_UID, "%d", xucred.cr_uid);
 		pw_properties_setf(props, PW_KEY_SEC_GID, "%d", xucred.cr_gid);
+		// this is what Linuxulator does at the moment, see sys/compat/linux/linux_socket.c
+		pw_properties_set(props, PW_KEY_SEC_LABEL, "unconfined");
 	}
 #endif
 
@@ -529,16 +594,8 @@ get_runtime_dir(void)
 	runtime_dir = getenv("PIPEWIRE_RUNTIME_DIR");
 	if (runtime_dir == NULL)
 		runtime_dir = getenv("XDG_RUNTIME_DIR");
-	if (runtime_dir == NULL)
-		runtime_dir = getenv("HOME");
 	if (runtime_dir == NULL)
 		runtime_dir = getenv("USERPROFILE");
-	if (runtime_dir == NULL) {
-		struct passwd pwd, *result = NULL;
-		char buffer[4096];
-		if (getpwuid_r(getuid(), &pwd, buffer, sizeof(buffer), &result) == 0)
-			runtime_dir = result ? result->pw_dir : NULL;
-	}
 	return runtime_dir;
 }
 
@@ -556,8 +613,8 @@ static int init_socket_name(struct server *s, const char *name)
 	pw_log_debug("name:%s runtime_dir:%s", name, runtime_dir);
 
 	if (runtime_dir == NULL && !path_is_absolute) {
-		pw_log_error("server %p: name %s is not an absolute path and no runtime dir found."
-				"set one of PIPEWIRE_RUNTIME_DIR, XDG_RUNTIME_DIR, HOME or "
+		pw_log_error("server %p: name %s is not an absolute path and no runtime dir found. "
+				"Set one of PIPEWIRE_RUNTIME_DIR, XDG_RUNTIME_DIR or "
 				"USERPROFILE in the environment", s, name);
 		return -ENOENT;
 	}
@@ -753,7 +810,7 @@ process_remote(struct client *impl)
 		if (res == 0)
 			break;
 
-		pw_log_trace(NAME" %p: got message %d from %u seq:%d",
+		pw_log_trace("%p: got message %d from %u seq:%d",
 			this, msg->opcode, msg->id, msg->seq);
 
 		this->recv_seq = msg->seq;
@@ -761,12 +818,15 @@ process_remote(struct client *impl)
 		if (debug_messages)
 			debug_msg("<<<<<< in", msg, false);
 
+		pre_demarshal(conn, msg, this, footer_core_demarshal,
+				SPA_N_ELEMENTS(footer_core_demarshal));
+
 		proxy = pw_core_find_proxy(this, msg->id);
 		if (proxy == NULL || proxy->zombie) {
 			if (proxy == NULL)
-				pw_log_error(NAME" %p: could not find proxy %u", this, msg->id);
+				pw_log_error("%p: could not find proxy %u", this, msg->id);
 			else
-				pw_log_debug(NAME" %p: zombie proxy %u", this, msg->id);
+				pw_log_debug("%p: zombie proxy %u", this, msg->id);
 
 			/* FIXME close fds */
 			continue;
@@ -774,7 +834,7 @@ process_remote(struct client *impl)
 
 		marshal = pw_proxy_get_marshal(proxy);
 		if (marshal == NULL || msg->opcode >= marshal->n_server_methods) {
-			pw_log_error(NAME" %p: invalid method %u for %u (%d)",
+			pw_log_error("%p: invalid method %u for %u (%d)",
 					this, msg->opcode, msg->id,
 					marshal ? marshal->n_server_methods : (uint32_t)-1);
 			continue;
@@ -782,7 +842,7 @@ process_remote(struct client *impl)
 
 		demarshal = marshal->client_demarshal;
 		if (!demarshal[msg->opcode].func) {
-                               pw_log_error(NAME" %p: function %d not implemented on %u",
+                               pw_log_error("%p: function %d not implemented on %u",
 					this, msg->opcode, msg->id);
 			continue;
 		}
@@ -793,7 +853,7 @@ process_remote(struct client *impl)
 		pw_proxy_unref(proxy);
 
 		if (res < 0) {
-			pw_log_error(NAME" %p: invalid message received %u for %u: %s",
+			pw_log_error("%p: invalid message received %u for %u: %s",
 					this, msg->opcode, msg->id, spa_strerror(res));
 			debug_msg("*invalid*", msg, true);
 		}
@@ -826,6 +886,21 @@ on_remote_data(void *data, int fd, uint32_t mask)
 			goto error;
 	}
 	if (mask & SPA_IO_OUT || impl->need_flush) {
+		if (!impl->connected) {
+			socklen_t len = sizeof res;
+
+			if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &res, &len) < 0) {
+				res = -errno;
+				pw_log_error("getsockopt: %m");
+				goto error;
+			}
+			if (res != 0) {
+				res = -res;
+				goto error;
+			}
+			impl->connected = true;
+			pw_log_debug("%p: connected, fd %d", impl, fd);
+		}
 		impl->need_flush = false;
 		res = pw_protocol_native_connection_flush(conn);
 		if (res >= 0) {
@@ -840,7 +915,7 @@ done:
 	pw_proxy_unref(core_proxy);
 	return;
 error:
-	pw_log_debug(NAME" %p: got connection error %d (%s)", impl, res, spa_strerror(res));
+	pw_log_debug("%p: got connection error %d (%s)", impl, res, spa_strerror(res));
 	if (impl->source) {
 		pw_loop_destroy_source(loop, impl->source);
 		impl->source = NULL;
@@ -881,6 +956,7 @@ static int impl_connect_fd(struct pw_protocol_client *client, int fd, bool do_cl
 	struct client *impl = SPA_CONTAINER_OF(client, struct client, this);
 	int res;
 
+	impl->connected = false;
 	impl->disconnecting = false;
 
 	pw_protocol_native_connection_set_fd(impl->connection, fd);
@@ -946,7 +1022,7 @@ static int impl_set_paused(struct pw_protocol_client *client, bool paused)
 
 	SPA_FLAG_UPDATE(mask, SPA_IO_IN, !paused);
 
-	pw_log_debug(NAME" %p: paused %d", client->protocol, paused);
+	pw_log_debug("%p: paused %d", client->protocol, paused);
 	pw_loop_update_io(impl->context->main_loop, impl->source, mask);
 
 	return paused ? 0 : process_remote(impl);
@@ -1007,7 +1083,7 @@ impl_new_client(struct pw_protocol *protocol,
 	if ((impl = calloc(1, sizeof(struct client))) == NULL)
 		return NULL;
 
-	pw_log_debug(NAME" %p: new client %p", protocol, impl);
+	pw_log_debug("%p: new client %p", protocol, impl);
 
 	this = &impl->this;
 	this->protocol = protocol;
@@ -1025,17 +1101,17 @@ impl_new_client(struct pw_protocol *protocol,
 		str = spa_dict_lookup(props, PW_KEY_REMOTE_INTENTION);
 		if (str == NULL &&
 		   (str = spa_dict_lookup(props, PW_KEY_REMOTE_NAME)) != NULL &&
-		    strcmp(str, "internal") == 0)
+		    spa_streq(str, "internal"))
 			str = "internal";
 	}
 	if (str == NULL)
 		str = "generic";
 
-	pw_log_debug(NAME" %p: connect %s", protocol, str);
+	pw_log_debug("%p: connect %s", protocol, str);
 
-	if (!strcmp(str, "screencast"))
+	if (spa_streq(str, "screencast"))
 		this->connect = pw_protocol_native_connect_portal_screencast;
-	else if (!strcmp(str, "internal"))
+	else if (spa_streq(str, "internal"))
 		this->connect = pw_protocol_native_connect_internal;
 	else
 		this->connect = pw_protocol_native_connect_local_socket;
@@ -1061,7 +1137,7 @@ static void destroy_server(struct pw_protocol_server *server)
 	struct server *s = SPA_CONTAINER_OF(server, struct server, this);
 	struct client_data *data, *tmp;
 
-	pw_log_debug(NAME" %p: server %p", s->this.protocol, s);
+	pw_log_debug("%p: server %p", s->this.protocol, s);
 
 	spa_list_remove(&server->link);
 
@@ -1134,7 +1210,7 @@ create_server(struct pw_protocol *protocol,
 
 	spa_list_append(&protocol->server_list, &this->link);
 
-	pw_log_debug(NAME" %p: created server %p", protocol, this);
+	pw_log_debug("%p: created server %p", protocol, this);
 
 	return s;
 }
@@ -1168,7 +1244,7 @@ impl_add_server(struct pw_protocol *protocol,
 	if ((s->resume = pw_loop_add_event(s->loop, do_resume, s)) == NULL)
 		goto error;
 
-	pw_log_info(NAME" %p: Listening on '%s'", protocol, name);
+	pw_log_info("%p: Listening on '%s'", protocol, name);
 
 	return this;
 
@@ -1178,7 +1254,7 @@ error:
 	return NULL;
 }
 
-static const struct pw_protocol_implementaton protocol_impl = {
+static const struct pw_protocol_implementation protocol_impl = {
 	PW_VERSION_PROTOCOL_IMPLEMENTATION,
 	.new_client = impl_new_client,
 	.add_server = impl_add_server,
@@ -1203,11 +1279,25 @@ static int impl_ext_get_proxy_fd(struct pw_proxy *proxy, uint32_t index)
 	return pw_protocol_native_connection_get_fd(impl->connection, index);
 }
 
+static void assert_single_pod(struct spa_pod_builder *builder)
+{
+	/*
+	 * Check the invariant that the message we just marshaled
+	 * consists of at most one POD.
+	 */
+	struct spa_pod *pod = builder->data;
+	spa_assert(builder->data == NULL ||
+			builder->state.offset < sizeof(struct spa_pod) ||
+			builder->state.offset == SPA_POD_SIZE(pod));
+}
+
 static int impl_ext_end_proxy(struct pw_proxy *proxy,
 			       struct spa_pod_builder *builder)
 {
 	struct pw_core *core = proxy->core;
 	struct client *impl = SPA_CONTAINER_OF(core->conn, struct client, this);
+	assert_single_pod(builder);
+	marshal_core_footers(&impl->footer_state, core, builder);
 	return core->send_seq = pw_protocol_native_connection_end(impl->connection, builder);
 }
 
@@ -1235,6 +1325,8 @@ static int impl_ext_end_resource(struct pw_resource *resource,
 {
 	struct client_data *data = resource->client->user_data;
 	struct pw_impl_client *client = resource->client;
+	assert_single_pod(builder);
+	marshal_client_footers(&data->footer_state, client, builder);
 	return client->send_seq = pw_protocol_native_connection_end(data->connection, builder);
 }
 static const struct pw_protocol_native_ext protocol_ext_impl = {
@@ -1285,6 +1377,9 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	const struct pw_properties *props;
 	int res;
 
+	PW_LOG_TOPIC_INIT(mod_topic);
+	PW_LOG_TOPIC_INIT(mod_topic_connection);
+
 	if (pw_context_find_protocol(context, PW_TYPE_INFO_PROTOCOL_Native) != NULL)
 		return 0;
 
@@ -1292,7 +1387,7 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	if (this == NULL)
 		return -errno;
 
-	debug_messages = pw_debug_is_category_enabled("connection");
+	debug_messages = mod_topic_connection->level >= SPA_LOG_LEVEL_DEBUG;
 
 	this->implementation = &protocol_impl;
 	this->extension = &protocol_ext_impl;
@@ -1300,7 +1395,7 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	pw_protocol_native_init(this);
 	pw_protocol_native0_init(this);
 
-	pw_log_debug(NAME" %p: new debug:%d", this, debug_messages);
+	pw_log_debug("%p: new debug:%d", this, debug_messages);
 
 	d = pw_protocol_get_user_data(this);
 	d->protocol = this;
