@@ -26,13 +26,21 @@
 
 #include <spa/pod/iter.h>
 #include <spa/pod/parser.h>
-#include <extensions/metadata.h>
+#include <spa/utils/result.h>
+#include <spa/utils/string.h>
+#include <pipewire/extensions/metadata.h>
+
+#include "log.h"
+#include "module-protocol-pulse/server.h"
+
+#define MAX_PARAMS 32
 
 #define manager_emit_sync(m) spa_hook_list_call(&m->hooks, struct pw_manager_events, sync, 0)
 #define manager_emit_added(m,o) spa_hook_list_call(&m->hooks, struct pw_manager_events, added, 0, o)
 #define manager_emit_updated(m,o) spa_hook_list_call(&m->hooks, struct pw_manager_events, updated, 0, o)
 #define manager_emit_removed(m,o) spa_hook_list_call(&m->hooks, struct pw_manager_events, removed, 0, o)
 #define manager_emit_metadata(m,o,s,k,t,v) spa_hook_list_call(&m->hooks, struct pw_manager_events, metadata,0,o,s,k,t,v)
+#define manager_emit_disconnect(m) spa_hook_list_call(&m->hooks, struct pw_manager_events, disconnect, 0)
 
 struct object;
 
@@ -56,7 +64,7 @@ struct object_info {
 
 struct object_data {
 	struct spa_list link;
-	const char *id;
+	const char *key;
 	size_t size;
 };
 
@@ -72,6 +80,8 @@ struct object {
 	struct spa_hook proxy_listener;
 	struct spa_hook object_listener;
 
+	int param_seq[MAX_PARAMS];
+
 	struct spa_list data_list;
 };
 
@@ -97,7 +107,8 @@ static uint32_t clear_params(struct spa_list *param_list, uint32_t id)
 	return count;
 }
 
-static struct pw_manager_param *add_param(struct spa_list *params, uint32_t id, const struct spa_pod *param)
+static struct pw_manager_param *add_param(struct spa_list *params,
+		int seq, int *param_seq, uint32_t id, const struct spa_pod *param)
 {
 	struct pw_manager_param *p;
 
@@ -109,13 +120,26 @@ static struct pw_manager_param *add_param(struct spa_list *params, uint32_t id,
 		id = SPA_POD_OBJECT_ID(param);
 	}
 
+	if (id >= MAX_PARAMS) {
+		pw_log_error("too big param id %d", id);
+		errno = EINVAL;
+		return NULL;
+	}
+
+	if (seq != param_seq[id]) {
+		pw_log_debug("ignoring param %d, seq:%d != current_seq:%d",
+				id, seq, param_seq[id]);
+		errno = EBUSY;
+		return NULL;
+	}
+
 	p = malloc(sizeof(*p) + (param != NULL ? SPA_POD_SIZE(param) : 0));
 	if (p == NULL)
 		return NULL;
 
 	p->id = id;
 	if (param != NULL) {
-		p->param = SPA_MEMBER(p, sizeof(*p), struct spa_pod);
+		p->param = SPA_PTROFF(p, sizeof(*p), struct spa_pod);
 		memcpy(p->param, param, SPA_POD_SIZE(param));
 	} else {
 		clear_params(params, id);
@@ -139,12 +163,10 @@ static bool has_param(struct spa_list *param_list, struct pw_manager_param *p)
 }
 
 
-static struct object *find_object(struct manager *m, uint32_t id)
+static struct object *find_object_by_id(struct manager *m, uint32_t id)
 {
 	struct object *o;
 	spa_list_for_each(o, &m->this.object_list, this.link) {
-		if (o->this.creating)
-			continue;
 		if (o->this.id == id)
 			return o;
 	}
@@ -174,8 +196,7 @@ static void object_destroy(struct object *o)
 	m->this.n_objects--;
 	if (o->this.proxy)
 		pw_proxy_destroy(o->this.proxy);
-	if (o->this.props)
-		pw_properties_free(o->this.props);
+	pw_properties_free(o->this.props);
 	if (o->this.message_object_path)
 		free(o->this.message_object_path);
 	clear_params(&o->this.param_list, SPA_ID_INVALID);
@@ -199,9 +220,9 @@ static void client_event_info(void *object, const struct pw_client_info *info)
 	struct object *o = object;
 	int changed = 0;
 
-        pw_log_debug("object %p: id:%d change-mask:%08"PRIx64, o, o->this.id, info->change_mask);
+	pw_log_debug("object %p: id:%d change-mask:%08"PRIx64, o, o->this.id, info->change_mask);
 
-        info = o->this.info = pw_client_info_update(o->this.info, info);
+	info = o->this.info = pw_client_info_merge(o->this.info, info, o->this.changed == 0);
 
 	if (info->change_mask & PW_CLIENT_CHANGE_MASK_PROPS)
 		changed++;
@@ -235,12 +256,12 @@ static const struct object_info client_info = {
 /* module */
 static void module_event_info(void *object, const struct pw_module_info *info)
 {
-        struct object *o = object;
+	struct object *o = object;
 	int changed = 0;
 
-        pw_log_debug("object %p: id:%d change-mask:%08"PRIx64, o, o->this.id, info->change_mask);
+	pw_log_debug("object %p: id:%d change-mask:%08"PRIx64, o, o->this.id, info->change_mask);
 
-        info = o->this.info = pw_module_info_update(o->this.info, info);
+	info = o->this.info = pw_module_info_merge(o->this.info, info, o->this.changed == 0);
 
 	if (info->change_mask & PW_MODULE_CHANGE_MASK_PROPS)
 		changed++;
@@ -279,7 +300,7 @@ static void device_event_info(void *object, const struct pw_device_info *info)
 
 	pw_log_debug("object %p: id:%d change-mask:%08"PRIx64, o, o->this.id, info->change_mask);
 
-	info = o->this.info = pw_device_info_update(o->this.info, info);
+	info = o->this.info = pw_device_info_merge(o->this.info, info, o->this.changed == 0);
 
 	if (info->change_mask & PW_DEVICE_CHANGE_MASK_PROPS)
 		changed++;
@@ -287,11 +308,17 @@ static void device_event_info(void *object, const struct pw_device_info *info)
 	if (info->change_mask & PW_DEVICE_CHANGE_MASK_PARAMS) {
 		for (i = 0; i < info->n_params; i++) {
 			uint32_t id = info->params[i].id;
+			int res;
 
 			if (info->params[i].user == 0)
 				continue;
 			info->params[i].user = 0;
 
+			if (id >= MAX_PARAMS) {
+				pw_log_error("too big param id %d", id);
+				continue;
+			}
+
 			switch (id) {
 			case SPA_PARAM_EnumProfile:
 			case SPA_PARAM_Profile:
@@ -301,12 +328,14 @@ static void device_event_info(void *object, const struct pw_device_info *info)
 			case SPA_PARAM_Route:
 				break;
 			}
-			add_param(&o->pending_list, id, NULL);
+			add_param(&o->pending_list, o->param_seq[id], o->param_seq, id, NULL);
 			if (!(info->params[i].flags & SPA_PARAM_INFO_READ))
 				continue;
 
-			pw_device_enum_params((struct pw_device*)o->this.proxy,
-					0, id, 0, -1, NULL);
+			res = pw_device_enum_params((struct pw_device*)o->this.proxy,
+					++o->param_seq[id], id, 0, -1, NULL);
+			if (SPA_RESULT_IS_ASYNC(res))
+				o->param_seq[id] = res;
 		}
 	}
 	if (changed) {
@@ -322,7 +351,7 @@ static struct object *find_device(struct manager *m, uint32_t card_id, uint32_t
 		struct pw_node_info *info;
 		const char *str;
 
-		if (strcmp(o->this.type, PW_TYPE_INTERFACE_Node) != 0)
+		if (!spa_streq(o->this.type, PW_TYPE_INTERFACE_Node))
 			continue;
 
 		if ((info = o->this.info) != NULL &&
@@ -343,13 +372,15 @@ static void device_event_param(void *object, int seq,
 	struct manager *m = o->manager;
 	struct pw_manager_param *p;
 
-	p = add_param(&o->pending_list, id, param);
+	p = add_param(&o->pending_list, seq, o->param_seq, id, param);
+	if (p == NULL)
+		return;
 
 	if (id == SPA_PARAM_Route && !has_param(&o->this.param_list, p)) {
-		uint32_t id, device;
+		uint32_t idx, device;
 		if (spa_pod_parse_object(param,
 				SPA_TYPE_OBJECT_ParamRoute, NULL,
-				SPA_PARAM_ROUTE_index, SPA_POD_Int(&id),
+				SPA_PARAM_ROUTE_index, SPA_POD_Int(&idx),
 				SPA_PARAM_ROUTE_device,  SPA_POD_Int(&device)) < 0)
 			return;
 
@@ -389,7 +420,7 @@ static void node_event_info(void *object, const struct pw_node_info *info)
 
 	pw_log_debug("object %p: id:%d change-mask:%08"PRIx64, o, o->this.id, info->change_mask);
 
-	info = o->this.info = pw_node_info_update(o->this.info, info);
+	info = o->this.info = pw_node_info_merge(o->this.info, info, o->this.changed == 0);
 
 	if (info->change_mask & PW_NODE_CHANGE_MASK_STATE)
 		changed++;
@@ -400,18 +431,26 @@ static void node_event_info(void *object, const struct pw_node_info *info)
 	if (info->change_mask & PW_NODE_CHANGE_MASK_PARAMS) {
 		for (i = 0; i < info->n_params; i++) {
 			uint32_t id = info->params[i].id;
+			int res;
 
 			if (info->params[i].user == 0)
 				continue;
 			info->params[i].user = 0;
 
+			if (id >= MAX_PARAMS) {
+				pw_log_error("too big param id %d", id);
+				continue;
+			}
+
 			changed++;
-			add_param(&o->pending_list, id, NULL);
+			add_param(&o->pending_list, o->param_seq[id], o->param_seq, id, NULL);
 			if (!(info->params[i].flags & SPA_PARAM_INFO_READ))
 				continue;
 
-			pw_node_enum_params((struct pw_node*)o->this.proxy,
-					0, id, 0, -1, NULL);
+			res = pw_node_enum_params((struct pw_node*)o->this.proxy,
+					++o->param_seq[id], id, 0, -1, NULL);
+			if (SPA_RESULT_IS_ASYNC(res))
+				o->param_seq[id] = res;
 		}
 	}
 	if (changed) {
@@ -425,7 +464,7 @@ static void node_event_param(void *object, int seq,
 		const struct spa_pod *param)
 {
 	struct object *o = object;
-	add_param(&o->pending_list, id, param);
+	add_param(&o->pending_list, seq, o->param_seq, id, param);
 }
 
 static const struct pw_node_events node_events = {
@@ -503,7 +542,7 @@ static const struct object_info *find_info(const char *type, uint32_t version)
 {
 	size_t i;
 	for (i = 0; i < SPA_N_ELEMENTS(objects); i++) {
-		if (strcmp(objects[i]->type, type) == 0 &&
+		if (spa_streq(objects[i]->type, type) &&
 		    objects[i]->version <= version)
 			return objects[i];
 	}
@@ -547,6 +586,7 @@ static void registry_event_global(void *data, uint32_t id,
 	struct manager *m = data;
 	struct object *o;
 	const struct object_info *info;
+	const char *str;
 	struct pw_proxy *proxy;
 
 	info = find_info(type, version);
@@ -564,10 +604,15 @@ static void registry_event_global(void *data, uint32_t id,
 		pw_proxy_destroy(proxy);
 		return;
 	}
+	str = props ? spa_dict_lookup(props, PW_KEY_OBJECT_SERIAL) : NULL;
+	if (!spa_atou64(str, &o->this.serial, 0))
+		o->this.serial = SPA_ID_INVALID;
+
 	o->this.id = id;
 	o->this.permissions = permissions;
 	o->this.type = info->type;
 	o->this.version = version;
+	o->this.index = o->this.serial < (1ULL<<32) ? o->this.serial : SPA_ID_INVALID;
 	o->this.props = props ? pw_properties_new_dict(props) : NULL;
 	o->this.proxy = proxy;
 	o->this.creating = true;
@@ -599,12 +644,13 @@ static void registry_event_global_remove(void *object, uint32_t id)
 	struct manager *m = object;
 	struct object *o;
 
-	if ((o = find_object(m, id)) == NULL)
+	if ((o = find_object_by_id(m, id)) == NULL)
 		return;
 
 	o->this.removing = true;
 
-	manager_emit_removed(m, &o->this);
+	if (!o->this.creating)
+		manager_emit_removed(m, &o->this);
 
 	object_destroy(o);
 }
@@ -618,7 +664,7 @@ static const struct pw_registry_events registry_events = {
 static void on_core_info(void *data, const struct pw_core_info *info)
 {
 	struct manager *m = data;
-	m->this.info = pw_core_info_update(m->this.info, info);
+	m->this.info = pw_core_info_merge(m->this.info, info, true);
 }
 
 static void on_core_done(void *data, uint32_t id, int seq)
@@ -650,10 +696,21 @@ static void on_core_done(void *data, uint32_t id, int seq)
 	}
 }
 
+static void on_core_error(void *data, uint32_t id, int seq, int res, const char *message)
+{
+	struct manager *m = data;
+
+	if (id == PW_ID_CORE && res == -EPIPE) {
+		pw_log_debug("connection error: %d, %s", res, message);
+		manager_emit_disconnect(m);
+	}
+}
+
 static const struct pw_core_events core_events = {
 	PW_VERSION_CORE_EVENTS,
 	.done = on_core_done,
 	.info = on_core_info,
+	.error = on_core_error
 };
 
 struct pw_manager *pw_manager_new(struct pw_core *core)
@@ -706,7 +763,7 @@ int pw_manager_set_metadata(struct pw_manager *manager,
 	char buf[1024];
 	char *value;
 
-	if ((s = find_object(m, subject)) == NULL)
+	if ((s = find_object_by_id(m, subject)) == NULL)
 		return -ENOENT;
 	if (!SPA_FLAG_IS_SET(s->this.permissions, PW_PERM_M))
 		return -EACCES;
@@ -742,7 +799,7 @@ int pw_manager_for_each_object(struct pw_manager *manager,
 	int res;
 
 	spa_list_for_each(o, &m->this.object_list, this.link) {
-		if (o->this.creating)
+		if (o->this.creating || o->this.removing)
 			continue;
 		if ((res = callback(data, &o->this)) != 0)
 			return res;
@@ -769,22 +826,22 @@ void pw_manager_destroy(struct pw_manager *manager)
 	free(m);
 }
 
-static struct object_data *object_find_data(struct object *o, const char *id)
+static struct object_data *object_find_data(struct object *o, const char *key)
 {
 	struct object_data *d;
 	spa_list_for_each(d, &o->data_list, link) {
-		if (strcmp(d->id, id) == 0)
+		if (spa_streq(d->key, key))
 			return d;
 	}
 	return NULL;
 }
 
-void *pw_manager_object_add_data(struct pw_manager_object *obj, const char *id, size_t size)
+void *pw_manager_object_add_data(struct pw_manager_object *obj, const char *key, size_t size)
 {
 	struct object *o = SPA_CONTAINER_OF(obj, struct object, this);
 	struct object_data *d;
 
-	d = object_find_data(o, id);
+	d = object_find_data(o, key);
 	if (d != NULL) {
 		if (d->size == size)
 			goto done;
@@ -793,13 +850,21 @@ void *pw_manager_object_add_data(struct pw_manager_object *obj, const char *id,
 	}
 
 	d = calloc(1, sizeof(struct object_data) + size);
-	d->id = id;
+	d->key = key;
 	d->size = size;
 
 	spa_list_append(&o->data_list, &d->link);
 
 done:
-	return SPA_MEMBER(d, sizeof(struct object_data), void);
+	return SPA_PTROFF(d, sizeof(struct object_data), void);
+}
+
+void *pw_manager_object_get_data(struct pw_manager_object *obj, const char *id)
+{
+	struct object *o = SPA_CONTAINER_OF(obj, struct object, this);
+	struct object_data *d = object_find_data(o, id);
+
+	return d ? SPA_PTROFF(d, sizeof(*d), void) : NULL;
 }
 
 int pw_manager_sync(struct pw_manager *manager)
@@ -810,57 +875,57 @@ int pw_manager_sync(struct pw_manager *manager)
 
 bool pw_manager_object_is_client(struct pw_manager_object *o)
 {
-	return strcmp(o->type, PW_TYPE_INTERFACE_Client) == 0;
+	return spa_streq(o->type, PW_TYPE_INTERFACE_Client);
 }
 
 bool pw_manager_object_is_module(struct pw_manager_object *o)
 {
-	return strcmp(o->type, PW_TYPE_INTERFACE_Module) == 0;
+	return spa_streq(o->type, PW_TYPE_INTERFACE_Module);
 }
 
 bool pw_manager_object_is_card(struct pw_manager_object *o)
 {
 	const char *str;
-	return strcmp(o->type, PW_TYPE_INTERFACE_Device) == 0 &&
+	return spa_streq(o->type, PW_TYPE_INTERFACE_Device) &&
 		o->props != NULL &&
 		(str = pw_properties_get(o->props, PW_KEY_MEDIA_CLASS)) != NULL &&
-		strcmp(str, "Audio/Device") == 0;
+		spa_streq(str, "Audio/Device");
 }
 
 bool pw_manager_object_is_sink(struct pw_manager_object *o)
 {
 	const char *str;
-	return strcmp(o->type, PW_TYPE_INTERFACE_Node) == 0 &&
+	return spa_streq(o->type, PW_TYPE_INTERFACE_Node) &&
 		o->props != NULL &&
 		(str = pw_properties_get(o->props, PW_KEY_MEDIA_CLASS)) != NULL &&
-		(strcmp(str, "Audio/Sink") == 0 || strcmp(str, "Audio/Duplex") == 0);
+		(spa_streq(str, "Audio/Sink") || spa_streq(str, "Audio/Duplex"));
 }
 
 bool pw_manager_object_is_source(struct pw_manager_object *o)
 {
 	const char *str;
-	return strcmp(o->type, PW_TYPE_INTERFACE_Node) == 0 &&
+	return spa_streq(o->type, PW_TYPE_INTERFACE_Node) &&
 		o->props != NULL &&
 		(str = pw_properties_get(o->props, PW_KEY_MEDIA_CLASS)) != NULL &&
-		(strcmp(str, "Audio/Source") == 0 ||
-		 strcmp(str, "Audio/Duplex") == 0 ||
-		 strcmp(str, "Audio/Source/Virtual") == 0);
+		(spa_streq(str, "Audio/Source") ||
+		 spa_streq(str, "Audio/Duplex") ||
+		 spa_streq(str, "Audio/Source/Virtual"));
 }
 
 bool pw_manager_object_is_monitor(struct pw_manager_object *o)
 {
 	const char *str;
-	return strcmp(o->type, PW_TYPE_INTERFACE_Node) == 0 &&
+	return spa_streq(o->type, PW_TYPE_INTERFACE_Node) &&
 		o->props != NULL &&
 		(str = pw_properties_get(o->props, PW_KEY_MEDIA_CLASS)) != NULL &&
-		(strcmp(str, "Audio/Sink") == 0);
+		(spa_streq(str, "Audio/Sink"));
 }
 
 bool pw_manager_object_is_virtual(struct pw_manager_object *o)
 {
 	const char *str;
 	struct pw_node_info *info;
-	return strcmp(o->type, PW_TYPE_INTERFACE_Node) == 0 &&
+	return spa_streq(o->type, PW_TYPE_INTERFACE_Node) &&
 		(info = o->info) != NULL && info->props != NULL &&
 		(str = spa_dict_lookup(info->props, PW_KEY_NODE_VIRTUAL)) != NULL &&
 		pw_properties_parse_bool(str);
@@ -874,19 +939,19 @@ bool pw_manager_object_is_source_or_monitor(struct pw_manager_object *o)
 bool pw_manager_object_is_sink_input(struct pw_manager_object *o)
 {
 	const char *str;
-	return strcmp(o->type, PW_TYPE_INTERFACE_Node) == 0 &&
+	return spa_streq(o->type, PW_TYPE_INTERFACE_Node) &&
 		o->props != NULL &&
 		(str = pw_properties_get(o->props, PW_KEY_MEDIA_CLASS)) != NULL &&
-		strcmp(str, "Stream/Output/Audio") == 0;
+		spa_streq(str, "Stream/Output/Audio");
 }
 
 bool pw_manager_object_is_source_output(struct pw_manager_object *o)
 {
 	const char *str;
-	return strcmp(o->type, PW_TYPE_INTERFACE_Node) == 0 &&
+	return spa_streq(o->type, PW_TYPE_INTERFACE_Node) &&
 		o->props != NULL &&
 		(str = pw_properties_get(o->props, PW_KEY_MEDIA_CLASS)) != NULL &&
-		strcmp(str, "Stream/Input/Audio") == 0;
+		spa_streq(str, "Stream/Input/Audio");
 }
 
 bool pw_manager_object_is_recordable(struct pw_manager_object *o)
@@ -896,5 +961,5 @@ bool pw_manager_object_is_recordable(struct pw_manager_object *o)
 
 bool pw_manager_object_is_link(struct pw_manager_object *o)
 {
-	return strcmp(o->type, PW_TYPE_INTERFACE_Link) == 0;
+	return spa_streq(o->type, PW_TYPE_INTERFACE_Link);
 }
