@@ -23,25 +23,28 @@
  * DEALINGS IN THE SOFTWARE.
  */
 
-#ifdef HAVE_CONFIG_H
 #include "config.h"
-#endif
 
 #include <stdio.h>
 #include <dlfcn.h>
 #include <dirent.h>
+#include <limits.h>
 #include <sys/stat.h>
 #include <errno.h>
 
+#include <spa/utils/string.h>
+
 #include "pipewire/impl.h"
 #include "pipewire/private.h"
 
-#define NAME "module"
+PW_LOG_TOPIC_EXTERN(log_module);
+#define PW_LOG_TOPIC_DEFAULT log_module
 
 /** \cond */
 struct impl {
 	struct pw_impl_module this;
 	void *hnd;
+	uint32_t destroy_work_id;
 };
 
 #define pw_module_resource_info(r,...)	pw_resource_call(r,struct pw_module_events,info,0,__VA_ARGS__)
@@ -49,7 +52,7 @@ struct impl {
 
 /** \endcond */
 
-static char *find_module(const char *path, const char *name)
+static char *find_module(const char *path, const char *name, int level)
 {
 	char *filename;
 	struct dirent *entry;
@@ -70,6 +73,8 @@ static char *find_module(const char *path, const char *name)
 	filename = NULL;
 
 	/* now recurse down in subdirectories and look for it there */
+	if (level <= 0)
+		return NULL;
 
 	dir = opendir(path);
 	if (dir == NULL) {
@@ -82,16 +87,16 @@ static char *find_module(const char *path, const char *name)
 	while ((entry = readdir(dir))) {
 		char *newpath;
 
-		if (strcmp(entry->d_name, ".") == 0 || strcmp(entry->d_name, "..") == 0)
+		if (spa_streq(entry->d_name, ".") || spa_streq(entry->d_name, ".."))
 			continue;
 
 		newpath = spa_aprintf("%s/%s", path, entry->d_name);
 		if (newpath == NULL)
 			break;
 
-		if (stat(newpath, &s) == 0 && S_ISDIR(s.st_mode)) {
-			filename = find_module(newpath, name);
-		}
+		if (stat(newpath, &s) == 0 && S_ISDIR(s.st_mode))
+			filename = find_module(newpath, name, level - 1);
+
 		free(newpath);
 
 		if (filename != NULL)
@@ -115,7 +120,7 @@ global_bind(void *_data, struct pw_impl_client *client, uint32_t permissions,
 	if (resource == NULL)
 		goto error_resource;
 
-	pw_log_debug(NAME" %p: bound to %d", this, resource->id);
+	pw_log_debug("%p: bound to %d", this, resource->id);
 	pw_global_add_resource(global, resource);
 
 	this->info.change_mask = PW_MODULE_CHANGE_MASK_ALL;
@@ -125,7 +130,7 @@ global_bind(void *_data, struct pw_impl_client *client, uint32_t permissions,
 	return 0;
 
 error_resource:
-	pw_log_error(NAME" %p: can't create module resource: %m", this);
+	pw_log_error("%p: can't create module resource: %m", this);
 	return -errno;
 }
 
@@ -147,10 +152,9 @@ static const struct pw_global_events global_events = {
  * \param context a \ref pw_context
  * \param name name of the module to load
  * \param args A string with arguments for the module
- * \param[out] error Return location for an error string, or NULL
+ * \param properties extra global properties
  * \return A \ref pw_impl_module if the module could be loaded, or NULL on failure.
  *
- * \memberof pw_impl_module
  */
 SPA_EXPORT
 struct pw_impl_module *
@@ -165,33 +169,42 @@ pw_context_load_module(struct pw_context *context,
 	const char *module_dir;
 	int res;
 	pw_impl_module_init_func_t init_func;
+	const char *state = NULL, *p;
+	size_t len;
+	char path_part[PATH_MAX];
+	static const char * const keys[] = {
+		PW_KEY_OBJECT_SERIAL,
+		PW_KEY_MODULE_NAME,
+		NULL
+	};
 
 	module_dir = getenv("PIPEWIRE_MODULE_DIR");
-	if (module_dir != NULL) {
-		char **l;
-		int i, n_paths;
-
+	if (module_dir == NULL) {
+		module_dir = MODULEDIR;
+		pw_log_debug("moduledir set to: %s", module_dir);
+	}
+	else {
 		pw_log_debug("PIPEWIRE_MODULE_DIR set to: %s", module_dir);
+	}
 
-		l = pw_split_strv(module_dir, "/", 0, &n_paths);
-		for (i = 0; l[i] != NULL; i++) {
-			filename = find_module(l[i], name);
-			if (filename != NULL)
-				break;
+	while ((p = pw_split_walk(module_dir, ":", &len, &state))) {
+		if ((res = spa_scnprintf(path_part, sizeof(path_part), "%.*s", (int)len, p)) > 0) {
+			filename = find_module(path_part, name, 8);
+			if (filename != NULL) {
+				pw_log_debug("trying to load module: %s (%s) args(%s)", name, filename, args);
+
+				hnd = dlopen(filename, RTLD_NOW | RTLD_LOCAL);
+				if (hnd != NULL)
+					break;
+
+				free(filename);
+				filename = NULL;
+			}
 		}
-		pw_free_strv(l);
-	} else {
-		pw_log_debug("moduledir set to: %s", MODULEDIR);
-
-		filename = find_module(MODULEDIR, name);
 	}
 
 	if (filename == NULL)
 		goto error_not_found;
-
-	pw_log_debug("trying to load module: %s (%s) args(%s)", name, filename, args);
-
-	hnd = dlopen(filename, RTLD_NOW | RTLD_LOCAL);
 	if (hnd == NULL)
 		goto error_open_failed;
 
@@ -208,6 +221,7 @@ pw_context_load_module(struct pw_context *context,
 		goto error_no_mem;
 
 	impl->hnd = hnd;
+	impl->destroy_work_id = SPA_ID_INVALID;
 	hnd = NULL;
 
 	this = &impl->this;
@@ -227,21 +241,23 @@ pw_context_load_module(struct pw_context *context,
 	this->global = pw_global_new(context,
 				     PW_TYPE_INTERFACE_Module,
 				     PW_VERSION_MODULE,
-				     pw_properties_new(
-					     PW_KEY_MODULE_NAME, name,
-					     NULL),
+				     NULL,
 				     global_bind,
 				     this);
 
 	if (this->global == NULL)
 		goto error_no_global;
 
-	spa_list_append(&context->module_list, &this->link);
+	spa_list_prepend(&context->module_list, &this->link);
 
 	this->info.id = this->global->id;
 	pw_properties_setf(this->properties, PW_KEY_OBJECT_ID, "%d", this->info.id);
+	pw_properties_setf(this->properties, PW_KEY_OBJECT_SERIAL, "%"PRIu64,
+			pw_global_get_serial(this->global));
 	this->info.props = &this->properties->dict;
 
+	pw_global_update_keys(this->global, &this->properties->dict, keys);
+
 	pw_impl_module_emit_initialized(this);
 
 	pw_global_add_listener(this->global, &this->global_listener, &global_events, this);
@@ -253,7 +269,7 @@ pw_context_load_module(struct pw_context *context,
 
 	pw_impl_module_emit_registered(this);
 
-	pw_log_debug(NAME" %p: loaded module: %s", this, this->info.name);
+	pw_log_debug("%p: loaded module: %s", this, this->info.name);
 
 	return this;
 
@@ -290,22 +306,20 @@ error_free_filename:
 	if (filename)
 		free(filename);
 error_cleanup:
-	if (properties)
-		pw_properties_free(properties);
+	pw_properties_free(properties);
 	errno = -res;
 	return NULL;
 }
 
 /** Destroy a module
  * \param module the module to destroy
- * \memberof pw_impl_module
  */
 SPA_EXPORT
 void pw_impl_module_destroy(struct pw_impl_module *module)
 {
 	struct impl *impl = SPA_CONTAINER_OF(module, struct impl, this);
 
-	pw_log_debug(NAME" %p: destroy", module);
+	pw_log_debug("%p: destroy", module);
 	pw_impl_module_emit_destroy(module);
 
 	if (module->global) {
@@ -314,7 +328,7 @@ void pw_impl_module_destroy(struct pw_impl_module *module)
 		pw_global_destroy(module->global);
 	}
 
-	pw_log_debug(NAME" %p: free", module);
+	pw_log_debug("%p: free", module);
 	pw_impl_module_emit_free(module);
 	free((char *) module->info.name);
 	free((char *) module->info.filename);
@@ -324,8 +338,12 @@ void pw_impl_module_destroy(struct pw_impl_module *module)
 
 	spa_hook_list_clean(&module->listener_list);
 
+	if (impl->destroy_work_id != SPA_ID_INVALID)
+		pw_work_queue_cancel(pw_context_get_work_queue(module->context),
+				     module, SPA_ID_INVALID);
+
 	if (!pw_in_valgrind() && dlclose(impl->hnd) != 0)
-		pw_log_warn(NAME" %p: dlclose failed: %s", module, dlerror());
+		pw_log_warn("%p: dlclose failed: %s", module, dlerror());
 	free(impl);
 }
 
@@ -357,7 +375,7 @@ int pw_impl_module_update_properties(struct pw_impl_module *module, const struct
 	changed = pw_properties_update(module->properties, dict);
 	module->info.props = &module->properties->dict;
 
-	pw_log_debug(NAME" %p: updated %d properties", module, changed);
+	pw_log_debug("%p: updated %d properties", module, changed);
 
 	if (!changed)
 		return 0;
@@ -386,3 +404,20 @@ void pw_impl_module_add_listener(struct pw_impl_module *module,
 {
 	spa_hook_list_append(&module->listener_list, listener, events, data);
 }
+
+static void do_destroy_module(void *obj, void *data, int res, uint32_t id)
+{
+	pw_impl_module_destroy(obj);
+}
+
+SPA_EXPORT
+void pw_impl_module_schedule_destroy(struct pw_impl_module *module)
+{
+	struct impl *impl = SPA_CONTAINER_OF(module, struct impl, this);
+
+	if (impl->destroy_work_id != SPA_ID_INVALID)
+		return;
+
+	impl->destroy_work_id = pw_work_queue_add(pw_context_get_work_queue(module->context),
+						  module, 0, do_destroy_module, NULL);
+}
