@@ -42,14 +42,17 @@
 #include <spa/param/audio/type-info.h>
 #include <spa/param/props.h>
 #include <spa/utils/result.h>
+#include <spa/utils/string.h>
 #include <spa/utils/json.h>
 #include <spa/debug/types.h>
+#include <spa/debug/pod.h>
 
 #include <pipewire/pipewire.h>
-#include <pipewire/global.h>
-#include <extensions/metadata.h>
+#include <pipewire/i18n.h>
+#include <pipewire/extensions/metadata.h>
 
 #include "midifile.h"
+#include "dsffile.h"
 
 #define DEFAULT_MEDIA_TYPE	"Audio"
 #define DEFAULT_MIDI_MEDIA_TYPE	"Midi"
@@ -87,6 +90,10 @@ typedef int (*fill_fn)(struct data *d, void *dest, unsigned int n_frames);
 struct target {
 	struct spa_list link;
 	uint32_t id;
+#define TARGET_TYPE_SINK	0
+#define TARGET_TYPE_SOURCE	1
+#define TARGET_TYPE_STREAM	2
+	uint32_t type;
 	char *name;
 	char *desc;
 	int prio;
@@ -114,7 +121,10 @@ struct data {
 
 	enum mode mode;
 	bool verbose;
-	bool is_midi;
+#define TYPE_PCM	0
+#define TYPE_MIDI	1
+#define TYPE_DSD	2
+	int data_type;
 	const char *remote_name;
 	const char *media_type;
 	const char *media_category;
@@ -151,6 +161,7 @@ struct data {
 	int sync;
 
 	struct spa_io_position *position;
+	struct spa_io_rate_match *rate_match;
 	bool drained;
 	uint64_t clock_time;
 
@@ -158,6 +169,11 @@ struct data {
 		struct midi_file *file;
 		struct midi_file_info info;
 	} midi;
+	struct {
+		struct dsf_file *file;
+		struct dsf_file_info info;
+		struct dsf_layout layout;
+	} dsf;
 };
 
 static inline int
@@ -166,17 +182,19 @@ sf_str_to_fmt(const char *str)
 	if (!str)
 		return -1;
 
-	if (!strcmp(str, "s8"))
+	if (spa_streq(str, "s8"))
 		return SF_FORMAT_PCM_S8;
-	if (!strcmp(str, "s16"))
+	if (spa_streq(str, "u8"))
+		return SF_FORMAT_PCM_U8;
+	if (spa_streq(str, "s16"))
 		return SF_FORMAT_PCM_16;
-	if (!strcmp(str, "s24"))
+	if (spa_streq(str, "s24"))
 		return SF_FORMAT_PCM_24;
-	if (!strcmp(str, "s32"))
+	if (spa_streq(str, "s32"))
 		return SF_FORMAT_PCM_32;
-	if (!strcmp(str, "f32"))
+	if (spa_streq(str, "f32"))
 		return SF_FORMAT_FLOAT;
-	if (!strcmp(str, "f64"))
+	if (spa_streq(str, "f64"))
 		return SF_FORMAT_DOUBLE;
 
 	return -1;
@@ -187,6 +205,8 @@ sf_fmt_to_str(int format)
 {
 	int sub_type = (format & SF_FORMAT_SUBMASK);
 
+	if (sub_type == SF_FORMAT_PCM_U8)
+		return "u8";
 	if (sub_type == SF_FORMAT_PCM_S8)
 		return "s8";
 	if (sub_type == SF_FORMAT_PCM_16)
@@ -202,7 +222,7 @@ sf_fmt_to_str(int format)
 	return "(invalid)";
 }
 
-#define STR_FMTS "(s8|s16|s32|f32|f64)"
+#define STR_FMTS "(u8|s8|s16|s32|f32|f64)"
 
 /* 0 = native, 1 = le, 2 = be */
 static inline int
@@ -221,8 +241,14 @@ sf_format_to_pw(int format)
 		return SPA_AUDIO_FORMAT_UNKNOWN;
 
 	switch (format & SF_FORMAT_SUBMASK) {
+	case SF_FORMAT_PCM_U8:
+		return SPA_AUDIO_FORMAT_U8;
 	case SF_FORMAT_PCM_S8:
 		return SPA_AUDIO_FORMAT_S8;
+	case SF_FORMAT_ULAW:
+		return SPA_AUDIO_FORMAT_ULAW;
+	case SF_FORMAT_ALAW:
+		return SPA_AUDIO_FORMAT_ALAW;
 	case SF_FORMAT_PCM_16:
 		return endianness == 1 ? SPA_AUDIO_FORMAT_S16_LE :
 		       endianness == 2 ? SPA_AUDIO_FORMAT_S16_BE :
@@ -254,6 +280,9 @@ sf_format_samplesize(int format)
 
 	switch (sub_type) {
 	case SF_FORMAT_PCM_S8:
+	case SF_FORMAT_PCM_U8:
+	case SF_FORMAT_ULAW:
+	case SF_FORMAT_ALAW:
 		return 1;
 	case SF_FORMAT_PCM_16:
 		return 2;
@@ -268,11 +297,11 @@ sf_format_samplesize(int format)
 	return -1;
 }
 
-static int sf_playback_fill_s8(struct data *d, void *dest, unsigned int n_frames)
+static int sf_playback_fill_x8(struct data *d, void *dest, unsigned int n_frames)
 {
 	sf_count_t rn;
 
-	rn = sf_read_raw(d->file, dest, n_frames);
+	rn = sf_read_raw(d->file, dest, n_frames * d->stride);
 	return (int)rn;
 }
 
@@ -319,7 +348,10 @@ sf_fmt_playback_fill_fn(int format)
 
 	switch (fmt) {
 	case SPA_AUDIO_FORMAT_S8:
-		return sf_playback_fill_s8;
+	case SPA_AUDIO_FORMAT_U8:
+	case SPA_AUDIO_FORMAT_ULAW:
+	case SPA_AUDIO_FORMAT_ALAW:
+		return sf_playback_fill_x8;
 	case SPA_AUDIO_FORMAT_S16_LE:
 	case SPA_AUDIO_FORMAT_S16_BE:
 		/* sndfile check */
@@ -349,11 +381,11 @@ sf_fmt_playback_fill_fn(int format)
 	return NULL;
 }
 
-static int sf_record_fill_s8(struct data *d, void *src, unsigned int n_frames)
+static int sf_record_fill_x8(struct data *d, void *src, unsigned int n_frames)
 {
 	sf_count_t rn;
 
-	rn = sf_write_raw(d->file, src, n_frames);
+	rn = sf_write_raw(d->file, src, n_frames * d->stride);
 	return (int)rn;
 }
 
@@ -400,7 +432,10 @@ sf_fmt_record_fill_fn(int format)
 
 	switch (fmt) {
 	case SPA_AUDIO_FORMAT_S8:
-		return sf_record_fill_s8;
+	case SPA_AUDIO_FORMAT_U8:
+	case SPA_AUDIO_FORMAT_ULAW:
+	case SPA_AUDIO_FORMAT_ALAW:
+		return sf_record_fill_x8;
 	case SPA_AUDIO_FORMAT_S16_LE:
 	case SPA_AUDIO_FORMAT_S16_BE:
 		/* sndfile check */
@@ -495,7 +530,7 @@ static unsigned int find_channel(const char *name)
 	int i;
 
 	for (i = 0; spa_type_audio_channel[i].name; i++) {
-		if (strcmp(name, spa_debug_type_short_name(spa_type_audio_channel[i].name)) == 0)
+		if (spa_streq(name, spa_debug_type_short_name(spa_type_audio_channel[i].name)))
 			return spa_type_audio_channel[i].type;
 	}
 	return SPA_AUDIO_CHANNEL_UNKNOWN;
@@ -507,7 +542,7 @@ static int parse_channelmap(const char *channel_map, struct channelmap *map)
 	char **ch;
 
 	for (i = 0; i < (int) SPA_N_ELEMENTS(maps); i++) {
-		if (strcmp(maps[i].name, channel_map) == 0) {
+		if (spa_streq(maps[i].name, channel_map)) {
 			map->n_channels = maps[i].channels;
 			spa_memcpy(map->channels, &maps[i].values,
 					map->n_channels * sizeof(unsigned int));
@@ -588,7 +623,7 @@ target_destroy(struct target *target)
 }
 
 static struct target *
-target_create(uint32_t id, const char *name, const char *desc, int prio)
+target_create(uint32_t id, uint32_t type, const char *name, const char *desc, int prio)
 {
 	struct target *target;
 
@@ -596,6 +631,7 @@ target_create(uint32_t id, const char *name, const char *desc, int prio)
 	if (!target)
 		return NULL;
 	target->id = id;
+	target->type = type;
 	target->name = strdup(name);
 	target->desc = strdup(desc ? : "");
 	target->prio = prio;
@@ -612,7 +648,7 @@ static void on_core_info(void *userdata, const struct pw_core_info *info)
 	struct data *data = userdata;
 
 	if (data->verbose)
-		fprintf(stdout, "remote %"PRIu32" is named \"%s\"\n",
+		printf("remote %"PRIu32" is named \"%s\"\n",
 				info->id, info->name);
 }
 
@@ -658,8 +694,8 @@ static int json_object_find(const char *obj, const char *key, char *value, size_
 	if (spa_json_enter_object(&it[0], &it[1]) <= 0)
 		return -EINVAL;
 
-	while (spa_json_get_string(&it[1], k, sizeof(k)-1) > 0) {
-		if (strcmp(k, key) == 0) {
+	while (spa_json_get_string(&it[1], k, sizeof(k)) > 0) {
+		if (spa_streq(k, key)) {
 			if (spa_json_get_string(&it[1], value, len) <= 0)
 				continue;
 			return 0;
@@ -677,13 +713,13 @@ static int metadata_property(void *object,
 	struct data *data = object;
 
 	if (subject == PW_ID_CORE) {
-		if (key == NULL || strcmp(key, "default.audio.sink") == 0) {
+		if (key == NULL || spa_streq(key, "default.audio.sink")) {
 			if (value == NULL ||
 			    json_object_find(value, "name",
 					data->default_sink, sizeof(data->default_sink)) < 0)
 				data->default_sink[0] = '\0';
 		}
-		if (key == NULL || strcmp(key, "default.audio.source") == 0) {
+		if (key == NULL || spa_streq(key, "default.audio.source")) {
 			if (value == NULL ||
 			    json_object_find(value, "name",
 					data->default_source, sizeof(data->default_source)) < 0)
@@ -706,8 +742,8 @@ static void registry_event_global(void *userdata, uint32_t id,
 	const struct spa_dict_item *item;
 	const char *name, *desc, *media_class, *prio_session;
 	int prio;
-	enum mode mode = mode_none;
 	struct target *target;
+	uint32_t ttype;
 
 	/* only once */
 	if (data->targets_listed)
@@ -717,9 +753,12 @@ static void registry_event_global(void *userdata, uint32_t id,
 	if (!data->list_targets)
 		return;
 
-	if (strcmp(type, PW_TYPE_INTERFACE_Metadata) == 0) {
+	if (spa_streq(type, PW_TYPE_INTERFACE_Metadata)) {
 		if (data->metadata != NULL)
 			return;
+		if ((name = spa_dict_lookup(props, PW_KEY_METADATA_NAME)) != NULL &&
+		    !spa_streq(name, "default"))
+			return;
 
 		data->metadata = pw_registry_bind(data->registry,
 				id, type, PW_VERSION_METADATA, 0);
@@ -729,7 +768,7 @@ static void registry_event_global(void *userdata, uint32_t id,
 
 		data->sync = pw_core_sync(data->core, 0, data->sync);
 
-	} else if (strcmp(type, PW_TYPE_INTERFACE_Node) == 0) {
+	} else if (spa_streq(type, PW_TYPE_INTERFACE_Node)) {
 		name = spa_dict_lookup(props, PW_KEY_NODE_NAME);
 		desc = spa_dict_lookup(props, PW_KEY_NODE_DESCRIPTION);
 		media_class = spa_dict_lookup(props, PW_KEY_MEDIA_CLASS);
@@ -739,15 +778,27 @@ static void registry_event_global(void *userdata, uint32_t id,
 		if (!name || !media_class)
 			return;
 
-		/* get allowed mode from the media class */
-		/* TODO extend to something else besides Audio/Source|Sink */
-		if (!strcmp(media_class, "Audio/Source"))
-			mode = mode_record;
-		else if (!strcmp(media_class, "Audio/Sink"))
-			mode = mode_playback;
+		if (desc == NULL)
+			desc = name;
 
-		/* modes must match */
-		if (mode != data->mode)
+		/* get allowed mode from the media class */
+		if (spa_streq(media_class, "Audio/Source")) {
+			if (data->mode != mode_record)
+				return;
+			ttype = TARGET_TYPE_SOURCE;
+		}
+		else if (spa_streq(media_class, "Stream/Output/Audio")) {
+			if (data->mode != mode_record)
+				return;
+			ttype = TARGET_TYPE_STREAM;
+		}
+		else if (spa_streq(media_class, "Audio/Sink")) {
+			if (data->mode != mode_playback &&
+			    data->mode != mode_record)
+				return;
+			ttype = TARGET_TYPE_SINK;
+		}
+		else
 			return;
 
 		prio = prio_session ? atoi(prio_session) : -1;
@@ -757,11 +808,11 @@ static void registry_event_global(void *userdata, uint32_t id,
 					id, type, name, media_class, desc ? : "", prio);
 
 			spa_dict_for_each(item, props) {
-				fprintf(stdout, "\t\t%s = \"%s\"\n", item->key, item->value);
+				printf("\t\t%s = \"%s\"\n", item->key, item->value);
 			}
 		}
 
-		target = target_create(id, name, desc, prio);
+		target = target_create(id, ttype, name, desc, prio);
 		if (target)
 			spa_list_append(&data->targets, &target->link);
 	}
@@ -828,19 +879,50 @@ on_io_changed(void *userdata, uint32_t id, void *data, uint32_t size)
 	case SPA_IO_Position:
 		d->position = data;
 		break;
+	case SPA_IO_RateMatch:
+		if (d->data_type == TYPE_PCM)
+			d->rate_match = data;
+		break;
 	default:
 		break;
 	}
 }
 
 static void
-on_param_changed(void *userdata, uint32_t id, const struct spa_pod *format)
+on_param_changed(void *userdata, uint32_t id, const struct spa_pod *param)
 {
 	struct data *data = userdata;
+	struct spa_audio_info info = { 0 };
+	int err;
 
 	if (data->verbose)
-		printf("stream param change: id=%"PRIu32"\n",
-				id);
+		printf("stream param change: id=%"PRIu32"\n", id);
+
+	if (id != SPA_PARAM_Format || param == NULL)
+		return;
+
+	if ((err = spa_format_parse(param, &info.media_type, &info.media_subtype)) < 0)
+		return;
+
+	if (info.media_type != SPA_MEDIA_TYPE_audio ||
+	    info.media_subtype != SPA_MEDIA_SUBTYPE_dsd)
+		return;
+
+	if (spa_format_audio_dsd_parse(param, &info.info.dsd) < 0)
+		return;
+
+	data->dsf.layout.interleave = info.info.dsd.interleave,
+	data->dsf.layout.channels = info.info.dsd.channels;
+	data->dsf.layout.lsb = info.info.dsd.bitorder == SPA_PARAM_BITORDER_lsb;
+
+	data->stride = data->dsf.info.channels * SPA_ABS(data->dsf.layout.interleave);
+
+	if (data->verbose) {
+		printf("DSD out: channels:%d bitorder:%s interleave:%d\n",
+				data->dsf.layout.channels,
+				data->dsf.layout.lsb ? "lsb" : "msb",
+				data->dsf.layout.interleave);
+	}
 }
 
 static void on_process(void *userdata)
@@ -868,6 +950,8 @@ static void on_process(void *userdata)
 	if (data->mode == mode_playback) {
 
 		n_frames = d->maxsize / data->stride;
+		if (data->rate_match && data->rate_match->size > 0)
+			n_frames = SPA_MIN((uint32_t)n_frames, data->rate_match->size);
 
 		n_fill_frames = data->fill(data, p, n_frames);
 
@@ -931,7 +1015,7 @@ static void do_print_delay(void *userdata, uint64_t expirations)
 	struct data *data = userdata;
 	struct pw_time time;
 	pw_stream_get_time(data->stream, &time);
-	printf("now=%li rate=%u/%u ticks=%lu delay=%li queued=%lu\n",
+	printf("now=%"PRIi64" rate=%u/%u ticks=%"PRIu64" delay=%"PRIi64" queued=%"PRIu64"\n",
 		time.now,
 		time.rate.num, time.rate.denom,
 		time.ticks, time.delay, time.queued);
@@ -1028,11 +1112,12 @@ static void show_usage(const char *name, bool is_error)
 	     DEFAULT_VOLUME,
 	     DEFAULT_QUALITY);
 
-	if (!strcmp(name, "pw-cat")) {
+	if (spa_streq(name, "pw-cat")) {
 		fputs(
 		   _("  -p, --playback                        Playback mode\n"
 		     "  -r, --record                          Recording mode\n"
 		     "  -m, --midi                            Midi mode\n"
+		     "  -d, --dsd                             DSD mode\n"
 		     "\n"), fp);
 	}
 }
@@ -1152,9 +1237,50 @@ static int setup_midifile(struct data *data)
 	return 0;
 }
 
+struct dsd_layout_info {
+	 uint32_t type;
+	 struct spa_audio_layout_info info;
+};
+static const struct dsd_layout_info dsd_layouts[] = {
+	{ 1, { SPA_AUDIO_LAYOUT_Mono, }, },
+	{ 2, { SPA_AUDIO_LAYOUT_Stereo, }, },
+	{ 3, { SPA_AUDIO_LAYOUT_2FC }, },
+	{ 4, { SPA_AUDIO_LAYOUT_Quad }, },
+	{ 5, { SPA_AUDIO_LAYOUT_3_1 }, },
+	{ 6, { SPA_AUDIO_LAYOUT_5_0R }, },
+	{ 7, { SPA_AUDIO_LAYOUT_5_1R }, },
+};
+
+static int dsf_play(struct data *d, void *src, unsigned int n_frames)
+{
+	return dsf_file_read(d->dsf.file, src, n_frames, &d->dsf.layout);
+}
+
+static int setup_dsffile(struct data *data)
+{
+	if (data->mode == mode_record)
+		return -ENOTSUP;
+
+	data->dsf.file = dsf_file_open(data->filename, "r", &data->dsf.info);
+	if (data->dsf.file == NULL) {
+		fprintf(stderr, "error: can't read dsf file '%s': %m\n", data->filename);
+		return -errno;
+	}
+
+	if (data->verbose)
+		printf("opened file \"%s\" channels:%d rate:%d bitorder:%s\n",
+				data->filename,
+				data->dsf.info.channels, data->dsf.info.rate,
+				data->dsf.info.lsb ? "lsb" : "msb");
+
+	data->fill = dsf_play;
+;
+	return 0;
+}
+
 static int fill_properties(struct data *data)
 {
-	static const char* table[] = {
+	static const char * const table[] = {
 		[SF_STR_TITLE] = PW_KEY_MEDIA_TITLE,
 		[SF_STR_COPYRIGHT] = PW_KEY_MEDIA_COPYRIGHT,
 		[SF_STR_SOFTWARE] = PW_KEY_MEDIA_SOFTWARE,
@@ -1162,6 +1288,7 @@ static int fill_properties(struct data *data)
 		[SF_STR_COMMENT] = PW_KEY_MEDIA_COMMENT,
 		[SF_STR_DATE] = PW_KEY_MEDIA_DATE
 	};
+
 	SF_INFO sfi;
 	SF_FORMAT_INFO fi;
 	int res;
@@ -1291,15 +1418,15 @@ static int setup_sndfile(struct data *data)
 		s++;
 	if (!*s)
 		data->latency_unit = unit_samples;
-	else if (!strcmp(s, "none"))
+	else if (spa_streq(s, "none"))
 		data->latency_unit = unit_none;
-	else if (!strcmp(s, "s") || !strcmp(s, "sec") || !strcmp(s, "secs"))
+	else if (spa_streq(s, "s") || spa_streq(s, "sec") || spa_streq(s, "secs"))
 		data->latency_unit = unit_sec;
-	else if (!strcmp(s, "ms") || !strcmp(s, "msec") || !strcmp(s, "msecs"))
+	else if (spa_streq(s, "ms") || spa_streq(s, "msec") || spa_streq(s, "msecs"))
 		data->latency_unit = unit_msec;
-	else if (!strcmp(s, "us") || !strcmp(s, "usec") || !strcmp(s, "usecs"))
+	else if (spa_streq(s, "us") || spa_streq(s, "usec") || spa_streq(s, "usecs"))
 		data->latency_unit = unit_usec;
-	else if (!strcmp(s, "ns") || !strcmp(s, "nsec") || !strcmp(s, "nsecs"))
+	else if (spa_streq(s, "ns") || spa_streq(s, "nsec") || spa_streq(s, "nsecs"))
 		data->latency_unit = unit_nsec;
 	else {
 		fprintf(stderr, "error: bad latency value %s (bad unit)\n", data->latency);
@@ -1338,9 +1465,12 @@ static int setup_sndfile(struct data *data)
 				sf_fmt_to_str(info.format),
 				data->samplesize,
 				data->stride, nom, (double)nom/data->rate);
+
 	if (nom)
 		pw_properties_setf(data->props, PW_KEY_NODE_LATENCY, "%u/%u", nom, data->rate);
 
+	pw_properties_setf(data->props, PW_KEY_NODE_RATE, "1/%u", data->rate);
+
 	if (data->quality >= 0)
 		pw_properties_setf(data->props, "resample.quality", "%d", data->quality);
 
@@ -1369,16 +1499,21 @@ int main(int argc, char *argv[])
 		prog = argv[0];
 
 	/* prime the mode from the program name */
-	if (!strcmp(prog, "pw-play"))
+	if (spa_streq(prog, "pw-play")) {
 		data.mode = mode_playback;
-	else if (!strcmp(prog, "pw-record"))
+		data.data_type = TYPE_PCM;
+	} else if (spa_streq(prog, "pw-record")) {
 		data.mode = mode_record;
-	else if (!strcmp(prog, "pw-midiplay")) {
+		data.data_type = TYPE_PCM;
+	} else if (spa_streq(prog, "pw-midiplay")) {
 		data.mode = mode_playback;
-		data.is_midi = true;
-	} else if (!strcmp(prog, "pw-midirecord")) {
+		data.data_type = TYPE_MIDI;
+	} else if (spa_streq(prog, "pw-midirecord")) {
 		data.mode = mode_record;
-		data.is_midi = true;
+		data.data_type = TYPE_MIDI;
+	} else if (spa_streq(prog, "pw-dsdplay")) {
+		data.mode = mode_playback;
+		data.data_type = TYPE_DSD;
 	} else
 		data.mode = mode_none;
 
@@ -1389,7 +1524,7 @@ int main(int argc, char *argv[])
 	/* initialize list every time */
 	spa_list_init(&data.targets);
 
-	while ((c = getopt_long(argc, argv, "hvprmR:q:", long_options, NULL)) != -1) {
+	while ((c = getopt_long(argc, argv, "hvprmdR:q:", long_options, NULL)) != -1) {
 
 		switch (c) {
 
@@ -1398,7 +1533,7 @@ int main(int argc, char *argv[])
 			return EXIT_SUCCESS;
 
 		case OPT_VERSION:
-			fprintf(stdout, "%s\n"
+			printf("%s\n"
 				"Compiled with libpipewire %s\n"
 				"Linked with libpipewire %s\n",
 				prog,
@@ -1419,7 +1554,11 @@ int main(int argc, char *argv[])
 			break;
 
 		case 'm':
-			data.is_midi = true;
+			data.data_type = TYPE_MIDI;
+			break;
+
+		case 'd':
+			data.data_type = TYPE_DSD;
 			break;
 
 		case 'R':
@@ -1444,7 +1583,7 @@ int main(int argc, char *argv[])
 
 		case OPT_TARGET:
 			data.target = optarg;
-			if (!strcmp(optarg, "auto")) {
+			if (spa_streq(optarg, "auto")) {
 				data.target_id = PW_ID_ANY;
 				break;
 			}
@@ -1509,10 +1648,14 @@ int main(int argc, char *argv[])
 	}
 
 	if (!data.media_type) {
-		if (data.is_midi)
+		switch (data.data_type) {
+		case TYPE_MIDI:
 			data.media_type = DEFAULT_MIDI_MEDIA_TYPE;
-		else
+			break;
+		default:
 			data.media_type = DEFAULT_MEDIA_TYPE;
+			break;
+		}
 	}
 	if (!data.media_category)
 		data.media_category = data.mode == mode_playback ?
@@ -1608,12 +1751,21 @@ int main(int argc, char *argv[])
 	data.sync = pw_core_sync(data.core, 0, data.sync);
 
 	if (!data.list_targets) {
-		struct spa_audio_info_raw info;
 
-		if (data.is_midi)
-			ret = setup_midifile(&data);
-		else
+		switch (data.data_type) {
+		case TYPE_PCM:
 			ret = setup_sndfile(&data);
+			break;
+		case TYPE_MIDI:
+			ret = setup_midifile(&data);
+			break;
+		case TYPE_DSD:
+			ret = setup_dsffile(&data);
+			break;
+		default:
+			ret = -ENOTSUP;
+			break;
+		}
 
 		if (ret < 0) {
 			fprintf(stderr, "error: open failed: %s\n", spa_strerror(ret));
@@ -1625,8 +1777,10 @@ int main(int argc, char *argv[])
 				goto error_usage;
 			}
 		}
-
-		if (!data.is_midi) {
+		switch (data.data_type) {
+		case TYPE_PCM:
+		{
+			struct spa_audio_info_raw info;
 			info = SPA_AUDIO_INFO_RAW_INIT(
 				.flags = data.channelmap.n_channels ? 0 : SPA_AUDIO_FLAG_UNPOSITIONED,
 				.format = data.spa_format,
@@ -1637,13 +1791,35 @@ int main(int argc, char *argv[])
 				memcpy(info.position, data.channelmap.channels, data.channels * sizeof(int));
 
 			params[0] = spa_format_audio_raw_build(&b, SPA_PARAM_EnumFormat, &info);
-		} else {
+			break;
+		}
+		case TYPE_MIDI:
 			params[0] = spa_pod_builder_add_object(&b,
 					SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat,
 					SPA_FORMAT_mediaType,		SPA_POD_Id(SPA_MEDIA_TYPE_application),
 					SPA_FORMAT_mediaSubtype,	SPA_POD_Id(SPA_MEDIA_SUBTYPE_control));
 
 			pw_properties_set(data.props, PW_KEY_FORMAT_DSP, "8 bit raw midi");
+			break;
+		case TYPE_DSD:
+		{
+			struct spa_audio_info_dsd info;
+			size_t i;
+
+			spa_zero(info);
+			info.channels = data.dsf.info.channels;
+			info.rate = data.dsf.info.rate / 8;
+
+			for (i = 0; i < SPA_N_ELEMENTS(dsd_layouts); i++) {
+				if (dsd_layouts[i].type != data.dsf.info.channel_type)
+					continue;
+				info.channels = dsd_layouts[i].info.n_channels;
+				memcpy(info.position, dsd_layouts[i].info.position,
+						info.channels * sizeof(uint32_t));
+			}
+			params[0] = spa_format_audio_dsd_build(&b, SPA_PARAM_EnumFormat, &info);
+			break;
+		}
 		}
 
 		data.stream = pw_stream_new(data.core, prog, data.props);
@@ -1714,14 +1890,31 @@ int main(int argc, char *argv[])
 			target_default = NULL;
 			spa_list_for_each(target, &data.targets, link) {
 				if (target_default == NULL ||
-				    strcmp(default_name, target->name) == 0 ||
+				    spa_streq(default_name, target->name) ||
 				    (default_name[0] == '\0' &&
 				     target->prio > target_default->prio))
 					target_default = target;
 			}
 			printf("Available targets (\"*\" denotes default): %s\n", default_name);
 			spa_list_for_each(target, &data.targets, link) {
-				printf("%s\t%"PRIu32": description=\"%s\" prio=%d\n",
+				if (target->type != TARGET_TYPE_SOURCE)
+					continue;
+				printf("%s\t%"PRIu32": source description=\"%s\" prio=%d\n",
+				       target == target_default ? "*" : "",
+				       target->id, target->desc, target->prio);
+			}
+			spa_list_for_each(target, &data.targets, link) {
+				if (target->type != TARGET_TYPE_SINK)
+					continue;
+				printf("%s\t%"PRIu32": %s description=\"%s\" prio=%d\n",
+					target == target_default ? "*" : "",
+					target->id, data.mode == mode_record ? "monitor" : "sink",
+					target->desc, target->prio);
+			}
+			spa_list_for_each(target, &data.targets, link) {
+				if (target->type != TARGET_TYPE_STREAM)
+					continue;
+				printf("%s\t%"PRIu32": stream description=\"%s\" prio=%d\n",
 				       target == target_default ? "*" : "",
 				       target->id, target->desc, target->prio);
 			}
@@ -1737,14 +1930,21 @@ int main(int argc, char *argv[])
 	}
 
 error_connect_fail:
-	if (data.stream)
+	if (data.stream) {
+		spa_hook_remove(&data.stream_listener);
 		pw_stream_destroy(data.stream);
+	}
 error_no_stream:
-	if (data.metadata)
+	if (data.metadata) {
+		spa_hook_remove(&data.metadata_listener);
 		pw_proxy_destroy((struct pw_proxy*)data.metadata);
-	if (data.registry)
+	}
+	if (data.registry) {
+		spa_hook_remove(&data.registry_listener);
 		pw_proxy_destroy((struct pw_proxy*)data.registry);
+	}
 error_no_registry:
+	spa_hook_remove(&data.core_listener);
 	pw_core_disconnect(data.core);
 error_ctx_connect_failed:
 	pw_context_destroy(data.context);
@@ -1753,8 +1953,7 @@ error_no_context:
 error_no_props:
 error_no_main_loop:
 error_bad_file:
-	if (data.props)
-		pw_properties_free(data.props);
+	pw_properties_free(data.props);
 	if (data.file)
 		sf_close(data.file);
 	if (data.midi.file)
