@@ -27,6 +27,7 @@
 #include <getopt.h>
 
 #include <spa/utils/result.h>
+#include <spa/utils/string.h>
 #include <spa/debug/pod.h>
 #include <spa/debug/format.h>
 #include <spa/debug/types.h>
@@ -505,7 +506,7 @@ static int draw_graph(struct data *d, const char *path)
 	/* draw the footer */
 	dot_str_add(&d->dot_str, "}\n");
 
-	if (strcmp(path, "-") == 0) {
+	if (spa_streq(path, "-")) {
 		/* wire the dot graph into to stdout */
 		fputs(d->dot_str, stdout);
 	} else {
@@ -608,8 +609,9 @@ static void removed_proxy(void *user_data)
 static void destroy_proxy(void *user_data)
 {
 	struct global *g = user_data;
-	if (g->props)
-		pw_properties_free(g->props);
+	spa_hook_remove(&g->object_listener);
+	spa_hook_remove(&g->proxy_listener);
+	pw_properties_free(g->props);
 	if (g->info)
 		g->info_destroy(g->info);
 }
@@ -634,7 +636,7 @@ static void registry_event_global(void *data, uint32_t id, uint32_t permissions,
         draw_t draw;
         struct global *g;
 
-	if (strcmp(type, PW_TYPE_INTERFACE_Port) == 0) {
+	if (spa_streq(type, PW_TYPE_INTERFACE_Port)) {
 		events = &port_events;
 		info_destroy = (pw_destroy_t)pw_port_info_free;
 		info_update = (info_update_t)pw_port_info_update;
@@ -642,7 +644,7 @@ static void registry_event_global(void *data, uint32_t id, uint32_t permissions,
 		client_version = PW_VERSION_PORT;
 		object_type = INTERFACE_Port;
 	}
-	else if (strcmp(type, PW_TYPE_INTERFACE_Node) == 0) {
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Node)) {
 		events = &node_events;
 		info_destroy = (pw_destroy_t)pw_node_info_free;
 		info_update = (info_update_t)pw_node_info_update;
@@ -650,7 +652,7 @@ static void registry_event_global(void *data, uint32_t id, uint32_t permissions,
 		client_version = PW_VERSION_NODE;
 		object_type = INTERFACE_Node;
 	}
-	else if (strcmp(type, PW_TYPE_INTERFACE_Link) == 0) {
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Link)) {
 		events = &link_events;
 		info_destroy = (pw_destroy_t)pw_link_info_free;
 		info_update = (info_update_t)pw_link_info_update;
@@ -658,7 +660,7 @@ static void registry_event_global(void *data, uint32_t id, uint32_t permissions,
 		client_version = PW_VERSION_LINK;
 		object_type = INTERFACE_Link;
 	}
-	else if (strcmp(type, PW_TYPE_INTERFACE_Client) == 0) {
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Client)) {
 		events = &client_events;
 		info_destroy = (pw_destroy_t)pw_client_info_free;
 		info_update = (info_update_t)pw_client_info_update;
@@ -666,7 +668,7 @@ static void registry_event_global(void *data, uint32_t id, uint32_t permissions,
 		client_version = PW_VERSION_CLIENT;
 		object_type = INTERFACE_Client;
 	}
-	else if (strcmp(type, PW_TYPE_INTERFACE_Device) == 0) {
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Device)) {
 		events = &device_events;
 		info_destroy = (pw_destroy_t)pw_device_info_free;
 		info_update = (info_update_t)pw_device_info_update;
@@ -674,7 +676,7 @@ static void registry_event_global(void *data, uint32_t id, uint32_t permissions,
 		client_version = PW_VERSION_DEVICE;
 		object_type = INTERFACE_Device;
 	}
-	else if (strcmp(type, PW_TYPE_INTERFACE_Factory) == 0) {
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Factory)) {
 		events = &factory_events;
 		info_destroy = (pw_destroy_t)pw_factory_info_free;
 		info_update = (info_update_t)pw_factory_info_update;
@@ -682,7 +684,7 @@ static void registry_event_global(void *data, uint32_t id, uint32_t permissions,
 		client_version = PW_VERSION_FACTORY;
 		object_type = INTERFACE_Factory;
 	}
-	else if (strcmp(type, PW_TYPE_INTERFACE_Module) == 0) {
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Module)) {
 		events = &module_events;
 		info_destroy = (pw_destroy_t)pw_module_info_free;
 		info_update = (info_update_t)pw_module_info_update;
@@ -690,7 +692,7 @@ static void registry_event_global(void *data, uint32_t id, uint32_t permissions,
 		client_version = PW_VERSION_MODULE;
 		object_type = INTERFACE_Module;
 	}
-	else if (strcmp(type, PW_TYPE_INTERFACE_Core) == 0) {
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Core)) {
 		/* sync to notify we are done with globals */
 		pw_core_sync(d->core, 0, 0);
 		return;
@@ -760,9 +762,9 @@ static void do_quit(void *data, int signal_number)
 	pw_main_loop_quit(d->loop);
 }
 
-static void show_help(const char *name)
+static void show_help(const char *name, bool error)
 {
-        fprintf(stdout, "%s [options]\n"
+        fprintf(error ? stderr : stdout, "%s [options]\n"
 		"  -h, --help                            Show this help\n"
 		"      --version                         Show version\n"
 		"  -a, --all                             Show all object types\n"
@@ -801,10 +803,10 @@ int main(int argc, char *argv[])
 	while ((c = getopt_long(argc, argv, "hVasdr:o:L9", long_options, NULL)) != -1) {
 		switch (c) {
 		case 'h' :
-			show_help(argv[0]);
+			show_help(argv[0], false);
 			return 0;
 		case 'V' :
-			fprintf(stdout, "%s\n"
+			printf("%s\n"
 				"Compiled with libpipewire %s\n"
 				"Linked with libpipewire %s\n",
 				argv[0],
@@ -840,7 +842,7 @@ int main(int argc, char *argv[])
 			fprintf(stderr, "orthogonal edges enabled\n");
 			break;
 		default:
-			show_help(argv[0]);
+			show_help(argv[0], true);
 			return -1;
 		}
 	}
@@ -891,7 +893,9 @@ int main(int argc, char *argv[])
 	draw_graph(&data, dot_path);
 
 	dot_str_clear(&data.dot_str);
+	spa_hook_remove(&data.registry_listener);
 	pw_proxy_destroy((struct pw_proxy*)data.registry);
+	spa_hook_remove(&data.core_listener);
 	pw_context_destroy(data.context);
 	pw_main_loop_destroy(data.loop);
 	pw_deinit();
